package edu.northwestern.cs.aqualab.cattle;

import java.io.Serializable;
import java.util.UUID;

import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.cs.aqualab.cattle.logging.Format;
import edu.northwestern.cs.aqualab.cattle.logging.Logging;



/**
 * Represents an individual data point to send home. The actual data is stored
 * in a {@link JSONObject} which can be accessed and modified via
 * {@link #getJSONObject()}. Instances simply represent units of data and must
 * be passed to a {@link ReportManager} for delivery.<br>
 * <br>
 * It is not safe to pass a Report to {@link ReportManager#send} twice. Doing so
 * will throw an IllegalStateException.
 */
public class Report implements Serializable {
    private static final long serialVersionUID = 1L;

    // Per-report configuration. Any field that still has an invalid value will
    // be set to ReportManger's defaults during send().
    protected String mHost = null;
    protected int mPort = -1;
    protected String mProject = null;
    protected String mDataset;  // Will always be specified by a constructor.
    protected JSONObject mBeef = new JSONObject();

    // Used to produce a unique filename for each report. Does not go out on the
    // wire.
    protected UUID mUUID = UUID.randomUUID();

    // Used to determine which reports should be deleted when disk usage reaches
    // the limit.
    protected long createdAt = System.currentTimeMillis();

    // Used to identify when a report needs to have its UUID rewritten.
    protected boolean mPreviouslyAsyncSent = false;

    // Bookkeeping for exponential backoff.
    protected long last_attempt = 0;
    protected int attempts = 0;


    /**
     * Construct a <code>Report</code> with the default project (as passed to
     * {@link ReportManager#init}).
     *
     * @param dataset
     *            The dataset to report to.
     */
    public Report(String dataset) {
        mDataset = dataset;
        Logging.debug("Initialized: %s", this.toString());
    }


    /**
     * Construct a <code>Report</code>, overriding the default project.
     *
     * @param project
     *            The project to use instead of the default.
     * @param dataset
     *            The dataset to report to.
     */
    public Report(String project, String dataset) {
        mProject = project;
        mDataset = dataset;
        Logging.debug("Report(): %s... pointed at %s/%s", mUUID.toString().substring(0, 6),
                      project, dataset);
    }


    /**
     * Two Reports are equal if their UUIDs match. This is helpful for comparing
     * instances before and after serialization.
     */
    @Override
    public boolean equals(Object other) {
        if (other == null) {
            return false;
        }

        if (!(other instanceof Report)) {
            return false;
        }

        Report otherReport = (Report) other;
        return this.mUUID.equals(otherReport.mUUID);
    }


    /**
     * This is overridden to maintain the contracts of both
     * <code>hashCode</code> and <code>equals</code>.
     */
    @Override
    public int hashCode() {
        return mUUID.hashCode();
    }


    @Override
    public String toString() {
        long age = System.currentTimeMillis() - createdAt;
        StringBuilder sb = new StringBuilder();
        sb.append("<Report ");
        sb.append(mUUID.toString().substring(0, 5));
        sb.append("... (");
        sb.append(Format.approximateFamiliarInterval(age));
        sb.append(" old, attempted ");
        sb.append(String.valueOf(attempts));
        sb.append(" times) facing ");
        sb.append(mHost);
        sb.append(":");
        sb.append(String.valueOf(mPort));
        sb.append("/");
        sb.append(mProject);
        sb.append("/");
        sb.append(mDataset);
        sb.append(">");
        return sb.toString();
    }


    /**
     * @return The {@link JSONObject} backing this Report.
     */
    public JSONObject getJSONObject() {
        return mBeef;
    }


    /**
     * Set the JSONObject backing this Report.
     */
    public void setJSONObject(JSONObject freshmeat) {
        mBeef = freshmeat;
    }


    /**
     * @return The body of this Report in the format it should have on the wire.
     */
    public String getBody() {
        return mBeef.toJSONString();
    }
    
    /**
     * @return The unique UUID for the report. Does not go out on the wire.
     */
    public UUID getUuid() {
        return mUUID;
    }

    /**
     * @return Creation time of this report.
     */
    public long getCreatedAt() {
        return createdAt;
    }

    
}
