package edu.northwestern.cs.aqualab.cattle;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Writer;
import java.net.Socket;
import java.net.UnknownHostException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.logging.Level;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocketFactory;

import edu.northwestern.cs.aqualab.cattle.exception.CattleInitException;
import edu.northwestern.cs.aqualab.cattle.exception.CattleStateException;
import edu.northwestern.cs.aqualab.cattle.filesystem.Filesystem;
import edu.northwestern.cs.aqualab.cattle.filesystem.FilesystemImpl;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONValue;
import edu.northwestern.cs.aqualab.cattle.logging.Logger;
import edu.northwestern.cs.aqualab.cattle.logging.Logging;
import edu.northwestern.cs.aqualab.cattle.socket.SelfSignedSSLSocketFactory;

/**
 * Singleton class that provides reliable delivery of {@link Report} instances.
 * It maintains some module-level state, necessitating an {@link #init} that
 * must be called before you can get an instance. <br>
 * <br>
 * Assuming you give a persistent directory to <code>init</code>, a
 * <code>ReportManager</code> instance will attempt to deliver all reports
 * passed in through {@link #send}, performing exponential backoff with those
 * that fail. It will eventually abandon reports, but if it does so, the first
 * and last tries will be at least a week apart.
 */
public class ReportManager {
    private static final String POST_FORMAT = "POST %s HTTP/1.0\r\n" + "Content-Length: %d\r\n"
                                              + "\r\n" + "%s\r\n";
    private static final String POST_PERSISTENT_FORMAT = "POST %s HTTP/1.0\r\n" + "Content-Length: %d\r\n" + 
                                               "Connection: Keep-Alive\r\n" + "\r\n" + "%s\r\n";

    // State under the hood.
    private static boolean sInit = false;
    private static Boolean sReportingEnabled = true;
    private static ReportManager sInstance = null;
    private static Object sInstanceLock = new Object();
    private static List<KeyStore> sTrustedKeystores = new ArrayList<KeyStore>();
    private static SocketFactory sSocketFactory = SSLSocketFactory.getDefault();
    private static ReportingThread sReportingThread = null;
    private static Object sReportingThreadLock = new Object();

    // Reporting configuration.
    private static UUID sClientUUID = null;
    private static Object sClientUUIDLock = new Object();
    private static ConcurrentMap<String, Callable<Object>> sAutokeys = makeDefaultAutokeys();
    private static String sWorkingDir = null;
    private static String sThreadNamePrefix = "";

    public static boolean sDebugMode = false;

    /*
     * The values are not used directly when sending reports. Instead, Report
     * constructors use them in lieu of default arguments. This makes it
     * possible to extend this API to allow changing any parameter on a
     * per-report basis with minimum pain.
     */
    private static String sDefaultHost = null;
    private static int sDefaultPort = 0;
    private static boolean sDefaultUseSSL = false;
    private static String sDefaultProject = null;

    protected static Filesystem filesystem = null;
    /*
     * Load the keystore with our self-signed cert. This works from inside the
     * jar.
     */
    static {
        String errmsg = "Couldn't load keystore (bad jar?): ";

        InputStream inputStream = ReportManager.class.getResourceAsStream(Const.KEYSTORE_PATH);
        if (inputStream == null) {
            throw new CattleInitException(errmsg + "Couldn't get resource!");
        }

        KeyStore ks = null;
        Exception exc = new Exception();
        try {
            ks = KeyStore.getInstance("JKS");
            ks.load(inputStream, "fadazani".toCharArray());
        } catch (KeyStoreException kse) {
            exc = kse;
        } catch (NoSuchAlgorithmException nsae) {
            exc = nsae;
        } catch (CertificateException ce) {
            exc = ce;
        } catch (IOException ioe) {
            exc = ioe;
        }

        if (ks == null) {
            throw new CattleInitException(errmsg + exc.getMessage());
        }
        ReportManager.addKeyStore(ks);
        
        // Default filesystem uses OS implementation
        filesystem = FilesystemImpl.getInstance();
    }

    /**
     * This must be invoked before calling {@link #getInstance}.<br>
     * <br>
     * Set default values for several members of {@link Report}. If appropriate,
     * these values will be applied to instances of that class when they are
     * passed as arguments to {@link #send}.<br>
     * <br>
     * While not strictly necessary, it is a good idea to call
     * {@link #addKeyStore} at the same time and place you call this method.
     *
     * @param host
     *            The domain running the wrangler server.
     * @param port
     *            The port the wrangler server is listening on.
     * @param useSSL
     *            Whether to enable SSL.
     * @param project
     *            The first-level resource to report to. If you want to report
     *            to "/foo/bar", this should be "foo".
     * @param workingDir
     *            A writable directory to store serialized reports that have not
     *            yet been successfully delivered. This directory should be
     *            within your project's private files and persistent across
     *            program and system restarts.
     * @param debug
     *            Whether to run in debug mode. Debug mode turns on verbose
     *            logging.
     */
    public static synchronized void init(String host, int port, boolean useSSL, String project,
                                         String workingDir, boolean debug) {
        sDebugMode = debug;

        if (host == null) { throw new CattleStateException("host cannot be null"); }
        if (project == null) { throw new CattleStateException("project cannot be null"); }
        if (workingDir == null) { throw new CattleStateException("working dir cannot be null"); }
        sDefaultHost = host;
        sDefaultPort = port;
        sDefaultUseSSL = useSSL;
        sDefaultProject = project;

        // setRoot will throw a CattleInitException if workingDir isn't usable.
        sWorkingDir = workingDir;
        filesystem.setRoot(sWorkingDir);

        sInit = true;
        Logging.debug("init():");
        Logging.debug("\thost:\t\t%s", host);
        Logging.debug("\tport:\t\t%s", port);
        Logging.debug("\tSSL:\t\t%s", useSSL ? "yep" : "nope");
        Logging.debug("\tproject:\t%s", project);
        Logging.debug("\tworkingDir:\t%s", workingDir);
    }


    public static synchronized void init(String host, int port, boolean useSSL, String project,
            String workingDir, boolean debug, Filesystem filesystemImpl) {
    	filesystem = filesystemImpl;
    	init(host, port, useSSL, project, workingDir, debug);    	
    }

    
    /**
     * Ensure all data is written to disk and stop the reliable delivery thread.
     * This method is not guaranteed to fully reset the library, so it is not
     * safe to reinitialize it and start sending data again. Do not attempt to
     * use a previously acquired instance of {@link ReportManger} after calling
     * this.
     */
    public static synchronized void shutdown() {
        Logging.debug("shutdown() called.");
        if (sInstance != null) {
            sInstance = null;
        }

        if (sReportingThread != null) {
            sReportingThread.shutdown();
            try {
                sReportingThread.join();
            } catch (InterruptedException ie) {
                // According to the docs, this happens only when someone calls
                // Thread.interrupt(). We promise not to do that.
                Logging.error("InterruptedException during sReportingThread.join()");
            }
            sReportingThread = null;
        }
        if (persistentSocket != null)
			try {
				persistentSocket.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
        
        Logging.debug("shutdown() complete.");
    }


    /**
     * This method exists for testing purposes. Instances of
     * {@link ReportManager} acquired before invoking this method may wreak
     * havoc in undefined ways if you use them after, so be <b>very</b> careful
     * if you decide to use it. <br>
     * <br>
     * Undo the effects of {@link init} and leave the static state of this class
     * the way it would be immediately after being loaded.
     */
    protected static synchronized void reset() {
        // This will set sInstance to null, as well as stopping the reporting
        // thread.
        shutdown();
        sDefaultHost = null;
        sDefaultPort = 0;
        sDefaultUseSSL = false;
        sDefaultProject = null;
        sWorkingDir = null;
        sAutokeys = makeDefaultAutokeys();
        sInit = false;
    }


    /**
     * Specify a self-signed certificate that our sockets should accept. This
     * can be called any time before a send occurs, but should be called at the
     * same time as {@link #init}.
     *
     * @param ks
     *            The new keystore. Acceptable formats vary among platforms, and
     *            particularly between J2SE and Android.
     * @throws RuntimeException
     *             The key management code in the standard library throws
     *             several exceptions that really ought to be RuntimeExceptions,
     *             but aren't. If any are thrown, this function will rethrow
     *             them as RuntimeExceptions with the same message and cause.
     */
    private static void addKeyStore(KeyStore ks) {
        Logging.debug("addKeyStore() called.");
        synchronized (sTrustedKeystores) {
            sTrustedKeystores.add(ks);
            try {
                sSocketFactory = new SelfSignedSSLSocketFactory(sTrustedKeystores);
            } catch (KeyManagementException kme) {
                throw new RuntimeException(kme.getMessage(), kme.getCause());
            } catch (NoSuchAlgorithmException nsae) {
                throw new RuntimeException(nsae.getMessage(), nsae.getCause());
            }
        }
    }


    /**
     * @return Whether reporting is globally enabled. See
     *         {@link #setReportingEnabled(boolean) setReportingEnabled()}.
     */
    public static boolean isReportingEnabled() {
        synchronized (sReportingEnabled) {
            return sReportingEnabled;
        }
    }


    /**
     * Globally set whether reporting is enabled. Calling this with
     * <code>false</code> has the following effects:
     * <ul>
     * <li>Calls to {@link #send(Report) send()} discard their arguments and do
     * nothing.</li>
     * <li>Calls to {@link #sendSync(Report) sendSync()} discard their arguments
     * and return {@link ReportResult}s that indicate successful delivery.</li>
     * <li>{@link ReportManager} remains initialized. Instances behave normally
     * aside from the above changes.</li>
     * <li>The {@link ReportingThread reliable delivery thread} continues
     * running and attempting to deliver queued reports.</li>
     * </ul>
     * This is useful in giving users a single switch to disable data collection
     * or network activity (with the caveat about the reliable delivery thread
     * above).
     */
    public static void setReportingEnabled(boolean enabled) {
        Logging.debug("setReportingEnabled(%s)", enabled);
        synchronized (sReportingEnabled) {
            sReportingEnabled = enabled;
        }
    }


    /**
     * Get the unique identifier for this client.
     */
    public static UUID getClientUUID() {
        if (!sInit) {
            throw new CattleStateException("getClientUUID called before init()!");
        }

        synchronized (sClientUUIDLock) {
            // If this is the first time we've been called, try to read our UUID
            // from disk. If that doesn't work, generate a new one and write it
            // back to disk before returning it.
            if (sClientUUID == null) {
                sClientUUID = filesystem.readClientUUID();
                if (sClientUUID == null) {
                    sClientUUID = UUID.randomUUID();
                    setClientUUID(sClientUUID);
                }
            }
            return sClientUUID;
        }
    }


    /**
     * Set the unique identifier for this client.
     */
    public static void setClientUUID(UUID newUUID) {
        Logging.debug("setClientUUID(%s)", newUUID.toString());
        if (!sInit) {
            throw new CattleStateException("setClientUUID called before init()!");
        }

        synchronized (sClientUUIDLock) {
            // UUID lacks a copy constructor.
            sClientUUID =
                new UUID(newUUID.getMostSignificantBits(), newUUID.getLeastSignificantBits());
            filesystem.writeClientUUID(sClientUUID);
        }
    }


    /**
     * @returns A map, suitable for assignment to sAutokeys, containing the
     *          built-in autokeys.
     */
    private static ConcurrentMap<String, Callable<Object>> makeDefaultAutokeys() {
        ConcurrentMap<String, Callable<Object>> autokeys =
            new ConcurrentHashMap<String, Callable<Object>>();

        autokeys.put("+client", new Callable<Object>() {
            public Object call() throws Exception {
                return getClientUUID().toString();
            }
        });

        autokeys.put("+clienttstamp", new Callable<Object>() {
            public Object call() throws Exception {
                return System.currentTimeMillis() / 1000;
            }
        });

        return autokeys;
    }


    /**
     * Get the autokeys map. Every time a report is sent, cattle will iterate
     * over the keys in this {@link Map} and add them to <b>every</b> outgoing
     * report. The value inserted for each key will be the return value of the
     * associated {@link Callable}, invoked when you call {@link #send(Report)}
     * or {@link #sendSync(Report)}. If a <code>Callable</code> does not return
     * a {@link Number} or {@link String}, the result of its {@link #toString()}
     * will end up in the final data. If a <code>Callable</code> throws an
     * {@link Exception}, that <code>Exception</code> will be used as its return
     * value. <br>
     * <br>
     * The returned map includes the built-in, automatic keys, so this can be
     * used to override them. For example, you can add a mapping for
     * "+clienttstamp" to a <code>Callable</code> that returns NTP-synchronized
     * time to improve the accuracy of those values.
     */
    public static Map<String, Callable<Object>> getAutokeys() {
        return sAutokeys;
    }


    /**
     * Set the string that will be prepended to all thread names. This is useful
     * for keeping track of all of a program's threads without having to
     * special-case Cattle's.
     */
    public static void setThreadNamePrefix(String prefix) {
        Logging.warn("setThreadNamePrefix(%s)", prefix);
        synchronized (sThreadNamePrefix) {
            sThreadNamePrefix = prefix;
        }
        updateThreadName();
    }


    /**
     * Updates the names of any threads we're maintaining.
     */
    private static void updateThreadName() {
        synchronized (sReportingThreadLock) {
            if (sReportingThread != null) {
                sReportingThread.setName(sThreadNamePrefix + Const.REPORTING_THREAD_NAME);
            }
        }
    }


    /**
     * Set the delay, in milliseconds, between iterations of the asynchronous
     * send loop.
     */
    public static void setThreadLoopDelay(long period) {
        Logging.warn("setThreadLoopDelay(%d)", period);
        synchronized (ReportingThread.sLoopDelay) {
            ReportingThread.sLoopDelay = period;
        }
    }


    /**
     * Set the disk usage limit you want the library to enforce. Note that this
     * is not a strict limit - large reports and other metadata may result in
     * slightly higher usage.
     */
    public static synchronized void setDiskUsageLimit(long bytes) {
        Logging.debug("setDiskUsageLimit(): now %,d B", bytes);
        filesystem.setDiskUsageLimit(bytes);
    }
    
    /**
     * Set the limit on the number of bytes to transmit to the server consecutively
     * before the reporting thread goes to sleep. Note that this
     * is not a strict limit - large reports and other metadata may result in
     * slightly higher usage.
     */
    protected static void setReportableBytesLimit(long bytes) {
    	Logging.debug("setReportableBytesLimit: %,d B", bytes);
        ReportingThread.setReportableBytesLimit(bytes);        
    }    


    /**
     * Retrieve an instance of {@link ReportManager}. {@link #init} must be
     * invoked before calling this.
     *
     * @return A ReportManager instance.
     * @throws CattleStateException
     *             If <code>init</code> has not been called.
     */
    public static ReportManager getInstance() {
        if (!sInit) { throw new CattleStateException(); }

        synchronized (sInstanceLock) {
            if (sInstance == null) {
                sInstance = new ReportManager();

                synchronized (sReportingThreadLock) {
                    sReportingThread = new ReportingThread(sInstance);
                    updateThreadName();
                    sReportingThread.start();
                }
            }
            return sInstance;
        }
    }


    /**
     * This class is not directly instantiable. Use {@link #getInstance}.
     */
    private ReportManager() {}


    /**
     * Add the client-side automatic keys to a Report.
     */
    private void addAutokeys(Report report) {
        JSONObject reportData = report.getJSONObject();

        if (sDebugMode) {
            reportData.put("+debug", true);
        }

        for (String key : sAutokeys.keySet()) {
            Object callableResult = null;
            try {
                callableResult = sAutokeys.get(key).call();
            } catch (Exception e) {
                callableResult = e;
            }

            if (callableResult instanceof Number) {
                reportData.put(key, callableResult);
            } else {
                reportData.put(key, callableResult.toString());
            }
        }
    }


    /**
     * Queue a report for delivery. This method returns immediately with no
     * return value; if you need to know whether a particular report worked
     * correctly, use {@link #sendSync(Report)}. <br>
     * <br>
     *
     *
     * @param report
     *            The <code>Report</code> to queue.
     */
    public void send(Report report) {
        Logging.debug("send(): %s", report);
        if (!sReportingEnabled) {
            return;
        }

        if (report.mPreviouslyAsyncSent) {
            /*
             * Best practices would have users instantiate a new Report for each
             * call to send(), but someone's going to reuse one someday. Rather
             * than have it clobber its past self or throw an Exception, we
             * change its UUID and update its createdAt, making a new Report as
             * far as the on-disk queue is concerned.
             *
             * There was one landmine here when I first wrote this. I don't
             * think there are any more. I'm pretty sure.
             */
            UUID oldUUID = report.mUUID;
            report.mUUID = UUID.randomUUID();
            report.createdAt = System.currentTimeMillis();

            String msg =
                String.format("send(): Report %s is now %s",
                              oldUUID.toString().substring(0, 6),
                              report.mUUID.toString().subSequence(0, 6));
            Logging.debug(msg);
        } else {
            report.mPreviouslyAsyncSent = true;
        }

        addAutokeys(report);
        filesystem.writeReport(report);
    }


    /**
     * Send a report and block until it completes or returns with an error.
     *
     * @param report
     *            The report to send.
     * @return A {@link ReportResult} structure.
     */
    public ReportResult sendSync(Report report) {
        if (!sReportingEnabled) {
            ReportResult fakeResult = new ReportResult();
            fakeResult.success = true;
            return fakeResult;
        }

        addAutokeys(report);
        return sendInternal(report);
    }


    /**
     * Implements the actual synchronous send behavior.
     *
     * @param report
     *            The report to send.
     * @return A {@link ReportResult} structure.
     */
    protected ReportResult sendInternal(Report report) {
        if (!sReportingEnabled) {
            ReportResult fakeResult = new ReportResult();
            fakeResult.success = true;
            return fakeResult;
        }

        // Set all uninitialized members of report to the values passed through
        // init.
        if (report.mHost == null) {
            report.mHost = sDefaultHost;
        }
        if (report.mPort == -1) {
            report.mPort = sDefaultPort;
        }
        if (report.mProject == null) {
            report.mProject = sDefaultProject;
        }
        Logging.debug("sendSync(): %s", report);

        // Validate those members.
        if (!Validate.validateHost(report.mHost)) {
            throw new CattleStateException("Invalid host: " + report.mHost);
        }
        if (!Validate.validatePort(report.mPort)) {
            throw new CattleStateException("Invalid port: " + String.valueOf(report.mPort));
        }
        if (!Validate.validateProject(report.mProject)) {
            throw new CattleStateException("Invalid project: " + report.mProject);
        }
        if (!Validate.validateDataset(report.mDataset)) {
            throw new CattleStateException("Invalid dataset: " + report.mDataset);
        }

        // We'll fill our ReportResult out as we go.
        ReportResult result = new ReportResult();
        Socket socket = null;
        Writer writer = null;
        BufferedReader reader = null;
        try {
            if (sDefaultUseSSL) {
                socket = sSocketFactory.createSocket(report.mHost, report.mPort);
            } else {
                socket = new Socket(report.mHost, report.mPort);
            }

            writer = new PrintWriter(socket.getOutputStream(), false);
            reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));

            String resource = String.format("/%s/%s", report.mProject, report.mDataset);
            String body = report.getBody();
            String post = String.format(POST_FORMAT, resource, body.length(), body);
            //Logging.debug("sendSync(): Final wire form:\n%s", post);
            writer.write(post);
            writer.flush();

            String reply = reader.readLine();
            if (reply == null) {
                if (sDefaultUseSSL) {
                    throw new IOException("no response read - maybe you need to turn off SSL?");
                } else {
                    throw new IOException("no response read - maybe you need to turn on SSL?");
                }
            }

            Logging.debug("sendSync(): Received reply: %s", reply);
            result.HTTPStatus = Integer.parseInt(reply.split("\\s")[1]);
            if (result.HTTPStatus == 200) {
                // Strip off HTTP headers and advance to the body.
                while (!(reply = reader.readLine()).equals("")) {}
                reply = reader.readLine();
                try {
                    result.response = (JSONObject) JSONValue.parse(reply);
                    result.success = !result.response.containsKey("error");
                } catch (ClassCastException cce) {
                    result.exception = cce;
                    result.success = false;
                }
            }
        } catch (UnknownHostException uhe) {
            result.exception = uhe;
        } catch (IOException ioe) {
            result.exception = ioe;
        } finally {
            try {
                if (writer != null) writer.close();
                if (reader != null) reader.close();
                if (socket != null) socket.close();
            } catch (IOException ioe) {}

            report.last_attempt = System.currentTimeMillis();
            report.attempts++;
        }

        Logging.debug("sendSync(): Returning result: %s", result);
        return result;
    }
    
    
    static TimedSocket persistentSocket = null;
    
    /**
     * Implements the actual synchronous send behavior using a persistent 
     * socket connection. The socket will disconnect after 5 seconds if unused.
     *
     * @param report
     *            The report to send.
     * @return A {@link ReportResult} structure.
     */
    protected ReportResult sendPersistentInternal(Report report) {
        if (!sReportingEnabled) {
            ReportResult fakeResult = new ReportResult();
            fakeResult.success = true;
            return fakeResult;
        }

        // Set all uninitialized members of report to the values passed through
        // init.
        if (report.mHost == null) {
            report.mHost = sDefaultHost;
        }
        if (report.mPort == -1) {
            report.mPort = sDefaultPort;
        }
        if (report.mProject == null) {
            report.mProject = sDefaultProject;
        }
        Logging.debug("sendSync(): %s", report);

        // Validate those members.
        if (!Validate.validateHost(report.mHost)) {
            throw new CattleStateException("Invalid host: " + report.mHost);
        }
        if (!Validate.validatePort(report.mPort)) {
            throw new CattleStateException("Invalid port: " + String.valueOf(report.mPort));
        }
        if (!Validate.validateProject(report.mProject)) {
            throw new CattleStateException("Invalid project: " + report.mProject);
        }
        if (!Validate.validateDataset(report.mDataset)) {
            throw new CattleStateException("Invalid dataset: " + report.mDataset);
        }
        
        // We'll fill our ReportResult out as we go.
        ReportResult result = new ReportResult();
        Socket socket = null;
        Writer writer = null;
        BufferedReader reader = null;
        try {

            if ( persistentSocket == null ) {     	
            	persistentSocket = new TimedSocket(sDefaultHost, sDefaultPort, 5*1000, 
            			sDefaultUseSSL? sSocketFactory : null);
            }
        	
        	socket = persistentSocket;
            writer = new PrintWriter(socket.getOutputStream(), false);
            reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));

            String resource = String.format("/%s/%s", report.mProject, report.mDataset);
            String body = report.getBody();
            String post = String.format(POST_PERSISTENT_FORMAT, resource, body.length(), body);
            //Logging.debug("sendSync(): Final wire form:\n%s", post);
            writer.write(post);
            writer.flush();

            String reply = reader.readLine();
            if (reply == null) {
            	
            	persistentSocket = null;
            	
                if (sDefaultUseSSL) {
                    throw new IOException("no response read - maybe you need to turn off SSL?");
                } else {
                    throw new IOException("no response read - maybe you need to turn on SSL?");
                }                
            }

            Logging.debug("sendSync(): Received reply: %s", reply);
            result.HTTPStatus = Integer.parseInt(reply.split("\\s")[1]);
            if (result.HTTPStatus == 200) {
                // Strip off HTTP headers and advance to the body.
                while (!(reply = reader.readLine()).equals("")) {}
                reply = reader.readLine();
                try {
                    result.response = (JSONObject) JSONValue.parse(reply);
                    result.success = !result.response.containsKey("error");
                } catch (ClassCastException cce) {
                    result.exception = cce;
                    result.success = false;
                }
            }
        } catch (UnknownHostException uhe) {
            result.exception = uhe;
        } catch (IOException ioe) {
            result.exception = ioe;
        } finally {
            report.last_attempt = System.currentTimeMillis();
            report.attempts++;
        }

        Logging.debug("sendSync(): Returning result: %s", result);
        return result;
    }
    

    public static void setDebugMode(boolean debug) {
    	sDebugMode = debug;
    }
    
    public static void setLogger(Logger logger) {
    	Logging.setLogger(logger);
    }
    
    public static void setLogger(final java.util.logging.Logger logger,
    		final String label) {
    	Logging.setLogger(new Logger() {

    	    public void error(String tag, String message, Object... args) {
    	    	if (sDebugMode)
    	    		logger.severe(stdLog(tag, "error", message, args));
    	    }

    	    public void warn(String tag, String message, Object... args) {
    	    	if (sDebugMode)
    	    		logger.warning(stdLog(tag, "warn", message, args));
    	    }

    	    public void debug(String tag, String message, Object... args) {
    	    	if (sDebugMode)
    	    		logger.info(stdLog(tag, "debug", message, args));    	        
    	    }

    	    private String stdLog(String tag, String level, String message, Object... args) {
    	        StringBuilder sb = new StringBuilder();
    	        sb.append(label);
    	        sb.append(":");
    	        sb.append(tag);
    	        sb.append(":");
    	        sb.append(level);
    	        sb.append(" ");
    	        sb.append(String.format(message, args));
    	        return sb.toString();    	        
    	    }
    		
    	});
    	
    	logger.setLevel(Level.INFO);
    }
}
