package edu.northwestern.cs.aqualab.cattle;

import java.util.Random;
import java.util.Set;

import edu.northwestern.cs.aqualab.cattle.filesystem.Filesystem;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.cs.aqualab.cattle.logging.Format;
import edu.northwestern.cs.aqualab.cattle.logging.Logging;



/**
 * Implements asynchronous delivery of {@link Report Reports} and exponential
 * backoff. This class is responsible for deciding which server responses
 * require a report to drop and when a report has been tried enough times to
 * give up on it. <br>
 * <br>
 * Reports are queued for delivery by writing them to the working directory
 * passed to {@link #init}. This class works with {@link Filesystem} to ensure
 * access is safe; so should you. Don't interact with the working directory
 * directly.
 */
public class ReportingThread extends Thread {
    // To introduce some jitter, we vary the base of our exponential backoff.
    private final static double BACKOFF_RADIX;
    private ReportManager mReportManager;
    private boolean mShouldRun = true;
    protected static Long sLoopDelay = Const.DEFAULT_LOOP_DELAY;
    private static long sAllowableReportBytes = 256 * 1024; // 256KB
    Filesystem filesystem;
    
    static {
        Random rnd = new Random();
        BACKOFF_RADIX = 2.0 + (0.015 * rnd.nextFloat());
    }    

    /**
     * @param reportManager
     *            The ReportManager instance this thread should serve.
     */
    protected ReportingThread(ReportManager reportManager) {
        super();
        this.setName(Const.REPORTING_THREAD_NAME);
        mReportManager = reportManager;
        filesystem = ReportManager.filesystem;
    }


    /**
     * @return Whether a report is due for a delivery attempt.
     */
    private long nextAttempt(Report report) {
        int attempts = report.attempts;
        long next_attempt = report.last_attempt;
        next_attempt += Math.pow(BACKOFF_RADIX, attempts) * 1000;
        return next_attempt;
    }


    /**
     * Deliver reports forever. Call {@link #shutdown} to interrupt this process
     * as soon as possible. It is advisable to call {@link Thread#join}
     * immediately after <code>shutdown</code>.
     */
    @Override
    public void run() {
        Logging.debug("Started.");
        while (mShouldRun) {
        	                    	
        	Set<String> queued = filesystem.readReportsNames();  
        	if (queued.size() > 0)
        		Logging.debug("Number of reports: %d.", queued.size());
            long totalSubmittedBytes = 0; 
            for ( String reportName : queued) {
            
            	Report report = filesystem.readReport(reportName);
                
            	long now = System.currentTimeMillis();
                if (now > nextAttempt(report)) {
                    Logging.debug("Report due for delivery: %s", report);
                    ReportResult result = mReportManager.sendPersistentInternal(report);

                    boolean deleteReport = false;
                    if (result.success) {
                        deleteReport = true;
                    } else {
                        if (result.HTTPStatus > 0 && result.HTTPStatus != 200) {
                            Logging.debug("Report %s received a %d",
                                    report.toString(), result.HTTPStatus);
                        }

                        JSONObject response = result.response;
                        if (response != null && response.containsKey("error")) {
                            Logging.warn("Discarding report due to server error: %s", response.get("error"));
                            deleteReport = true;
                        } else if (report.attempts == Const.MAX_REPORT_ATTEMPTS) {
                            long age = System.currentTimeMillis() - report.createdAt;
                            Logging.warn("Discarding report after %d attempts (%s): %s", report.attempts,
                                         Format.approximateFamiliarInterval(age), report);
                            deleteReport = true;
                        }
                    }
                    
                    // Keep track of the number of submitted bytes in this iteration.
                    totalSubmittedBytes += filesystem.getReportSize(reportName);                    

                    if (deleteReport) {
                        filesystem.deleteReport(report);
                    } else {
                        // Write back with updated attempt count and last
                        // attempt.
                        filesystem.writeReport(report);
                    }
                }

                // Check whether shutdown() has been called between reports.
                // Java has no breakn, but there's a check for the exact same
                // condition outside this loop that will have the same effect.
                if (!mShouldRun) break;
                    
                // If the number exceeds the maximum permitted, break.
                if (totalSubmittedBytes > sAllowableReportBytes) {
                	Logging.debug(String.format("%,d B already submitted, the limit is %,d... to to sleep...",totalSubmittedBytes, sAllowableReportBytes));
                	break;
                }
                
                // Sleep briefly to avoid long periods of activity.
                try {
                    Thread.sleep(sLoopDelay / 20);
                } catch (InterruptedException ie) {}
            }

            // Check before sleeping for a little extra responsiveness.
            if (!mShouldRun) break;

            // Check the allocated number of bytes on the disk.
            filesystem.cleanReports();
            
            try {
                Thread.sleep(sLoopDelay);
            } catch (InterruptedException i) {}
           
        }

        Logging.debug("Shutting down.");
    }
    
    /**
     * Set the limit on the number of bytes to transmit to the server consecutively
     * before the reporting thread goes to sleep. Note that this
     * is not a strict limit - large reports and other metadata may result in
     * slightly higher usage.
     */
    protected static void setReportableBytesLimit(long bytes) {
        sAllowableReportBytes = bytes;
        Logging.debug("setReportableBytesLimit: %,d B", bytes);
    }


    /**
     * Request that this thread stop running as soon as possible. It is
     * advisable to call {@link Thread#join} before assuming this thread has
     * actually exited cleanly.
     */
    public void shutdown() {
        mShouldRun = false;
        filesystem.close();
    }
}
