package edu.northwestern.cs.aqualab.cattle.example;

import java.util.Map;
import java.util.UUID;
import java.util.concurrent.Callable;
import java.util.concurrent.atomic.AtomicInteger;

import edu.northwestern.cs.aqualab.cattle.Report;
import edu.northwestern.cs.aqualab.cattle.ReportManager;
import edu.northwestern.cs.aqualab.cattle.ReportResult;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;



public class ContrivedExampleProgram {
    // You should store this stuff somewhere more central.
    private static final String HOST = "reports.aqualab.cs.northwestern.edu";
    private static final int PORT = 443;
    private static final String PROJECT = "dummy";
    private static final String DATASET = "dummy";
    private static final String WORKING_DIR = System.getProperty("java.io.tmpdir");

    /*
     * ReportManager is a singleton class. It makes no difference if you have a
     * single, shared reference, or if you call ReportManager.getInstance()
     * everywhere you want to report from.
     */
    private static ReportManager sReportManager;


    /*
     * JSON is inherently tricky to implement in statically-typed languages. In
     * the implementation we're using, JSONObject extends an unparameterized
     * Map, which means compiler warnings at every put. I'm sorry.
     *
     * For what it's worth, the first implementation I used required you to
     * catch an Exception on every .put(). This is the lesser of two evils.
     */
    public static void main(String[] args) {
        System.out.println("Setting up...");
        setUp();
        sReportManager = ReportManager.getInstance();

        /*
         * Send a blocking report. This is useful for testing whether your
         * program and the server agree on what constitutes a valid report, but
         * you probably don't want to do this in production.
         */
        Report pingReport = new Report(DATASET);
        JSONObject reportData = pingReport.getJSONObject();
        reportData.put("foo", "Contrived examples are the best!");
        ReportResult result = sReportManager.sendSync(pingReport);
        if (!result.success) {
            System.out.println("sendSync failed: " + result.toString());
            shutdown();
            return;
        }

        /*
         * Send a few async reports. Watch stderr to see them go out on the
         * wire.
         */
        for (int i = 0; i < 5; i++) {
            System.out.println("Sending report " + String.valueOf(i));
            Report ithReport = new Report(DATASET);
            ithReport.getJSONObject().put("foo", "bar");
            // Returns nothing immediately.
            sReportManager.send(ithReport);

            // Stagger the calls to send().
            long sleepFor = (i * 500) % 1501;
            try {
                Thread.sleep(sleepFor);
            } catch (InterruptedException ie) {}
        }

        // Give the reliable delivery thread a couple seconds to finish up.
        try {
            Thread.sleep(2000);
        } catch (InterruptedException ie) {}

        shutdown();
    }


    private static void setUp() {
        /*
         * The default logger goes to stderr. Implement an appropriate one for
         * your platform.
         */
//        Logging.setLogger(new Logger() {
//            @Override
//            public void warn(String tag, String message, Object... args) {}
//
//
//            @Override
//            public void error(String tag, String message, Object... args) {}
//
//
//            @Override
//            public void debug(String tag, String message, Object... args) {}
//        });


        /*
         * The autokeys functionality automatically runs Callables when send()
         * or sendSync() are invoked and puts their return values into every
         * report. Read the Javadoc for getAutokeys() before using this.
         */
        Map<String, Callable<Object>> autokeys = ReportManager.getAutokeys();
        final AtomicInteger largeValueOfThree = new AtomicInteger(3);
        autokeys.put("largeValueOfThree", new Callable<Object>() {
            public Object call() throws Exception {
                int three = largeValueOfThree.getAndIncrement();
                return "three is now " + String.valueOf(three);
            }
        });


        /*
         * Disable all reporting (with some caveats) by uncommenting this. Read
         * the Javadoc for setReportingEnabled() to understand what you're
         * getting.
         */
        // ReportManager.setReportingEnabled(false);


        /*
         * Sets common configuration for future calls to ReportManager().send()
         * and .sendSync(), and starts the reliable delivery thread.
         *
         * The final argument, debug, turns on verbose logging that is not
         * useful during normal operation. Messages at the warn and error levels
         * are always sent to the Logger.
         */
        ReportManager.init(HOST, PORT, true, PROJECT, WORKING_DIR, true);


        /*
         * cattle automatically generates a UUID for this install the first time
         * it is run. If you have one you would rather it use, pass it here.
         *
         * cattle will save the new value on disk, so you only ever need to call
         * this once.
         *
         * This call needs to know where the working directory is, so it has to
         * happen after init().
         */
        UUID nullUUID = new UUID(0, 0);
        ReportManager.setClientUUID(nullUUID);


        /*
         * cattle can automatically purge reports if disk consumption exceeds a
         * soft limit.
         */
        ReportManager.setDiskUsageLimit((long) Math.pow(2, 20));


        /*
         * The reporting thread can be decorated and influenced.
         */
        ReportManager.setThreadNamePrefix("ContrivedExample-");
        ReportManager.setThreadLoopDelay(1000);
    }


    private static void shutdown() {
        // Safely stops the reliable delivery thread.
        System.out.println("Shutting down...");
        ReportManager.shutdown();
        System.out.print("Exiting!");
    }
}
