package edu.northwestern.cs.aqualab.cattle.filesystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Set;
import java.util.UUID;

import edu.northwestern.cs.aqualab.cattle.Report;
import edu.northwestern.cs.aqualab.cattle.exception.CattleStateException;
import edu.northwestern.cs.aqualab.cattle.logging.Logging;


/**
 * Contains static methods for interacting with the filesystem. There is a
 * configurable limit on total report disk usage, which is enforced by
 * automatic, oldest-first deletion.
 */
public interface Filesystem {
	
    /**
     * Generate a filename for a {@link Report}. This filename will always be
     * the same for the same object, but may change between cattle versions.
     */
    String mkFilename(Report report);

    /**
     * @return Whether the configured root dir is set and usable.
     */
    public boolean isRootUsable();

    /**
     * Set the root directory for all future operations.
     */
    public void setRoot(String root);

    /**
     * Read all {@link Report Reports} in the given working directory and return
     * them to the caller.
     *
     * @return A Set of Reports
     */
    public Set<Report> readReports();
    
    /**
     * Read a single {@link Report Report} in the given working directory represented
     * by the reportName and return it to the caller if found.
     *
     * @return Report or null if not found.
     */
    public Report readReport(final String reportName); 

    /**
     * Write a single report to the appropriate place on disk.
     */
    public void writeReport(Report report);
    
    /**
     * Delete a report from disk. If the report is not already stored on disk,
     * this method will log a warning, but will not throw an exception.
     */
    public void deleteReport(Report report);

    /**
     * Delete all reports from disk.
     */
    public void deleteAllReports();


    /**
     * Get the persistent, unique client ID from disk.
     *
     * @return The saved client UUID, or null if it could not be read.
     */
    public UUID readClientUUID();

    /**
     * Set the persistent, unique client ID on disk.
     */
    public void writeClientUUID(UUID clientUUID);


    /**
     * Set the disk usage limit you want the library to enforce. Note that this
     * is not a strict limit - large reports and other metadata may result in
     * slightly higher usage.
     */
    public void setDiskUsageLimit(long bytes);
    
    /**
     * Perform filesystem closing operations if necessary to maintain
     * the consistency of stored reports.
	*/
    public void close();
    
    /**
     * Reduce disk usage by deleting some reports. <br>
     * <br>
     */
    public void cleanReports();

    /**
     * Read all {@link Report Reports} in the given working directory and return
     * their Names to the caller.
     *
     * @return A Set of Report Names
     */
    public Set<String> readReportsNames();
    
    /**
     * Get the size of report of name reportName.<br>
     * 
     * @return Size of report, -1 if it doesn't exist.
     */
    public long getReportSize(String reportName);

    
}
