package edu.northwestern.cs.aqualab.cattle.filesystem;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;
import java.util.UUID;

import edu.northwestern.cs.aqualab.cattle.Report;
import edu.northwestern.cs.aqualab.cattle.exception.CattleInitException;
import edu.northwestern.cs.aqualab.cattle.exception.CattleStateException;
import edu.northwestern.cs.aqualab.cattle.logging.Logging;


/**
 * Contains methods for interacting with the filesystem. There is a
 * configurable limit on total report disk usage, which is enforced by
 * automatic, oldest-first deletion.
 */
public class FilesystemImpl implements Filesystem {
    // Reports will be in files named REPORT_PREFIX + hash(body).
    private static final String REPORT_PREFIX = "report-";
    private static final String UUID_FILENAME = "uuid";
    // We maintain exclusive access to the reports directory with an OS-provided
    // advisory lock.
    // XXX: This does not work on older OSes, possibly including OS 10.6 (Snow
    // Leopard). cattle will log a message at level error and assume it is safe
    // to proceed on such platforms.
    private static final String LOCK_FILENAME = "lock";

    // Protects against concurrent access from two different processes.
    private static FileLock sExternalDiskLock = null;
    // To support Java <1.7, we need to keep track of both of these.
    private static FileOutputStream sExternalDiskLockStream = null;
    private static FileChannel sExternalDiskLockChannel = null;

    // We have a single, global lock on disk access. While coarse, I think it'll
    // be fine - we're dealing with small chunks of data, low throughput, and
    // no latency demands.
    private static Object sInternalDiskLock = new Object();
    private static File sRootDir = null;
    private static long sAllowableReportBytes = 10 * 1024 * 1024; // 10 MiB

    private static FilesystemImpl instance = null;
    
    /* Singleton patter */
    public static FilesystemImpl getInstance() {
    	
    	if (instance == null)
    		instance = new FilesystemImpl();
    	
    	return instance;
    }

    /**
     * Generate a filename for a {@link Report}. This filename will always be
     * the same for the same object, but may change between cattle versions.
     */
    public String mkFilename(Report report) {
        return REPORT_PREFIX + report.getUuid().toString();
    }


    /**
     * @return Whether the configured root dir is set and usable.
     */
    public boolean isRootUsable() {
        return sRootDir != null && sRootDir.isDirectory() && sRootDir.canRead()
               && sRootDir.canWrite();
    }


    /**
     * Set the root directory for all future operations.
     */
    public void setRoot(String root) {
        sRootDir = new File(root);
        sRootDir.mkdirs();
        if (!isRootUsable()) throw new CattleInitException("Insufficient permissions on " + root);
        if (!lockRoot()) throw new CattleInitException("Couldn't lock " + root);
    }


    /**
     * Acquire an OS-implemented advisory lock on a file relative to the root
     * dir.
     */
    protected static boolean lockRoot() {
        if (sExternalDiskLock != null && sExternalDiskLock.isValid()) {
            Logging.debug("We already have a lock for %s", sRootDir);
            return true;
        }

        File lockFile = new File(sRootDir, LOCK_FILENAME);
        lockFile.deleteOnExit();
        try {
            sExternalDiskLockStream = new FileOutputStream(lockFile);
            sExternalDiskLockChannel = sExternalDiskLockStream.getChannel();

            // tryLock() returns null if and only if it fails to acquire a lock
            // because another process has it. It throws an IOException if we're
            // on an old OS that doesn't support these locks, or if cattle has
            // seriously screwed up. In either case, we have to take our
            // chances.
            sExternalDiskLock = sExternalDiskLockChannel.tryLock();
            if (sExternalDiskLock == null) {
                unlockRoot();
                return false;
            }

            // Explicitly release the lock on shutdown.
            Runtime.getRuntime().addShutdownHook(new Thread() {
                @Override
                public void run() {
                    unlockRoot();
                }
            });
        } catch (IOException ioe) {
            ioe.printStackTrace();
            Logging.error("IOException: %s. Unsupported platform? Tread carefully.");
        }
        return true;
    }


    /**
     * Release the root dir's lock.
     */
    protected static void unlockRoot() {
        // This releases the OS lock and associated resources. The actual lock
        // file is deleted by the JVM due to a .deleteOnExit() in lockRoot().

        if (sExternalDiskLock != null) {
            try {
                sExternalDiskLock.release();
            } catch (IOException ioe) {}
            sExternalDiskLock = null;
        }

        if (sExternalDiskLockChannel != null) {
            try {
                sExternalDiskLockChannel.close();
            } catch (IOException ioe) {}
            sExternalDiskLockChannel = null;
        }

        if (sExternalDiskLockStream != null) {
            try {
                sExternalDiskLockStream.close();
            } catch (IOException ioe) {}
        }
    }
    
    /**
     * Read a single {@link Report Report} in the given working directory represented
     * by the reportName and return it to the caller if found.
     *
     * @return Report or null if not found.
     */
    public Report readReport(final String reportName) {
        synchronized (sInternalDiskLock) {
            assert isRootUsable();

            File report_files[] = sRootDir.listFiles(new FilenameFilter() {
                public boolean accept(File directory, String fileName) {
                    return fileName.equals(reportName);
                }
            });
            
            if ( report_files.length == 0) {
            	Logging.warn("readReports(): Report disappeared?! %s", reportName);
            	return null;
            }
            
            else if ( report_files.length > 1 ) {
                String msg =
                        "readReports(): "
                            + "caught ClassNotFoundException: Something is really, really wrong.";
                    Logging.error(msg);
                    throw new CattleStateException(msg);	
            }
        
            File file = report_files[0];
            
            ObjectInputStream ois = null;
            Report report = null;
            try {
                ois = new ObjectInputStream(new FileInputStream(file));
                report = (Report) ois.readObject();
            } catch (ClassNotFoundException cnfe) {
                String msg =
                    "readReports(): "
                        + "caught ClassNotFoundException: Something is really, really wrong.";
                Logging.error(msg);
                throw new CattleStateException(msg);
            } catch (FileNotFoundException fnfe) {
                Logging.warn("readReports(): Report disappeared?! %s", file.getName());
            } catch (IOException ioe) {
                Logging.warn("readReports(): IOException while reading %s: %s", file.getName(),
                             ioe.getMessage());
            } catch (ClassCastException cce) {
                Logging
                    .warn("readReports(): Unable to cast report %s to a report. Discarding.",
                          file.getName());
            } finally {
                if (ois != null) {
                    try {
                        ois.close();
                    } catch (IOException ioe) {}
                }
            }

            if (report == null) {
                file.delete();
                return null;
            }

            assert file.getName().equals(mkFilename(report));
            return report;
        }
    	
    }
    
    
    /**
     * Read all {@link Report Reports} in the given working directory and return
     * their Names to the caller.
     *
     * @return A Set of Report Names
     */
    public Set<String> readReportsNames() {
    	
    	Set<String> report_files = new HashSet<String>();
    	
        synchronized (sInternalDiskLock) {
            assert isRootUsable();

            // My kingdom for a list comprehension!
            File[] all_files = sRootDir.listFiles();            
            for (File file : all_files) {
                if (file.getName().startsWith(REPORT_PREFIX)) {
                    report_files.add(file.getName());
                }
            }
                        
        }        
        return report_files;
    }


    /**
     * Read all {@link Report Reports} in the given working directory and return
     * them to the caller.
     *
     * @return A Set of Reports
     */
    public Set<Report> readReports() {
        synchronized (sInternalDiskLock) {
            assert isRootUsable();

            // My kingdom for a list comprehension!
            File[] all_files = sRootDir.listFiles();
            Set<File> report_files = new HashSet<File>();
            for (File file : all_files) {
                if (file.getName().startsWith(REPORT_PREFIX)) {
                    report_files.add(file);
                }
            }

            Set<Report> reports = new HashSet<Report>();
            for (File file : report_files) {

                ObjectInputStream ois = null;
                Report report = null;
                try {
                    ois = new ObjectInputStream(new FileInputStream(file));
                    report = (Report) ois.readObject();
                } catch (ClassNotFoundException cnfe) {
                    String msg =
                        "readReports(): "
                            + "caught ClassNotFoundException: Something is really, really wrong.";
                    Logging.error(msg);
                    throw new CattleStateException(msg);
                } catch (FileNotFoundException fnfe) {
                    Logging.warn("readReports(): Report disappeared?! %s", file.getName());
                } catch (IOException ioe) {
                    Logging.warn("readReports(): IOException while reading %s: %s", file.getName(),
                                 ioe.getMessage());
                } catch (ClassCastException cce) {
                    Logging
                        .warn("readReports(): Unable to cast report %s to a report. Discarding.",
                              file.getName());
                } finally {
                    if (ois != null) {
                        try {
                            ois.close();
                        } catch (IOException ioe) {}
                    }
                }

                if (report == null) {
                    file.delete();
                    continue;
                }

                assert file.getName().equals(mkFilename(report));
                reports.add(report);
            }

            return reports;
        }
    }


    /**
     * Write a single report to the appropriate place on disk.
     */
    public void writeReport(Report report) {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            String report_filename = mkFilename(report);
            File report_file = new File(sRootDir, report_filename);
            ObjectOutputStream oos = null;
            try {
                oos = new ObjectOutputStream(new FileOutputStream(report_file));
                oos.writeObject(report);
            } catch (IOException ioe) {
                Logging.warn("writeReport(): Couldn't write report %s to disk. Discarding.",
                             report_filename);
            } finally {
                if (oos != null) {
                    try {
                        oos.close();
                    } catch (IOException ioe) {}
                }
            }
        }
    }


    /**
     * Delete a report from disk. If the report is not already stored on disk,
     * this method will log a warning, but will not throw an exception.
     */
    public void deleteReport(Report report) {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            String report_filename = mkFilename(report);
            File report_file = new File(sRootDir, report_filename);
            if (!report_file.exists()) {
                Logging.warn("deleteReport(): Tried to delete nonexistent %s.", report_filename);
                return;
            }
            report_file.delete();
        }
    }

    /**
     * Delete a report from disk. If the report is not already stored on disk,
     * this method will log a warning, but will not throw an exception.
     */
    public void deleteReport(String report_filename) {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            
            File report_file = new File(sRootDir, report_filename);
            if (!report_file.exists()) {
                Logging.warn("deleteReport(): Tried to delete nonexistent %s.", report_filename);
                return;
            }
            report_file.delete();
        }
    }

    /**
     * Delete all reports from disk.
     */
    public void deleteAllReports() {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            Set<Report> reports = readReports();
            for (Report report : reports) {
                deleteReport(report);
            }
        }
    }

    /**
     * Get the persistent, unique client ID from disk.
     *
     * @return The saved client UUID, or null if it could not be read.
     */
    public UUID readClientUUID() {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            File UUIDFile = new File(sRootDir, UUID_FILENAME);
            UUID clientUUID = null;
            ObjectInputStream ois = null;
            try {
                ois = new ObjectInputStream(new FileInputStream(UUIDFile));
                clientUUID = (UUID) ois.readObject();
            } catch (FileNotFoundException fnfe) {
                Logging.debug("readClientUUID(): No saved UUID found.");
            } catch (IOException ioe) {
                Logging.warn("readClientUUID(): IOError: " + ioe.getMessage());
            } catch (ClassNotFoundException cnfe) {
                String msg =
                    "readClientUUID(): ClassNotFoundException: Something is really, really wrong.";
                Logging.error(msg);
            } catch (ClassCastException cce) {
                Logging.warn("readClientUUID(): Unable to cast UUID file to a UUID. Discarding.");
                UUIDFile.delete();
            } finally {
                if (ois != null) {
                    try {
                        ois.close();
                    } catch (IOException ioe) {}
                }
            }
            return clientUUID;
        }
    }


    /**
     * Set the persistent, unique client ID on disk.
     */
    public void writeClientUUID(UUID clientUUID) {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            File UUIDFile = new File(sRootDir, UUID_FILENAME);

            ObjectOutputStream oos = null;
            try {
                oos = new ObjectOutputStream(new FileOutputStream(UUIDFile));
                oos.writeObject(clientUUID);
            } catch (IOException ioe) {
                Logging.error("writeClientUUID(): Couldn't write UUID to disk.");
            } finally {
                if (oos != null) {
                    try {
                        oos.close();
                    } catch (IOException ioe) {}
                }
            }
        }
    }


    /**
     * Set the disk usage limit you want the library to enforce. Note that this
     * is not a strict limit - large reports and other metadata may result in
     * slightly higher usage.
     */
    public void setDiskUsageLimit(long bytes) {
        sAllowableReportBytes = bytes;
        Logging.debug("setDiskUsageLimit: %,d B", bytes);
    }
    
    /**
     * Perform filesystem closing operations if necessary to maintain
     * the consistency of stored reports.
	*/
    public void close() {
    	// Nothing needed here
    }


    /**
     * Reduce disk usage by deleting some reports. <br>
     * <br>
     */
    public void cleanReports() {
    	
        assert isRootUsable();
        
        // Sort records by creation time 
        TreeSet<File> reports = new TreeSet<File>( new Comparator<File>() {
            public int compare(File one, File another) {            	
                Long oneCreatedAt = one.lastModified();
                Long anotherCreatedAt = another.lastModified();
                
            	if (oneCreatedAt.equals(anotherCreatedAt))
            		anotherCreatedAt += 1;
            		
                return oneCreatedAt.compareTo(anotherCreatedAt);
            } 
        });
        reports.addAll(readReportsFiles());
        
        long totalReportBytes = 0;
        
        // Compute record sizes
        for ( File file : reports ) {
        	totalReportBytes += file.length();
        }
        
        if ( reports.size() > 0)
        	Logging.debug("cleanReports(): Disk usage %,d by %d reports.", totalReportBytes, reports.size());

        // Remove records if the limit has been exceeded
        if (totalReportBytes > sAllowableReportBytes) {
        	
            Logging.warn("cleanReports(): Maximum disk usage %,d B exceeded %,d B; cleaning.", 
            		totalReportBytes, sAllowableReportBytes);
                        
        	// Remove enough records to trim excess bytes
            long toDeleteBytes = totalReportBytes - sAllowableReportBytes;            
            Iterator<File> iter = reports.iterator();
            Integer deleted = 0;
            while ( iter.hasNext() && toDeleteBytes > 0 ) {
            	File file = iter.next();
            	iter.remove();
				toDeleteBytes -= file.length();
				deleteReport(file.getName());
				deleted += 1;
            }
            
            Logging.warn("cleanReports(): Removed %,d records, for a total of %,d B; %d records remaining.", 
            		deleted, totalReportBytes - sAllowableReportBytes, reports.size());
            
        }
    }
    
    /**
     * Read all {@link Report Reports} in the given working directory and return
     * their Files to the caller.
     *
     * @return A Set of Report Files
     */
    public Set<File> readReportsFiles() {
    	
    	Set<File> report_files = new HashSet<File>();
    	
        synchronized (sInternalDiskLock) {
            assert isRootUsable();

            // My kingdom for a list comprehension!
            File[] all_files = sRootDir.listFiles();            
            for (File file : all_files) {
                if (file.getName().startsWith(REPORT_PREFIX)) {
                    report_files.add(file);
                }
            }
                        
        }        
        return report_files;
    }
    
    /**
     * Get the size of report of name reportName.<br>
     * 
     * @return Size of report, -1 if it doesn't exist.
     */
    public long getReportSize(final String reportName) {
    	
    	long size = -1l;
    	
        synchronized (sInternalDiskLock) {
            assert isRootUsable();

            File report_files[] = sRootDir.listFiles(new FilenameFilter() {
                public boolean accept(File directory, String fileName) {
                    return fileName.equals(reportName);
                }
            });
            
            if ( report_files.length == 0) {
            	Logging.warn("readReports(): Report disappeared?! %s", reportName);
            	return size;
            }
            
            else if ( report_files.length > 1 ) {
                String msg =
                        "readReports(): "
                            + "caught ClassNotFoundException: Something is really, really wrong.";
                    Logging.error(msg);
                    throw new CattleStateException(msg);	
            }
        
            size = report_files[0].length();
        }
    	
        return size;
    	
    }

}
