package edu.northwestern.cs.aqualab.cattle.logging;

import edu.northwestern.cs.aqualab.cattle.ReportManager;



/**
 * We have our own wrapper for logging to ease support for multiple Java-y
 * platforms. Methods here will pass their arguments to the appropriate library
 * for your JRE, Android runtime, or ToasterVM®. <br>
 * <br>
 * I find the traditional set of levels to be a few too many, so this class only
 * contains:<br>
 * <ul>
 * <li><em>error</em>: Indicates a fatal error. Usually accompanies a newly
 * thrown exception.</li>
 * <li><em>warn</em>: Indicates a recoverable error.</li>
 * <li><em>debug</em>: Indicates the program is doing things. If the library is
 * initialized with <code>debug = false</code>, messages at this level will be
 * discarded.</li>
 * </ul>
 */
public class Logging {
    private static Logger sLogger = new StderrLogger();


    /**
     * Override the default {@link Logger} implementation.
     */
    public static void setLogger(Logger logger) {
        sLogger = logger;
    }


    /**
     * Generate a tag from a current stack trace identifying the caller's
     * caller. e.g., if <code>someClass.somefunc()</code> calls
     * <code>Logging.error</code> and <code>Logging.error</code> calls this, it
     * will return a tag for <code>someClass.somefunc()</code>.
     */
    private static String makeTag() {
        StackTraceElement[] stack = null;
        try {
            stack = Thread.currentThread().getStackTrace();
        } catch (SecurityException se) {
            return "??? (SecurityException while accessing stack trace)";
        }

        if (stack.length < 4) { return "??? (stack trace was truncated)"; }

        // stack[0] == Thread.getStackTrace()
        // stack[1] == this function
        // stack[2] == Logging.{error,warn,debug}()
        // stack[3] == The callsite of interest.
        StackTraceElement grandparent = stack[3];
        return String.format("%s.%s():%d", grandparent.getClassName(), grandparent.getMethodName(),
                             grandparent.getLineNumber());
    }


    /**
     * Format and log a message at level error, indicating an unrecoverable
     * problem. This message will reach the log target regardless of the debug
     * flag.
     *
     * @param msg
     *            The message to log. Will be passed to String.format.
     * @param args
     *            The formatting arguments for <code>msg</code>.
     */
    public static void error(String msg, Object... args) {
        sLogger.error(makeTag(), msg, args);
    }


    /**
     * Format and log a message at level warn, indicating a recoverable problem.
     * This message will reach the log target regardless of the debug flag.
     *
     * @param msg
     *            The message to log. Will be passed to String.format.
     * @param args
     *            The formatting arguments for <code>msg</code>.
     */
    public static void warn(String msg, Object... args) {
        sLogger.warn(makeTag(), msg, args);
    }


    /**
     * Format and log a message at level debug. If cattle was initialized with
     * <code>debug = False</code>, this method is a no-op.
     *
     * @param msg
     *            The message to log. Will be passed to String.format.
     * @param args
     *            The formatting arguments for <code>msg</code>.
     */
    public static void debug(String msg, Object... args) {
        if (ReportManager.sDebugMode) {
            sLogger.debug(makeTag(), msg, args);
        }
    }
}
