/**
 * Dasu Project
 *
 * File:         DasuConfiguration.java
 * RCS:          $Id$
 * Description:  DasuConfiguration class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 24, 2010 at 2:23:49 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLClassLoader;
import java.security.KeyStore;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.UUID;
import java.util.logging.Logger;

import org.gudy.azureus2.core3.util.Debug;

import com.azureus.plugins.aznetmon.main.RSTPacketStats;

import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.parser.JSONParser;
import edu.northwestern.dasu.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez / David Choffnes &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The DasuConfiguration class ...
 */

@SuppressWarnings({"unchecked", "unused"})
public class DasuConfiguration {
	
	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	
	private static DasuConfiguration self = null;
		
	/** active scripts */
	public static final String DASU_CONFIG = "/dasu/ws/dasuConfig.properties"; /**/
	public static final String DASU_MESSAGES = "/dasu/ws/message-list.dat";	 /**/
	
	/** new variables **/
	public static final String DASU_FILENAME_PREFIX = "DasuUiRulesModules_";
	public static final String DASU_FILENAME_SUFFIX = ".jar";
	public static final String DASU_AME_FILENAME_PREFIX = "AMExperiment";
	public static final String DASU_AME_FILENAME_SUFFIX = ".drl";
	public static final String DASU_AME_SIGNATURE_SUFFIX = ".signature";
	
	public static final String DASU_REPORT_SERVER_HOST = "reports.aqualab.cs.northwestern.edu";
	public static final Integer DASU_REPORT_SERVER_PORT = 8088;
	public static final String DASU_REPORT_RESOURCE = "dasuv5";
	public static final String DASU_REPORT_DIRECTORY = "reports";
	
	public static final String DASU_CONFIG_SERVER_HOST = "https://config.aqualab.cs.northwestern.edu";
	public static final Integer DASU_CONFIG_SERVER_PORT = 443;
	//public static final String DASU_CONFIG_SERVER_RESOURCE = "/config";
	public static final String DASU_CONFIG_SERVER_RESOURCE = "";
	
	public static final String DASU_CONFIG_DYNAMIC = "/data/dasu/dyn_version/";
	public static final String DASU_CONFIG_PROPERTIES = "/data/dasu/config_properties";
	public static final String DASU_MESSAGE_LIST = "/data/dasu/message_list";
	
	public static final String DASU_SCRIPT_NDT = "/data/dasu/dynamic/ndt.php";
	public static final String DASU_SCRIPT_AME = "/data/dasu/dynamic/AMExperiment.php";
	public static final String DASU_SCIPT_UPDATE_CONFIG = "/data/dasu/dynamic/UpdateConfig.php";
	public static final String DASU_SCRIPT_AME_ON_DEMAND = "/data/dasu/dynamic/communicateWithTaskManager.php"; 
	
	public static final String KEYSTORE_JKS = "dasu_keyStore.jks";
	public static final String KEYSTORE_PASS = "fadazani";
	
	protected static ArrayList<String> propertiesLocations = new ArrayList<String>();
	public static String propertiesFileName = "dasu.properties";
	protected static JSONObject propsJson = null;
	private static KeyStore dasuKeyStore = null;
	
	public static final String PROBE_MODULES_PROPERTIES = "probe.modules.properties";				
	private Properties probeModulesProperties = new Properties();

	private final int maxHistoricSignalsHARDLIMIT = 40;
	private final int maxNumEntriesPerHistoricSignalHARDLIMIT = 740;
	private final int maxPingsHARDLIMIT = 40;
	private final int maxAMExperimentsHARDLIMIT = 3;
	
	private static List<String> dasuDebugFilter = new ArrayList<String>(); 
	
	private DasuConfiguration() {
		
	}
	
	public synchronized static DasuConfiguration getInstance() {
		
		if (self==null){
			self = new DasuConfiguration();
			try {
				initProperties();
				loadKeyStore();
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}
		return self;
	}
	
	static Hashtable<String, DasuProperty> dasuProperties = new Hashtable<String, DasuProperty>();
	
	public static class DasuProperty<T> {
		
		private T value;
		private Boolean overritableByServer;

		public DasuProperty(T value) {
			this(value, true);			
		}

		public DasuProperty(T value, boolean overritableByServer) {
			this.value = value;
			this.overritableByServer = overritableByServer;
		}
		
		public T getValue() {
			return value;			
		}
		
		public void setValue(T value) {
			this.value = value;			
		}
		
		public boolean isOverritableByServer() {
			return overritableByServer;
		}
		
		public String toString() {
			return "value: " + this.value + " overritableByServer: " + this.overritableByServer;  
		}
		
	}
	
	private static void initProperties() {
		
		dasuProperties.put("dasu.config.dynamic.resource", new DasuProperty<String>("release.php"));		
		
		/** Variables that can only be set locally (not overwritten by server) **/
		dasuProperties.put("dasu.enableStats", new DasuProperty<Boolean>(true, false));
		dasuProperties.put("dasu.sendToDatabase", new DasuProperty<Boolean>(true, false));
		dasuProperties.put("dasu.enablePopups", new DasuProperty<Boolean>(true, false));
		dasuProperties.put("dasu.runExperiments", new DasuProperty<Boolean>(true, false));
		
		/** the sample interval for news, in seconds */
		dasuProperties.put("dasu.sampleInterval", new DasuProperty<Integer>(30));
		dasuProperties.put("dasu.RSTimeWindowSize", new DasuProperty<Integer>(1)); /*minutes*/
		dasuProperties.put("dasu.RSTimeSpan", new DasuProperty<Integer>(5)); /*minutes*/
		
		dasuProperties.put("dasu.reportingTimeout", new DasuProperty<Integer>(60));
		dasuProperties.put("dasu.ignoreStateChangeRST", new DasuProperty<Boolean>(false));
		dasuProperties.put("dasu.stateChangeRSTInterval", new DasuProperty<Integer>(2)); /*seconds*/
		 
		dasuProperties.put("dasu.wrangler.threadLoopDelay", new DasuProperty<Long>(1000l)); /*milliseconds*/
		/* dasu.wrangler.maxAllowableReportBytes */
		dasuProperties.put("dasu.maxAllowableReportBytes", new DasuProperty<Long>(10*1024*1024l)); //10 Mib
		dasuProperties.put("dasu.maxLogEntries", new DasuProperty<Integer>(200));
		dasuProperties.put("dasu.maxHistoricSignals", new DasuProperty<Integer>(20));
		dasuProperties.put("dasu.inference.version", new DasuProperty<Float>(5.0f));	
	
		dasuProperties.put("dasu.maxNumEntriesPerHistoricSignal", new DasuProperty<Integer>(370));
		dasuProperties.put("dasu.maxPings", new DasuProperty<Integer>(10));
		dasuProperties.put("dasu.maxAMExperiments", new DasuProperty<Integer>(1));
		dasuProperties.put("dasu.maxMemJVM", new DasuProperty<String>("64M"));

		/** dasu.resource.profiler.limitUtilizationInterval **/
		dasuProperties.put("dasu.resourceProfilerLimitUtilizationInterval", new DasuProperty<Integer>(300)); /*seconds*/		
		/** dasu.resource.profiler.recordInterval **/
		dasuProperties.put("dasu.resourceProfilerRecordInterval", new DasuProperty<Integer>(1)); /*seconds*/
		/** dasu.resource.profiler.maxAllowedCpuUtilization **/
		dasuProperties.put("dasu.maxAllowedCpuUtilization", new DasuProperty<Float>(0.40f));
		
		dasuProperties.put("dasu.resource.profiler.enableReports", new DasuProperty<Boolean>(true));
		dasuProperties.put("dasu.resource.profiler.reportTimeInterval", new DasuProperty<Integer>(300)); /*seconds*/
		dasuProperties.put("dasu.resource.profiler.reportUtilizationLimit", new DasuProperty<Float>(0.05f)); 
		
		dasuProperties.put("dasu.maxBWUnitConsumption", new DasuProperty<Integer>(20)); /*units*/		
		dasuProperties.put("dasu.scheduler.maxBWUtilizationAllowedScheduler", new DasuProperty<Float>(0.8f));
		dasuProperties.put("dasu.scheduler.maxCpuUtilizationAllowedScheduler", new DasuProperty<Float>(0.2f));
		dasuProperties.put("dasu.perprobe.profiler.enabled", new DasuProperty<Boolean>(false));
		dasuProperties.put("dasu.perprobe.profiler.updateInterval", new DasuProperty<Integer>(5)); /*seconds*/
		
		/** dasu.cron.schedulerInterval **/
		dasuProperties.put("dasu.cronSchedulerInterval", new DasuProperty<Integer>(302)); /* 5*60 +2 secods */
		/** dasu.amc.sleepInterval **/
		dasuProperties.put("dasu.AMCSleepInterval", new DasuProperty<Integer>(2)); /*seconds*/
		/** dasu.manager.sleepInterval **/
		dasuProperties.put("dasu.managerSleepInterval", new DasuProperty<Integer>(30)); /*seconds*/
		/** dasu.ame.checkupInterval **/
		dasuProperties.put("dasu.AMECheckupInterval", new DasuProperty<Integer>(5*60)); /*seconds*/
		dasuProperties.put("dasu.messageManagerInterval", new DasuProperty<Integer>(60*60)); /*seconds*/
		dasuProperties.put("dasu.db.staticFlushInterval", new DasuProperty<Integer>(15*60)); /*seconds, 15 mins*/
		dasuProperties.put("dasu.db.managerSleepInterval", new DasuProperty<Integer>(2*60)); /*seconds, 2 mins*/
		
		dasuProperties.put("dasu.enableExecutionPathTimestamps", new DasuProperty<Boolean>(true));
		dasuProperties.put("dasu.enableAzureusSpeedTest", new DasuProperty<Boolean>(true));
		dasuProperties.put("dasu.stats.peerUpdate", new DasuProperty<Integer>(30)); //interval (seconds) for subsequent updates for the peerManager et al.		
		dasuProperties.put("dasu.doTraceroute", new DasuProperty<Boolean>(true));
		dasuProperties.put("dasu.doAnaLookup", new DasuProperty<Boolean>(true));
		dasuProperties.put("dasu.stats.tracerouteTimeout", new DasuProperty<Integer>(60)); /*seconds*/
		dasuProperties.put("dasu.stats.dnsTimeout", new DasuProperty<Integer>(1)); /*seconds*/
		dasuProperties.put("dasu.update.updateCheck", new DasuProperty<Integer>(600)); /*seconds*/
		
	}
	
	public String getResourceScriptRequest() {
		
		StringBuilder resource = new StringBuilder();
		
		//basic information
		resource.append("?ip=" + Main.getPublicIpAddress() +
			"&uuid=" + DasuConfiguration.getInstance().getUUID());
		
		//geographic location
		if ((NetworkAdminASNLookupImpl.getInstance() != null) 
				&& (NetworkAdminASNLookupImpl.getInstance().getASNInfo() != null)) {

			resource.append("&asn=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS() +
				"&cc=" + NetworkAdminASNLookupImpl.getInstance().getASNInfo().getCountry());
		}

		//rules and client version
		resource.append( "&clientversion=" + Main.getPluginInterface().getPluginVersion() +
				"&interfversion=" + getInferenceVersion() );
		
		resource.append("&osname=" + System.getProperty("os.name").replace(" ", "_").replace("&", "_") +
				"&osversion=" + System.getProperty("os.version").replace(" ", "_").replace("&", "_") +
				"&osarch=" + System.getProperty("os.arch").replace(" ", "_").replace("&", "_"));
				
		return resource.toString();
		
	}
	
	public boolean isResourceProfilerReportsEnabled() {
		return (Boolean) dasuProperties.get("dasu.resource.profiler.enableReports").getValue();
	}
	
	public Integer getResourceProfilerReportTimeInterval() {
		return (Integer) dasuProperties.get("dasu.resource.profiler.reportTimeInterval").getValue();
	}
	
	public Float getResourceProfilerReportUtilizationLimit() {
		return (Float) dasuProperties.get("dasu.resource.profiler.reportUtilizationLimit").getValue();
	}
	
	public List<String> getDasuDebugFilter() {
		return dasuDebugFilter;
	}
	
	public void setDasuDebugFilter(String[] filter) {
		dasuDebugFilter = Arrays.asList(filter);
	}
	
	public KeyStore getKeyStore() {
		return dasuKeyStore;
	}

	public Long getWranglerSetLoopDelay() {
		return (Long) dasuProperties.get("dasu.wrangler.threadLoopDelay").getValue();
	}
	
	public boolean isPerProbeProfilerEnabled() {
		return (Boolean) dasuProperties.get("dasu.perprobe.profiler.enabled").getValue();
	}
	
	public int getPerProbeProfilerUpdateInterval() {
		return  (Integer) dasuProperties.get("dasu.perprobe.profiler.updateInterval").getValue();
	}
	
	public int getTracerouteTimeout() {
		return (Integer) dasuProperties.get("dasu.stats.tracerouteTimeout").getValue();
	}

	public int getDnsTimeout() {
		return (Integer) dasuProperties.get("dasu.stats.dnsTimeout").getValue();
	}

	private void setDnsTimeout(int dnsTimeout) {
		dasuProperties.get("dasu.stats.dnsTimeout").setValue(dnsTimeout);
	}

	public int getUpdateCheckFreqSec() {
		return (Integer) dasuProperties.get("dasu.update.updateCheck").getValue();
	}

	private void setUpdateCheckFreqSec(int updateCheckFreqSec) {
		dasuProperties.get("dasu.update.updateCheck").setValue(updateCheckFreqSec);
	}

	public boolean isDoTraceroute() {
		return (Boolean) dasuProperties.get("dasu.doTraceroute").getValue();		
	}

	public boolean isDoAnaLookup() {
		return (Boolean) dasuProperties.get("dasu.doAnaLookup").getValue();
	}
	
	private void setDoTraceroute(boolean doTraceroute) {
		dasuProperties.get("dasu.doTraceroute").setValue(doTraceroute);
	}

	private void setDoAnaLookup(boolean doAnaLookup) {
		dasuProperties.get("dasu.doAnaLookup").setValue(doAnaLookup);
	}
	
	public int getPeerUpdateIntervalSec() {
		return (Integer) dasuProperties.get("dasu.stats.peerUpdate").getValue();
	}
	
	private void setPeerUpdateIntervalSec(int peerUpdateIntervalSec) {
		dasuProperties.get("dasu.stats.peerUpdate").setValue(peerUpdateIntervalSec);
	}

	public boolean isRecordStats() {
		return (Boolean) dasuProperties.get("dasu.enableStats").getValue();		
	}

	private void setRecordStats(boolean recordStats) {
		dasuProperties.get("dasu.enableStats").setValue(recordStats);
	}
	
	public boolean isEnablePopups() {
		return (Boolean) dasuProperties.get("dasu.enablePopups").getValue();
	}
	
	public void setEnablePopups(boolean enablePopups) {
		dasuProperties.get("dasu.enablePopups").setValue(enablePopups);
	}

	public boolean isEnableExecutionPathTimestamps() {
		return (Boolean) dasuProperties.get("dasu.enableExecutionPathTimestamps").getValue();
	}

	private void setEnableExecutionPathTimestamps(boolean enableExecutionPathTimestamps) {
		dasuProperties.get("dasu.enableExecutionPathTimestamps").setValue(enableExecutionPathTimestamps);
	}

	public boolean isEnableAzureusSpeedTest() {
		return (Boolean) dasuProperties.get("dasu.enableAzureusSpeedTest").getValue();
	}

	private void setEnableAzureusSpeedTest(boolean enableAzureusSpeedTest) {
		dasuProperties.get("dasu.enableAzureusSpeedTest").setValue(enableAzureusSpeedTest);
	}

	public boolean isSendToDatabase() {
		return (Boolean) dasuProperties.get("dasu.sendToDatabase").getValue();
	}

	public void setSendToDatabase(boolean sendToDatabase) {
		dasuProperties.get("dasu.sendToDatabase").setValue(sendToDatabase);
	}

	public void setProbeModulesProperties(Properties prop) {
		probeModulesProperties = prop;
	}
	
	private static void loadKeyStore() throws Exception {
		
        dasuKeyStore = KeyStore.getInstance("JKS");
        InputStream is = null;
        
        if ( Main.getPluginInterface() != null)
        	is = Main.getPluginInterface().getPluginClassLoader().getResourceAsStream(KEYSTORE_JKS);
        else
        	is = ClassLoader.getSystemResourceAsStream(KEYSTORE_JKS);
        
        if ( is == null)
        	throw new Exception(String.format("keystore %s could not be found..", KEYSTORE_JKS));
        
        dasuKeyStore.load(is, KEYSTORE_PASS.toCharArray());

	}
	
	private void setReportingTimeout(int timeout) {
		dasuProperties.get("dasu.reportingTimeout").setValue(timeout);
	}

	public int getReportingTimeout() {
		return (Integer) dasuProperties.get("dasu.reportingTimeout").getValue();
	}

	private void setMaxPings(int value) {
		
		if (value <= maxPingsHARDLIMIT)
			dasuProperties.get("dasu.maxPings").setValue(value);
		
	}
	
	public int getMaxPings(){
		return (Integer) dasuProperties.get("dasu.maxPings").getValue();
	}

	private void setMaxAMExperiments(int value) {
		
		if (value <= maxAMExperimentsHARDLIMIT)
			dasuProperties.get("dasu.maxAMExperiments").setValue(value);
		
	}
	
	public int getMaxAMExperiments(){
		return (Integer) dasuProperties.get("dasu.maxAMExperiments").getValue();
	}
	
	private void setMaxMemJVM(String value) {
		dasuProperties.get("dasu.maxMemJVM").setValue(value);
	}
	
	public String getMaxMemJVM(){
		return (String) dasuProperties.get("dasu.maxMemJVM").getValue();		
	}

	public void setInferenceVersion(float version) {
		dasuProperties.get("dasu.inference.version").setValue(version);
	}

	public float getInferenceVersion(){
		return (Float) dasuProperties.get("dasu.inference.version").getValue();
	}

	public String getUUID() {
		return (String) dasuProperties.get("dasu.uuid").getValue();
	}
	
	private void setUUID(String uuid){
		dasuProperties.get("dasu.uuid").setValue(uuid);
	}

	public boolean isRunExperiments() {
		return (Boolean) dasuProperties.get("dasu.runExperiments").getValue();
	}
	
	public void setRunExperiments(boolean runExperiments){
		dasuProperties.get("dasu.runExperiments").setValue(runExperiments);
	}

	public boolean isIgnoreStateChangeRST() {
		return (Boolean) dasuProperties.get("dasu.ignoreStateChangeRST").getValue();
	}
	
	public int getSampleInterval() {		
		return (Integer) dasuProperties.get("dasu.sampleInterval").getValue();
	}

	private void setSampleInterval(int sampleInterval) {
		dasuProperties.get("dasu.sampleInterval").setValue(sampleInterval);
	}

	public int getCronSchedulerInterval() {
		return (Integer) dasuProperties.get("dasu.cronSchedulerInterval").getValue();	
	}

	private void setCronSchedulerInterval(int interval) {
		dasuProperties.get("dasu.cronSchedulerInterval").setValue(interval);
	}

	public int getResourceProfilerRecordInterval() {
		return (Integer) dasuProperties.get("dasu.resourceProfilerRecordInterval").getValue();
	}

	private void setResourceProfilerRecordInterval(int interval){
		dasuProperties.get("dasu.resourceProfilerRecordInterval").setValue(interval);
	}

	public int getResourceProfilerLimitUtilizationInterval() {
		return (Integer) dasuProperties.get("dasu.resourceProfilerLimitUtilizationInterval").getValue();
	}

	private void setResourceProfilerLimitUtilizationInterval(int interval){
		dasuProperties.get("dasu.resourceProfilerLimitUtilizationInterval").getValue();
	}

	public String getDasuConfigDynamicResource() {
		return (String) dasuProperties.get("dasu.config.dynamic.resource").getValue();
	}
	
	private void setDasuConfigDynamicResource(String resource) {
		dasuProperties.get("dasu.config.dynamic.resource").setValue(resource);
	}
	
	public int getAMECheckupInterval() {
		return (Integer) dasuProperties.get("dasu.AMECheckupInterval").getValue();
	}
	
	private void setAMECheckupInterval(int interval){
		dasuProperties.get("dasu.AMECheckupInterval").setValue(interval);
	}
	
	public int getMessageManagerInterval() {
		return (Integer) dasuProperties.get("dasu.messageManagerInterval").getValue();
	}
	
	private void setMessageManagerInterval(int interval){
		dasuProperties.get("dasu.messageManagerInterval").setValue(interval);
	}

	public int getStateChangeRSTInterval() {
		return (Integer) dasuProperties.get("dasu.stateChangeRSTInterval").getValue();
	}

	private void setStateChangeRSTInterval(int interval) {
		dasuProperties.get("dasu.stateChangeRSTInterval").setValue(interval);
	}


	public int getAMCSleepInterval() {
		return (Integer) dasuProperties.get("dasu.AMCSleepInterval").getValue();
	}

	private void setAMCSleepInterval(int interval){
		dasuProperties.get("dasu.AMCSleepInterval").setValue(interval);
	}
	
	public int getDBManagerSleepInterval() {
		return (Integer) dasuProperties.get("dasu.db.managerSleepInterval").getValue();
	}

	private void setDBManagerSleepInterval(int interval){
		dasuProperties.get("dasu.db.managerSleepInterval").setValue(interval);
	}
	
	public int getDbStaticFlushInterval() {
		return (Integer) dasuProperties.get("dasu.db.staticFlushInterval").getValue();
	}

	private void setDbStaticFlushInterval(int interval){
		dasuProperties.get("dasu.db.staticFlushInterval").setValue(interval);
	}
	
	public int getDasuManagerSleepInterval() {
		return (Integer) dasuProperties.get("dasu.managerSleepInterval").getValue();
	}

	private void setDasuManagerSleepInterval(int interval){
		dasuProperties.get("dasu.managerSleepInterval").setValue(interval);
	}

	public int getMaxLogEntries() {
		return (Integer) dasuProperties.get("dasu.maxLogEntries").getValue();
	}
	
	private void setMaxLogEntries(int entries){
		dasuProperties.get("dasu.maxLogEntries").setValue(entries);
	}

	public long getMaxAllowableReportBytes() {
		return (Long) dasuProperties.get("dasu.maxAllowableReportBytes").getValue();
	}
	
	private void setMaxAllowableReportBytes(long bytes){
		dasuProperties.get("dasu.maxAllowableReportBytes").setValue(bytes);
	}

	public int getMaxBWUnitConsumptionScheduler() {
		return (Integer) dasuProperties.get("dasu.maxBWUnitConsumption").getValue();
	}
	
	private void setMaxBWUnitConsumptionScheduler(int size){
		dasuProperties.get("dasu.maxBWUnitConsumption").setValue(size);
	}

	public float getMaxBWUtilizationAllowedScheduler() {
		return (Float) dasuProperties.get("dasu.scheduler.maxBWUtilizationAllowedScheduler").getValue();
	}
	
	private void setMaxBWUtilizationAllowedScheduler(float max) {
		dasuProperties.get("dasu.scheduler.maxBWUtilizationAllowedScheduler").setValue(max);
	}

	public float getMaxCpuUtilizationAllowedScheduler() {
		return (Float) dasuProperties.get("dasu.scheduler.maxCpuUtilizationAllowedScheduler").getValue();
	}
	
	private void setMaxCpuUtilizationAllowedScheduler(float max){
		dasuProperties.get("dasu.scheduler.maxCpuUtilizationAllowedScheduler").setValue(max);
	}

	public float getMaxAllowedCpuUtilizationProfiler() {
		return (Float) dasuProperties.get("dasu.maxAllowedCpuUtilization").getValue();		
	}
	
	private void setMaxAllowedCpuUtilizationProfiler(float maxAllowedCpuUtilization){
		dasuProperties.get("dasu.maxAllowedCpuUtilization").setValue(maxAllowedCpuUtilization);
	}

	public int getMaxHistoricSignals() {
		return (Integer) dasuProperties.get("dasu.maxHistoricSignals").getValue();
	}
	
	private void setMaxHistoricSignals(int entries){
		
		if (entries <= maxHistoricSignalsHARDLIMIT) {
			dasuProperties.get("dasu.maxHistoricSignals").setValue(entries);
			
		}
	}

	public int getMaxNumEntriesPerHistoricSignal() {		
		return (Integer) dasuProperties.get("dasu.maxNumEntriesPerHistoricSignal").getValue();
	}
	
	private void setMaxNumEntriesPerHistoricSignal(int entries){

		if (entries <= maxNumEntriesPerHistoricSignalHARDLIMIT)
			dasuProperties.get("dasu.maxNumEntriesPerHistoricSignal").setValue(entries);

	}

	private void setIgnoreStateChangeRST(boolean ignoreStateChangeRST) {
		dasuProperties.get("dasu.ignoreStateChangeRST").setValue(ignoreStateChangeRST);
	}

	public int getRSTimeWindowSize() {
		return (Integer) dasuProperties.get("dasu.RSTimeWindowSize").getValue();
	}

	public int getRSTTimeSpan() {
		return (Integer) dasuProperties.get("dasu.RSTimeSpan").getValue();
	}

	private void setRSTimeWindowSize(int timeWindowSize) {
		dasuProperties.get("dasu.RSTimeWindowSize").setValue(timeWindowSize);
	}

	private void setRSTTimeSpan(int timeSpan) {
		dasuProperties.get("dasu.RSTimeSpan").setValue(timeSpan);
	}

	private void setTracerouteTimeout(Integer val) {
		dasuProperties.get("dasu.stats.tracerouteTimeout").setValue(val);
	}
	
	public Properties getProbeModulesProperties() {
		return probeModulesProperties;
	}

	public void updatePropertiesJson(JSONObject json ) {		
		updatePropertiesJson(json, false);		
	}
	
	private void updatePropertiesJson(JSONObject json, boolean override) {

		LOGGER.info("CONFIG: Updating properties...");
		try {
					
			Iterator<String> iter = json.keySet().iterator();
			while ( iter.hasNext() ) {
				String key = iter.next().toString();
				Object value = json.get(key);
	
				try {
										
					if (key.equals("dasu.uuid")){
						//only override the uuid if it is null in the first place
						if ( !dasuProperties.containsKey("dasu.uuid")) {
							dasuProperties.put("dasu.uuid", new DasuProperty<String>(value.toString()));
						}
							
					} 
					else if ( key.equals("dasu.debug.filter") ) {
						String[] args = value.toString().split(",");
						Main.setLogFilter(args);
						continue;
					}
					else if ( !dasuProperties.containsKey(key)) {
						LOGGER.warning("Unknown key: *"+key+"* removing...");
						iter.remove();
					}
					
					if (dasuProperties.get(key).isOverritableByServer() || override) {
												
						if ( dasuProperties.get(key).getValue() instanceof Integer )
							dasuProperties.get(key).setValue(Integer.parseInt(value.toString()));
						else if ( dasuProperties.get(key).getValue() instanceof Float )
							dasuProperties.get(key).setValue(Float.parseFloat(value.toString()));
						else if ( dasuProperties.get(key).getValue() instanceof Boolean )
							dasuProperties.get(key).setValue(Boolean.parseBoolean(value.toString()));
						else if ( dasuProperties.get(key).getValue() instanceof Long )
							dasuProperties.get(key).setValue(Long.parseLong(value.toString()));						
						else
							dasuProperties.get(key).setValue(value.toString());
						
						if (key.equals("dasu.sampleInterval")){
							int sampleInterval = Integer.valueOf(value.toString());
							RSTPacketStats.INTERVAL = sampleInterval*1000;
						}

					}
					
				} catch (Exception e){
					System.err.println("Error processing property... " + key);
					e.printStackTrace();
				} catch (Error e){
					System.err.println("Error processing property... " + key);
					e.printStackTrace();
				}
				
			}
			
			// if no uuid, generate it now
			//if (uuid==null){
			//	uuid = UUID.randomUUID().toString();
			//	json.put("dasu.uuid", uuid);
			//	//write = true;				
			//}
			if ( !dasuProperties.containsKey("dasu.uuid") ) {
				LOGGER.warning("generating new uuid");
				String uuid = UUID.randomUUID().toString();
				dasuProperties.put("dasu.uuid", new DasuProperty<String>(uuid.toString()));
				json.put("dasu.uuid", uuid);
			}
			
			writePropertiesJson(json);
			
		}
		catch (Exception e) {
			e.printStackTrace();			
			LOGGER.severe("CONFIG: empty properties file" );
		}
		
		LOGGER.info("UUID " + dasuProperties.get("dasu.uuid"));
		
	}
		
	private void writePropertiesJson(JSONObject props) {
		
		for (String prop : dasuProperties.keySet()) {
			props.put(prop, dasuProperties.get(prop).getValue());
		}
		
		for (String filename : propertiesLocations){
			try {
				
				File f = new File(filename);
				FileWriter fow = new FileWriter(f);
				fow.write(props.toJSONString());
				//props.write(fow);
				fow.close();
				break;
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
	}
	
	private boolean loadPropertiesJsonFromLocation() {
		
		boolean found = false;
		try {
			for (String location : propertiesLocations) {
				File properties_file = new File(location);
				// if properties file exists on its own then override any
				// properties file potentially held within a jar
				if (properties_file.exists()) {
					FileInputStream fis = null;

					try {
						fis = new FileInputStream(properties_file);						
				    	BufferedReader bufferedreader = new BufferedReader(
				    			new InputStreamReader(fis));
						String string;
				    	StringBuilder sb = new StringBuilder();
				    	while ((string = bufferedreader.readLine()) != null) {	    	    
				    	    sb.append(string);
				    	}
				    	
					    propsJson = (JSONObject)new JSONParser().parse(sb.toString());
						found = true;
						LOGGER.info("CONFIG: Succesfully loaded file " + location );
						break;
					}
					catch (Exception e) {
						
						String msg = "Can't read '" + location + "' for plugin '"
								+ Main.getPluginInterface().getPluginName() + "': file may be missing";
				
						LOGGER.warning("CONFIG: " + msg);
						LOGGER.warning("CONFIG: " + e.getMessage());
						
					} finally {
						if (fis != null) {
							fis.close();
						}
					}
				}
			}
			
			if (!found) {
				// properties file wasn't there, so use the jar
				ClassLoader classLoader = Main.getPluginInterface().getPluginClassLoader();

				if (classLoader instanceof URLClassLoader) {
					URLClassLoader current = (URLClassLoader) classLoader;

					URL url = current.findResource("dasu.properties.default");

					if (url != null) {
						
				    	BufferedReader bufferedreader = new BufferedReader(
				    			new InputStreamReader(url.openStream()));
						String string;
				    	StringBuilder sb = new StringBuilder();
				    	while ((string = bufferedreader.readLine()) != null) {	    	    
				    	    sb.append(string);
				    	}
				    	propsJson = (JSONObject)new JSONParser().parse(sb.toString());
				    	LOGGER.info("CONFIG: Succesfully loaded file dasu.properties.default from jars" );
				    	found = true;
				    	
					} else {
						throw (new Exception(
								"failed to load dasu.properties.default from jars"));
					}
				}
			}
			
		
		} catch (Throwable e) {
			Debug.printStackTrace(e);
	
			String msg = "Can't read 'dasu.properties.default' for plugin '"
					+ Main.getPluginInterface().getPluginName() + "': file may be missing";
	
			LOGGER.warning("CONFIG: " + msg);
		}
		
		return found;
		
	}

	protected void loadPropertiesJson() {
		updatePropertiesJson(propsJson, true);
	}
	
	public void updatePropertiesJson() {
		writePropertiesJson(propsJson);
	}

	public void checkPropertiesPreviousVersionDasu(String domain) {
			
		try {
			
			String propsFileName = Main.getDir() + File.separator  + propertiesFileName.concat("." + domain);
			File f = new File(propsFileName);
			if ( f.exists() ) {
				Properties p = new Properties();
				FileInputStream fis = new FileInputStream(f);
				p.load(fis);
				//System.out.println(p);
				
				String uuid = null;
				Integer UUIDid = -1;
				
				if ( p.containsKey("dasu.uuid") ) {
					uuid = p.getProperty("dasu.uuid");
					DasuConfiguration.getInstance().setUUID(uuid);
				}
				
				ReportableFact rf = new ReportableFact("dasuUpdated");
				rf.addContentsObject("timestamp", Util.currentDasuTime());
				rf.addContentsObject("uuid", uuid);
				rf.addContentsObject("UUIDid", UUIDid);
				Statistics.getInstance().addReportableFact(rf);
				
				f.delete();				
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	
	public boolean initializeProperties(String domain) {
		
		checkPropertiesPreviousVersionDasu(domain);
		
		String asn = NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS();
		if ( !Util.isNumeric(asn) ) 
			asn = "0";
			
		propertiesFileName = propertiesFileName.concat(".AS" + asn);		
		propertiesLocations.add(Main.getDir()
				+ java.io.File.separator + propertiesFileName);
		LOGGER.info("CONFIG: loading properties from " + propertiesFileName);		
		
		/* load properties */
		if (loadPropertiesJsonFromLocation()) {
			DasuConfiguration.getInstance().loadPropertiesJson();
			return true;
		}
		
		return false;
		
	}
	

}
