/**
 * Dasu Project
 *
 * File:         DasuManager.java
 * RCS:          $Id$
 * Description:  DasuManager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 24, 2010 at 2:27:54 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu;

import java.io.File; 
import java.security.KeyStore;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.concurrent.TimeUnit;

import org.gudy.azureus2.core3.config.COConfigurationManager;
import org.gudy.azureus2.core3.config.ParameterListener;
import org.gudy.azureus2.core3.logging.Logger;

import edu.northwestern.dasu.DasuData.DasuDataListener;
import edu.northwestern.dasu.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.dasu.drools.FactDidEvent;
import edu.northwestern.dasu.drools.FactMeasurementLpsd;
import edu.northwestern.dasu.drools.FactMeasurementSystem;
import edu.northwestern.dasu.drools.FactMeasurementUPnP;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.drools.RuleEngine;
import edu.northwestern.dasu.measurement.*;
import edu.northwestern.dasu.measurement.ame.AMEManager;
import edu.northwestern.dasu.measurement.monitors.DownloadMonitor;
import edu.northwestern.dasu.measurement.monitors.ResetMonitor;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.netstat.NetstatBanwidthPerformer;
import edu.northwestern.dasu.profilers.PerProbeProfiler;
import edu.northwestern.dasu.profilers.Profiler;
import edu.northwestern.dasu.profilers.ResourceProfiler;
import edu.northwestern.dasu.simulator.InterferenceSimulatorV3;
import edu.northwestern.dasu.stats.*;
import edu.northwestern.dasu.timer.DasuCronScheduler;
import edu.northwestern.dasu.timer.DasuMessageManager;
import edu.northwestern.dasu.timer.DasuTimerManager;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.timer.TimerEntry;
import edu.northwestern.dasu.upnp.UPnPBanwidthPerformer;
import edu.northwestern.dasu.util.DasuClassLoader;
import edu.northwestern.dasu.util.HttpsChannel;
import edu.northwestern.dasu.util.PluginInterface;
import edu.northwestern.dasu.util.UpdaterV2;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 * 
 *         The DasuManager class ...
 */
public class DasuManager implements Runnable {
	
	public final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger("edu.northwestern.dasu");
	
	private static boolean isActive = false;
	public Object halt = new Object();
	public static boolean isInitialized = false;
	private int sleepInterval = DasuConfiguration.getInstance().getDasuManagerSleepInterval();
	public static boolean USE_SIMULATION_V2 = false;
	private static DasuManager self;
	
	/** Interference Simulator **/
	public static InterferenceSimulatorV3 interferenceSimulator = null;
	
	/** Monitor for added/removed downloads */
	public DownloadMonitor dlm;
	public NetstatBanwidthPerformer nbp;
	public UPnPBanwidthPerformer ubp;
	public ResetMonitor rm;
	
	/* DROOLS ENGINES */
	public static RuleEngine engine = null;
	private static Hashtable<String, ParameterListener> COConfigurationListeners = new Hashtable<String, ParameterListener>();
	
	private static Profiler profiler = null;
	private static PerProbeProfiler perProbeProfiler = null;
	
	/* FACTS */
	public static FactMeasurementSystem systemFact = null;
	public static FactMeasurementLpsd localPeerFact = null;
	public static FactMeasurementUPnP localNetworkFact = null;
	
	public static PersistentData persistentDataManager = new PersistentData();
	protected static PluginInterface pi = Main.getPluginInterface();
	
	//private static String persistentDataFileName = Main.getDir() + File.separatorChar + "dasu.data." + Main.getDomainName();
	private static String persistentDataFileName = Main.getDir() + File.separatorChar + "dasu.data.AS" + 
			NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS();

	/* Pertaining to updating the UI */
	public static UpdaterV2 updater = null;
	
	/* Pertaining to the simulator */
	public static long startUpTime = Util.currentTimeMillis();
	
	/* Pertaining the Profiler*/
	protected volatile int yieldInterval = -1; //milliseconds
	
	private static DasuClassLoader dasuClassloader = null;

	/* Configuration service*/
	public static HttpsChannel configService = null;
	
	public static AMEManager ameManager = null;
	
	private static ResourceProfiler resourceProfiler = null;
		
	//ITimer AMETJVMStartTimer = null;
	//ITimer PersistentDataDumpTimer = null;
	//ITimer tickTimer;
	
	public void initialize() {
		
		// Initialize
		initializeConfigService();	
		
		// Start updating the UI and rules
		updater = new UpdaterV2(LOGGER);
		
		if (!Main.SKIP_UPDATE_UI_RULES) {			
			updater.run(); // just inline this :) don't run as a thread.
		}
		
		if(!updater.isReady() || !updater.haveValidFile()) {
			LOGGER.severe("Updater failed to find a working set of rules/UI.");
		}
		
		dasuClassloader = new DasuClassLoader(this.getClass().getClassLoader(), 
				updater.getJarPath(DasuConfiguration.getInstance().getInferenceVersion()) );
		dasuClassloader.loadExternalClasses();
		
		LOGGER.info("DEBUG: loading persistent data");
		Object pd = Main.loadObject(persistentDataFileName);
		if (pd != null)  {
			persistentDataManager = (PersistentData) pd;
		}
	}
	
	private DasuManager() {
		
		self = this;
		isActive = true;

		initialize();
		StatManager.registerUsage();
		StatManager.start();
		
	}

	public static DasuManager getInstance() {
		
		if ( self == null)
			self = new DasuManager();
		
		return self;
	}
	
	private static void initializeConfigService() {
        
		KeyStore ks = DasuConfiguration.getInstance().getKeyStore();
        HttpsChannel.addKeyStore(ks);
        configService = new HttpsChannel(DasuConfiguration.DASU_CONFIG_SERVER_HOST,
        		DasuConfiguration.DASU_CONFIG_SERVER_PORT, 
        		DasuConfiguration.DASU_CONFIG_SERVER_RESOURCE);
		
	}
	
	public HttpsChannel getConfigService() {
		return configService;
	}
	
	public static ClassLoader getDasuClassLoader() {		
		return dasuClassloader;
	}
	
	public static PersistentData getPersistentDataManager() {
		return persistentDataManager;
	}
	
	public static ResourceProfiler getResourceProfiler() {
		return resourceProfiler;
	}

	private void launchConfigurationChangeListener() {

		final HashMap<String, String> paramsEvent = new HashMap<String, String>();
		paramsEvent.put("Max Download Speed KBs", "MaxDownloadChanged");
		paramsEvent.put("Max Upload Speed KBs", "MaxUploadChanged");
		paramsEvent.put("Max.Peer.Connections.Total", "MaxConnectionsChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent", "MaxConnectionsPerTorrentChanged");
		paramsEvent.put("Max Seeds Per Torrent", "MaxSeedsPerTorrentChanged");
		paramsEvent.put("Use Lazy Bitfield", "LazyBitfieldChanged");
		paramsEvent.put("network.transport.encrypted.require", "EncryptionRequiredChanged");
		paramsEvent.put("network.transport.encrypted.min_level", "EncrytionLevelChanged");
		paramsEvent.put("network.transport.encrypted.fallback.outgoing", "EncryptionOutgoingChanged");
		paramsEvent.put("network.transport.encrypted.fallback.incoming", "EncryptionIncominChanged");
		paramsEvent.put("network.transport.encrypted.use.crypto.port", "EncryptionCryptoPortChanged");
		paramsEvent.put("TCP.Listen.Port", "TCPPortChanged");
		paramsEvent.put("UDP.Listen.Port", "UDPPortChanged");
		paramsEvent.put("Max.Peer.Connections.Per.Torrent.When.Seeding.Enable", "MaxConnectionsPerTorrentWhenSeedingChanged");
		paramsEvent.put("enable.seedingonly.maxuploads", "MaxUploadWhenSeedingChanged");
		
		for (String param : paramsEvent.keySet()) {
			
			ParameterListener pl = new ParameterListener() {
				public void parameterChanged(String parameterName) {					
					AMCoordinatorV2.assertToEngine(SourceType.RE, new FactDidEvent(paramsEvent.get(parameterName), -1L));
				}
			};
			
			COConfigurationManager.addParameterListener(param, pl);
			COConfigurationListeners.put(param, pl);
			
		}		

	}
	
	private void printDebug() {
	
		if ((localPeerFact != null) && (systemFact != null)) {

			StringBuilder sb = new StringBuilder();
			
			sb.append("local lpsd: avgUploadRate: "
					+ (localPeerFact.getBytesSent()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (localPeerFact.getBytesReceived()/(localPeerFact.getShortDeltaTime()!=0?localPeerFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (localPeerFact.getShortDeltaTime()) + "\n");
			
			sb.append("local netstat: avgUploadRate: "
					+ (systemFact.getBytesSent()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (systemFact.getBytesReceived()/(systemFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (systemFact.getShortDeltaTime()) + "\n");
			
			sb.append("local upnpstat: avgUploadRate: "
					+ (localNetworkFact.getTotalBytesSent()/(localNetworkFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " avgDownloadRate: "
					+ (localNetworkFact.getTotalBytesReceived()/(localNetworkFact.getShortDeltaTime()!=0?systemFact.getShortDeltaTime():1)/1000) + " kB/s"
					+ " deltaTime: "
					+ (localNetworkFact.getShortDeltaTime()) + "\n");


			sb.append("local lpsd: uploadRate: "
					+ localPeerFact.getUploadRate()
					+ " avgUpload: "
					+ localPeerFact.getAverageUpload()
					+ " stdevUpload: "
					+ localPeerFact.getStdevUpload()
					+ " downloadRate: "
					+ localPeerFact.getDownloadRate()
					+ " avgDownload: "
					+ localPeerFact.getAverageDownload()
					+ " stdevDownload: "
					+ localPeerFact.getStdevDownload()
					+ " deltaTime: "
					+ localPeerFact.getLongDeltaTime() + "\n");
			
			sb.append("local netstat: bytesSent: "
					+ systemFact.getBytesSent()
					+ " avgBytesSent: "
					+ systemFact.getAverageBytesSent()
					+ " stdevBytesSent: "
					+ systemFact.getStdevBytesSent()
					+ " bytesReceived: "
					+ systemFact.getBytesReceived()
					+ " avgBytesReceived: "
					+ systemFact.getAverageBytesReceived()
					+ " stdevBytesReceived: "
					+ systemFact.getStdevBytesReceived()
					+ " deltaTime: "
					+ systemFact.getLongDeltaTime() + "\n");
			
			sb.append("local lpsd: numTorrents: "
					+ localPeerFact.getNumTorrents()
					+ " numConnctedSeeds: "
					+ localPeerFact.getNumConnectedSeeds()
					+ " avgNumConnectedSeeds: "
					+ localPeerFact.getAverageNumConnectedSeeds()
					+ " numConnctedLeechers: "
					+ localPeerFact.getNumConnectedLeechers()
					+ " avgNumConnectedLeechers: "
					+ localPeerFact.getAverageNumConnectedLeechers()
					+ " maxPeerConnections: "
					+ localPeerFact.getMaxPeerConnectionsGlobal()
					+ " maxPeerConnectionsPerTorrent: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrent()
					+ " maxPeerConnectionsPerTorrentSeeding: "
					+ localPeerFact.getMaxPeerConnectionsPerTorrentSeeding()
					+ " maxUploadRateSeeding: "
					+ localPeerFact.getMaxUploadRateSeeding()
					+ " maxSeedsPerTorrent: "
					+ localPeerFact.getMaxSeedsPerTorrent()
					+ " lazyBitfield: "
					+ localPeerFact.isLazyBitfield()
					+ " requireEncryption: "
					+ localPeerFact.isRequireEncryption()
					+ " encryptionMinLevel: "
					+ localPeerFact.getEncryptionMinLevel()
					+ " allowNonEncryptedOut: "
					+ localPeerFact.isAllowNonEncryptedOut()
					+ " allowNonEncryptedIn: "
					+ localPeerFact.isAllowNonEncryptedIn()
					+ " useCryptoPortExtension: "
					+ localPeerFact.isUseCryptoPortExtension()
					+ " tcpPortIn: " 
					+ localPeerFact.getTcpPortIn()
					+ " udpPortIn: "
					+ localPeerFact.getUdpPortIn() + "\n");
			
			LOGGER.info("DEBUG: " + sb.toString());
		}
		
		dlm.printDebug();		

		if (engine != null)
			engine.displayFacts();
		
	}

	
	public String twoDigit(Integer number) {
		
		if ( number < 10 && number >= 0) 
			return "0" + number;
		
		return number.toString();
	}
	
	public String dateToString(long date) {
		
		Calendar c = Calendar.getInstance();
		c.setTimeInMillis(date);
		
		String ret = "";
		
		ret = ret.concat(twoDigit(c.get(Calendar.MONTH)+1));
		ret = ret.concat(twoDigit(c.get(Calendar.DAY_OF_MONTH)));		
		ret = ret.concat(twoDigit(c.get(Calendar.YEAR)));
		ret = ret.concat(twoDigit(c.get(Calendar.HOUR_OF_DAY)));
		ret = ret.concat(twoDigit(c.get(Calendar.MINUTE)));
		ret = ret.concat(twoDigit(c.get(Calendar.SECOND)));
		ret = ret.concat(twoDigit(c.get(Calendar.MILLISECOND)));
		//ret = ret.concat(twoDigit(c.get(Calendar.ZONE_OFFSET)));
		
		return ret;

	}
	
	public void startUiUpdater() {
		
		//every second
		DasuTimerManager.getInstance().createEvent("NonSimulatorUiTimer", new TimerEntry( 1000, true ),
				
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {

						LOGGER.info("TIMER: NonSimulatorUiTimer fired");
						
						Long ret = Util.currentTimeMillis();
						long millis = ret - DasuManager.startUpTime; 
						String date = String.format("%d weeks, %d days, %d hour, %d min, %d sec",
								TimeUnit.MILLISECONDS.toDays(millis)/7,
								TimeUnit.MILLISECONDS.toDays(millis) - 7*(TimeUnit.MILLISECONDS.toDays(millis)/7),  		
								TimeUnit.MILLISECONDS.toHours(millis) - TimeUnit.DAYS.toHours(TimeUnit.MILLISECONDS.toDays(millis)),
								TimeUnit.MILLISECONDS.toMinutes(millis) - TimeUnit.HOURS.toMinutes(TimeUnit.MILLISECONDS.toHours(millis)),			    
							    TimeUnit.MILLISECONDS.toSeconds(millis) - TimeUnit.MINUTES.toSeconds(TimeUnit.MILLISECONDS.toMinutes(millis))
							);
						
						DasuData.getInstance().update("simulator.elapsed.time", date);
												
					}
		});		
	}
	
	public void startSimulation(Object simulationId, String baseDir) {
				
		Object seconds = DasuData.getInstance().get("simulator.second.equivalence");
		//seconds = 300;
		
		if (seconds == null)
			seconds = 300;
		
		if ( simulationId == null)
			simulationId = "65601790"; //default
			
		if (interferenceSimulator == null) {
			//interferenceSimulator = new InterferenceSimulatorV3("/files2/tmp/dasu-simulator/65601790/", 5*60);
			interferenceSimulator = new InterferenceSimulatorV3(baseDir + "/" + simulationId, Integer.parseInt(seconds.toString()));			
		    DasuData.getInstance().update("simulator.second.equivalence", seconds.toString());
		}

	}
	
	
	public void launchSimulatorShutdownListener() {
	
		DasuData.getInstance().registerListener("simulator.request.restart", new DasuDataListener() {

			public void updated(String key, Object val) {
				
					if ( ((Boolean)val) ) {
						
						try {
							
							LOGGER.info("SIMULATOR: DasuManager.class: saving simulation variables (do and id)");
							DasuData.getInstance().markPersistent("simulator.simulation.id");
							LOGGER.info("SIMULATOR: DasuManager.class: shutdown!!");
							//PluginManager.restartAzureus();
							//PluginManager.stopAzureus();
							
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				}
			});
	}
	
	protected void setYieldInterval(int timeInMillis) {
		yieldInterval = timeInMillis;
	}
	
	protected void clearYieldInterval() {
		yieldInterval = -1;
	}

	public void startAME() {
				
		if ( DasuConfiguration.getInstance().isRunExperiments() ) {
			try {
				
				if ( !AMEManager.getInstance().isJvmActive() ) {
					LOGGER.info("AME: Starting/Restarting AMEManager " + new Date());
					ameManager = AMEManager.getInstance();
					ameManager.initialize();
				}
			}
			catch (Exception e) {
				LOGGER.severe("AME: ERROR: couldn't initilize AMEManager!!! " + e.getLocalizedMessage() );
			}
		}
		else {
			LOGGER.info("Running experiments disabled by user...");
		}
		
	}
	
	public void stopAME() {

		if ( ameManager != null && ameManager.isActive() ) {
			ameManager.stop();
			ameManager = null;
		}

	}
	
	public void startMessageManager() {
		
		if ( DasuConfiguration.getInstance().isEnablePopups() )
			if ( !DasuMessageManager.getInstance().isActive() )
				pi.getUtilities().createThread("DasuMessageManager", DasuMessageManager.getInstance());
	}
	
	public void stopMessageManager() {
		DasuMessageManager.getInstance().stop();
	}

	/**
	 * Perform periodic data-collections
	 */
	  
	  public void run() {
		
		//testing();

		resourceProfiler = new ResourceProfiler(DasuConfiguration.getInstance().getResourceProfilerRecordInterval(),
				DasuConfiguration.getInstance().getResourceProfilerLimitUtilizationInterval(),
				DasuConfiguration.getInstance().getMaxAllowedCpuUtilizationProfiler(),
				new String[]{"dasu","reaper"});		
		Main.createThread("DasuResourceProfiler", resourceProfiler);

		if (!Logger.isEnabled())
			COConfigurationManager.setParameter("Logger.Enabled", true);

		LOGGER.info("DYNAMIC: loading persistentData" );
		
		synchronized(persistentDataManager) {
			Iterator<String> iter = persistentDataManager.UIValues.keySet().iterator();
			synchronized(DasuData.values) {
				while (iter.hasNext()) {
					String key = iter.next();
					DasuData.values.put(key, persistentDataManager.UIValues.get(key));
					iter.remove();
				}
			}
		}
		
		/** SIMULATION **/

		//DasuData.values.put("simulator.simulation.id", "65601790");
		//DasuData.values.put("simulator.simulation.id", "370433691");
		//DasuData.getInstance().update("simulator.play.command", "pause");
		
		Object simulationId = null;
							
		if (USE_SIMULATION_V2 && (simulationId = DasuData.getInstance().get("simulator.simulation.id")) != null) {
			startSimulation(simulationId, "/files2/tmp/dasu-simulator/");
			launchSimulatorShutdownListener();
			startUiUpdater();	
		}

		/*************/
		
		// Has to be before AMC does init
		nbp = new NetstatBanwidthPerformer(Main.getInterfaceName(), "Netstat BW");
		nbp.runOnce(1000);
		
		ubp = new UPnPBanwidthPerformer(Main.getInterfaceName(), "UPnP BW");
		ubp.runOnce(1000);
	
		Main.createThread("DasuAMCoordinator", AMCoordinatorV2.getInstance());
		if (DasuConfiguration.getInstance().isPerProbeProfilerEnabled()) {
	        perProbeProfiler = new PerProbeProfiler();
	        AMCoordinatorV2.setPerProbeProfiler(perProbeProfiler);
		}
				
		LOGGER.info("DEBUG: DasuManager.class: initializing facts");
		systemFact = new FactMeasurementSystem(new NetstatSystemStats()); /**/
		localPeerFact = new FactMeasurementLpsd(new DynamicLocalPeerStat(Util.currentGMTTime())); /**/
		localNetworkFact = new FactMeasurementUPnP(new UPnPSystemStats()); /**/


		LOGGER.info("DEBUG: DasuManager.class: starting rule engine");
		try {

			engine = new RuleEngine(SourceType.RE);
			engine.setDaemon(true);
			engine.initialize();
			
			//wait for AMC to start before starting the engine			
			int wait = 0;
			while ( !AMCoordinatorV2.isActive() && !Main.isShuttingDown()) {
				Thread.sleep(2*1000);			
				LOGGER.info("DYNAMIC: AMC hasn't started, waiting...");
				
				if ( (wait+=2) > 120 ) {
					LOGGER.severe("The AMC never started... waiting 2 minutes! unloading Dasu!!");					
					Main.startUnloadConfigTimer();
					return;
				}
			}
			
			if ( Main.isShuttingDown() )
				return;

			if (!engine.isAlive() && engine.isInitialized()) 
				engine.start();
			
			if(Main.DEBUG) {
				profiler = new Profiler();
				engine.profile(profiler);
				AMCoordinatorV2.profile(profiler);				
			} 
		}
		catch(Exception e) {
			LOGGER.severe("We couldn't load any rules, stopping Dasu client load...");
			
			ReportableFact rf = new ReportableFact("ErrorEvent");
			rf.addContentsObject("report", "DasuShutdown");
			rf.addContentsObject("reason", "We couldn't load any rules, stopping Dasu client load...");
			Statistics.getInstance().addReportableFact(rf);

			Main.startUnloadConfigTimer();
			return;
		}
		LOGGER.info("DEBUG: started rule engine");
						
		pi.getUtilities().createThread("DasuCronScheduler", DasuCronScheduler.getInstance());
		startMessageManager();
				
		if (DasuConfiguration.getInstance().isRunExperiments()) {
			
			if ( !USE_SIMULATION_V2 ) {
				
				Integer AMECheckupInterval = DasuConfiguration.getInstance().getAMECheckupInterval();
				if (Main.AME_DEBUG_SKIP_DOWNLOAD)
					AMECheckupInterval = 15;
				LOGGER.info("AME: Starting AMEManager after " + AMECheckupInterval + " seconds..." + new Date());
				
				//Start JVM after a delay for AMECheckupInterval seconds
				
				//AMETJVMStartTimer = Main.createTimer("DasuAMEJVMTimer");
				//AMETJVMStartTimer.addPeriodicEvent(AMECheckupInterval*1000,
				DasuTimerManager.getInstance().createEvent("DasuAMEJVMTimer", 
						new TimerEntry(AMECheckupInterval*1000, true), 				 
						new ITimerEventPerformer(){ 

							public void perform(ITimerEvent event) {
					
								LOGGER.info("TIMER: DasuAMEJVMTimer fired");
								
								if ( !Main.isShuttingDown()) {								
									DasuManager.getInstance().startAME();
								}
							}
						}
				);
				
			}
			else
				LOGGER.info("SIMULATOR: ActiveMeasurement Engine halted due to SIMULATION mode!!\n\n");
				
		}
		
		DasuTimerManager.getInstance().createEvent("DasuPersistentDataDumpTimer", 
				new TimerEntry( DasuConfiguration.getInstance().getDbStaticFlushInterval()*1000, true ),
				new ITimerEventPerformer() {
					public void perform(ITimerEvent event) {

						LOGGER.info("TIMER: DasuPersistentDataDumpTimer fired");
						
						if ( !Main.isShuttingDown()) {		
							DasuManager.getInstance().writePersistentData();
						}
					}
				}
		);
		
		isInitialized = true;
		
		Main.logPlugin("Inference engine enabled...");

		dlm = new DownloadMonitor(engine); /**/
		
		rm = new ResetMonitor(dlm); /**/
		rm.runOnce("Netstat Reset Monitor", 1000);
		
		launchConfigurationChangeListener();
					
		/*
		 * while is simulation, this while active loop should be performed 
		 */
		
		//AMCoordinatorV2.addPriorityProbe("38.102.0.97", ProbeType.NDT2);
		//ProbeTask pt = new ProbeTask(-1, -1, -1l, -1l, Arrays.asList(new String[]{"http://eecs.northwestern.edu"}), ProbeType.BROWSER);
		//AMCoordinatorV2.addProbeTask(pt);
		
		while (isRunning()) {
			
			//System.out.println(resourceProfiler.getAverageMemoryUsage());
			//System.out.println(resourceProfiler.getAverageMemoryUsageMB());
			//System.out.println(resourceProfiler.getMemoryUsage());
			
			if ( USE_SIMULATION_V2 ) {
				interferenceSimulator.increaseTick();
				interferenceSimulator.iterGetNextTickEvents();				
			}
			
			/*
			 * update measurement facts associated with the collected
			 * variables
			 */
					
			boolean again = false;

			do {
				
				if ( USE_SIMULATION_V2) {
					interferenceSimulator.iterNext();
				}
				
				rm.update(); /**/
				nbp.update(); /**/
				ubp.update(); /**/
	
				if (engine != null) {
	
					//Rpss facts are inserted by the PeerStats Manager 
					//asynchronously as peers disconnect, only for peers
					//that actually transmitted/received something from us
					
					dlm.updateFacts();
					
					DynamicLocalPeerStat dlps = new DynamicLocalPeerStat(Util.currentGMTTime());
					if ( dlps.uploadRate != -1 ) {						
						localPeerFact.addEntry(dlps); /**/
						engine.addFact(localPeerFact); /**/						
					}
					
					NetstatSystemStats nss = new NetstatSystemStats();
					if ( nss.timeRecorded != -1) {
						
						//System.out.println("total bytesReceived: " + nss.bytesReceived);
						
						if ((systemFact.entryList.size() ==0) || 
								(systemFact.entryList.size() >0) && ( systemFact.entryList.getLast().timeRecorded != nss.timeRecorded)) {
							
							systemFact.addEntry(nss); /**/
							engine.addFact(systemFact); /**/
							
						}
					}					
					//System.out.println("since last bytesReceived: " + systemFact.getBytesReceived() );					
					
					UPnPSystemStats uss = new UPnPSystemStats();
					if ( uss.timeRecorded != -1) {
						localNetworkFact.addEntry(new UPnPSystemStats()); /**/
						engine.addFact(localNetworkFact); /**/
					}
	
				}
		
				LOGGER.info("DEBUG: DasuManager: sleeping for " + sleepInterval + " seconds...");
	
				if ( USE_SIMULATION_V2 )
					LOGGER.info("SIMULATOR: DasuManager.class: BEFOREHALT: " + new Date(Util.currentTimeMillis())  + " eventQueue.size(): " + interferenceSimulator.eventQueue.size());
				
				synchronized (halt) {
	
					try {
						if (isActive && !again) {

							if (Main.DEBUG) {
								
								//LOGGER.info("STOP: halting...");
								halt.wait(5 * 1000);
								//LOGGER.info("STOP: halting done...");
								
							}
							else
								halt.wait(sleepInterval * 1000);
						}
					} catch (InterruptedException e) {
						e.printStackTrace();
					}	
				}
				
				if ( USE_SIMULATION_V2 ) {
					LOGGER.info("SIMULATOR: DasuManager.class: AFTERHALT: " + new Date(System.currentTimeMillis()) + " eventQueue.size(): " + interferenceSimulator.eventQueue.size());
					checkPausedSimulation();
					
					if (interferenceSimulator.iterHasNext())
						again = true;
					else
						again = false;
				}
				
				if ( yieldInterval != -1) {
					try {
						Thread.sleep(yieldInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				
				if ( Main.DEBUG ) {
					printDebug();
					LOGGER.info("#######\nPERSISTENT: " + DasuData.getInstance().values + "\n########");
				}
				
			} while (again);
			
		}// while
		
		LOGGER.info("STOP: Adios...");

	}	
	
	private void checkPausedSimulation() {
		
		String command = (String) DasuData.getInstance().get("simulator.play.command");										
		while ( (command != null) && command.equals("pause") ) {

			try {
				LOGGER.info("SIMULATOR: DasuManager.class: paused!");
				Thread.sleep(1*1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			command = (String) DasuData.getInstance().get("simulator.play.command");
		}
		
		if ( command != null && command.equals("step") )
			DasuData.getInstance().update("simulator.play.command", "pause");
		
	}
	
	public boolean isRunning() {
		return isActive;
	}

	public void stop() {
		
		LOGGER.info("STOP: initiating DasuManager unload...");
		
		isActive = false;
		synchronized (halt) {
			halt.notify();
			LOGGER.info("STOP: halting keeps going now!");
		}

		LOGGER.info("STOP: Stopping rule engines");

		if (engine != null && engine.isAlive()) 
			engine.stopEngine();

		if ( nbp != null)
			nbp.stop();
		if (ubp != null)
			ubp.stop();
				
		if(Main.DEBUG) {
			if ( profiler != null)
				profiler.stop();
		}
		
		writePersistentData();
		
		DasuCronScheduler.getInstance().stop();
		stopMessageManager();
		stopAME();
		
		if (AMCoordinatorV2.isActive()) {
			AMCoordinatorV2.stop();
		}
		
		if (resourceProfiler.isActive()) {
			LOGGER.info("STOP: Stopping ResourceProfiler");
			resourceProfiler.stop();
		}

//		LOGGER.info("STOP: Stopping rule engines");
//
//		if (engine != null && engine.isAlive()) 
//			engine.stopEngine();
		
		LOGGER.info("STOP: Destryoing timers...");
		if(dlm != null) dlm.remove();		
		if(rm != null) rm.destroy();
		DasuTimerManager.getInstance().destroyTimer("DasuAMEJVMTimer");
		DasuTimerManager.getInstance().destroyTimer("DasuNonSimulatorUiTimer");
		DasuTimerManager.getInstance().destroyTimer("DasuPersistentDataDumpTimer");
		//if (AMETJVMStartTimer != null ) AMETJVMStartTimer.destroy();
		//if (tickTimer != null) tickTimer.destroy();
		//if (PersistentDataDumpTimer != null) PersistentDataDumpTimer.destroy();
				
		for (String param : COConfigurationListeners.keySet()) {
			COConfigurationManager.removeParameterListener(param, COConfigurationListeners.get(param));
		}

		Main.logPlugin("Inference engine disabled...");		
		try {
			finalize();
		} catch (Throwable e) {
			e.printStackTrace();
		}
		
		LOGGER.info("STOP: DasuManager: Buh-bye");
		self = null;

	}
	
	public void writePersistentData() {
		
		synchronized(persistentDataManager) {
					
			for ( String key : DasuData.persistentKeys ) {
				if ( DasuData.values.containsKey(key)) {
					persistentDataManager.UIValues.put(key, DasuData.values.get(key));
				}
			}
			
			LOGGER.info("SERIALIZING: serializing persistent data...");
			Main.saveObject(persistentDataManager, persistentDataFileName);
		}
		
	}
	
	public InterferenceSimulatorV3 getSimulator() {
		return interferenceSimulator;
	}
	

}
