/**
 * Dasu Project
 *
 * File:         ActiveMeasurementEngine.java
 * RCS:          $Id$
 * Description:  ActiveMeasurementEngine class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:28:49 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.drools
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.drools;

import java.io.DataInputStream;
import java.net.URL;
import java.util.logging.Logger;

import org.drools.compiler.DrlParser;
import org.drools.io.Resource;
import org.drools.io.ResourceFactory;
import org.drools.lang.descr.FunctionDescr;
import org.drools.lang.descr.ImportDescr;
import org.drools.lang.descr.PackageDescr;
import org.drools.lang.descr.RuleDescr;

import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.measurement.AMCoordinatorV2Interface;
import edu.northwestern.dasu.security.DigitalSignatureManager;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The ActiveMeasurementEngine class ...
 */
public class ActiveMeasurementEngine extends RuleBasedEngine {

	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	private AMCoordinatorV2Interface AMCoordinator = null;		
	private StringBuilder rulesLoadFailedReason = new StringBuilder();
	
	public ActiveMeasurementEngine(String filename, String sigFilename, Integer taskId) throws Exception {
		super("DasuActiveMeasurementEngine", filename, sigFilename, taskId);
	}

	public ActiveMeasurementEngine(String filename, String sigFilename, Integer taskId,
			AMCoordinatorV2Interface AMCoordinator, String engineId) throws Exception {
		super("DasuActiveMeasurementEngine_" + engineId, filename, sigFilename, taskId);
		this.AMCoordinator = AMCoordinator;
	}

	public boolean loadRules() {
		
		LOGGER.info("RULEENGINE: Verify signature of downloaded file...");		
		
		boolean verified = DigitalSignatureManager.verify(filename, sigFilename);
		Integer inVer = extractTaskId(filename);
		
		if ( !verified ) {
			LOGGER.warning("AME: Experiment could not be verified. Invalid digital signature! ");
			appendFailReason("Experiment could not be verified. Invalid digital signature!");			
		}
		
		if (!inVer.equals(taskId)) {
			verified = false;
			LOGGER.warning("AME: Invalid file! TaskId differ! " + taskId + " " + inVer);
			appendFailReason("Invalid file! TaskId differ! " + taskId + " " + inVer);
		}
		
		if ( !ASTVerification(filename) ) {
			verified = false;
			LOGGER.warning("AME: Invalid experiment file! AST verification failed");
			appendFailReason("Invalid experiment file! AST verification failed");
		}
		
		if(Main.AME_DEBUG_SKIP_VERIFICATION) {
			LOGGER.info("AME: ignoring verification because of DEBUG mode");
			verified = true;
			inVer = taskId;
		}
		
		if ((verified) && ( inVer.equals(taskId))) {
			
			kbase = loadRules(filename);
			ksession = kbase.newStatefulKnowledgeSession();
			FactHalt halt= new FactHalt(false);
			ksession.setGlobal("halt", halt);
			if ( AMCoordinator != null) {
				ksession.setGlobal("AMCoordinator", AMCoordinator);
			}
			
			return true;
		}
		
		return false;
	}
	
	public void setAMCoordinator(AMCoordinatorV2Interface coord) {

		ksession.setGlobal("AMCoordinator", coord);
		
	}
	
	@Override
	protected void destroySession() {
		
		if (ksession != null)
			ksession.dispose();
		
		setActive(false);
		
	}
	
	protected Integer extractTaskId(String drlFileName) {

		drlFileName = "file:" + drlFileName;
		Integer version = -1;

		try {
			URL url = new URL(drlFileName);
			if (url != null) {
	
				DataInputStream in = new DataInputStream(url.openStream());
				String line;
	
				while ( (line = in.readLine()) != null ) {
					
					if ( line.contains("_v_") ) {
						String[] f = line.split("_");
						try {
							version = Integer.parseInt(f[2]);
						}
						catch (Exception e) {
						}
					}
				}
				
			}
		} catch (Exception e) {
			
		}
		
		return version;
	}
	
	private boolean ASTVerification(String rulesfile) {
		
		try {
			Resource r = ResourceFactory.newFileResource(rulesfile);
			DrlParser dp = new org.drools.compiler.DrlParser();			
			PackageDescr pkgd = dp.parse(r.getInputStream());			
			
			//limit imports outside java namespace and edu.northwestern
			for ( ImportDescr imp : pkgd.getImports()) {
				if ( !imp.toString().contains("edu.northwestern") &&
						!imp.toString().contains("import java")) {
						//!imp.toString().contains("java.util") && 
						//!imp.toString().contains("java.lang") &&
						//!imp.toString().contains("java.io")) {
					LOGGER.info("AME: unauthrized import, abort... " + imp.toString() );
					appendFailReason("AST: unauthrized import, abort... " + imp.toString());
					return false;
				}
			}

			//loop over functions and avoid while loops
			for ( FunctionDescr f : pkgd.getFunctions()) {
				
				if ( f.getText().contains("while")) {
					LOGGER.info("AME: while loop found in function, abort...");
					appendFailReason("AST: while loop found in function, abort...");
					return false;
				}
			}
			
			//loop over rules and avoid while loops
			for ( RuleDescr rd : pkgd.getRules()) {
				
				if ( rd.getConsequence().toString().contains("while")) {
					LOGGER.info("AME: while loop found in rule, abort...");
					appendFailReason("AST: while loop found in rule, abort...");
					return false;
				}
				
			}
			

		}
		catch (Exception e ){
			LOGGER.warning("AME: " + e.getMessage());			
			return false;
		}

		return true;
	}
	
	private void appendFailReason(String reason) {		
		this.rulesLoadFailedReason.append("- " + reason + "\n");
	}

	public String getRulesLoadFailedReason() {		
		return rulesLoadFailedReason.toString();
	}
	
	
}
