/**
 * Dasu Project
 *
 * File:         CuratorConnection.java
 * RCS:          $Id: CuratorConnection.java,v 1.0 Oct 23, 2013 3:09:01 PM mas939 Exp $
 * Description:  CuratorConnection class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 23, 2013 at 3:09:01 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.dynamic.coordination
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2013, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.coordination;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The CuratorConnection class ...
 */

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import org.apache.zookeeper.CreateMode;
import org.apache.zookeeper.KeeperException.NoNodeException;
import org.apache.zookeeper.data.Stat;

import com.netflix.curator.RetryPolicy;
import com.netflix.curator.framework.CuratorFrameworkFactory;
import com.netflix.curator.framework.CuratorTempFramework;
import com.netflix.curator.retry.ExponentialBackoffRetry;
import com.netflix.curator.retry.RetryNTimes;

import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONArray;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONValue;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.parser.JSONParser;
import edu.northwestern.dasu.measurement.types.ProbeType;

public class CuratorConnectionNonPersistent {
	
	CuratorTempFramework client;	
	String taskID;
	String UUID;
	JSONObject defaultNodeInfo;	
	Stat leaseZNode;
	
	protected final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	
	private Boolean connectionFailed = false;
	private Long connectionFailedSince = -1l;
	
	private final String nameSpace_format = "dasu/%s"; // "dasu/TASKID"; 	
	private final String reportProbePath_format = "%s/probes"; // "/UUID/probes/";
	private final String leasePath_format = "%s/lease"; // "/UUID/lease";
		
	String reportProbePath;
	String leasePath;	
	
	private static boolean DEBUG = false;

	public CuratorConnectionNonPersistent(String hostPort, String taskID, String UUID, String ip, String cc, String asn) throws Exception {
				
		this.taskID = taskID;
		this.UUID = UUID;
		this.defaultNodeInfo = new JSONObject(); 
		defaultNodeInfo.put("ip", ip);
		defaultNodeInfo.put("cc", cc);
		defaultNodeInfo.put("asn", asn);
		 		
		// retry connecting to a server 2 times when disconnected
		client = createWithOptions(hostPort, String.format(nameSpace_format, taskID), new RetryNTimes(2, 1000));

		reportProbePath = String.format(reportProbePath_format, UUID);
		leasePath = String.format(leasePath_format, UUID);		
		createPath( reportProbePath );
		createPath( leasePath );
		updateMockLease();

	}
	
	private void setConnectionFailed(boolean failed) {
		
		if (failed) {
		
			//update connectionFailedSince if connection just failed
			if (!connectionFailed)
				connectionFailedSince = System.currentTimeMillis();
		}
		else {
			
			//otherwise rest timestamp
			connectionFailedSince = -1l;
		}
				
		connectionFailed = failed;
	}
	
	public boolean isConnectionFailed() {
		return connectionFailed;
	}
	
	public Long getConnectionFailedSince() {
		return connectionFailedSince;
	}
	
	// path must exist
 	private void setDataPath(String path, String data) throws Exception {
		client.inTransaction().setData().forPath(path, data.getBytes()).and().commit();
	}
	
	private void createPath(String createPath) throws Exception {

		String path = "";
		for ( String p : createPath.split("/") ) {
			path = path.concat("/" + p);
			LOGGER.info(path);

			try  {
				client.inTransaction().check().forPath(path).and().commit();
			}
			catch (NoNodeException e) {				
				client.inTransaction().create().withMode(CreateMode.PERSISTENT).forPath(path).and().commit();
			}
			
			if ( p.equals(UUID)) {
				defaultNodeInfo.put("localTime", System.currentTimeMillis());
				client.inTransaction().setData().forPath(path, defaultNodeInfo.toJSONString().getBytes()).and().commit();
			}			
			
		}
		
	}
	
	/**
	 * 	@returns JSONObject representing a lease with the following format:
	 *   
	 * { 'version': 1.0, 
	 *  'leaseExpiration': date,
	 *	'reportInterval': seconds,    	
	 *	'budget': {
	 *				'unitTimeForBudgetLimits': seconds,
	 *				'probeTypes: {
	 *						'PING': [ ('0.0.0.0/0', 30, 0.1), ('207.110.223.0/19', 30, 0.2).... ],
	 *						'TRACEROUTE': [ ('0.0.0.0/0',  -1, 1), (192.168.0.1/24', 22, 0.1)....],
	 *						....
	 *						}
	 *				}
	 * }
	 * 
	 * or null if no lease was successfully fetched.
	 */
	
	public void updateMockLease() throws Exception {
		setDataPath( leasePath, mockLease().toJSONString() );
	}
	
	public JSONObject PullUpdatedLease() throws Exception {
		
		setConnectionFailed(true);
		Stat tempLeaseZNode = new Stat();	
		client.getData().storingStatIn(tempLeaseZNode).forPath(leasePath);
		
		if ( leaseZNode != null ) {
			
			// lease hasn't been updated
			if (leaseZNode.getVersion() == tempLeaseZNode.getVersion()) {
				LOGGER.info("Lease hasn't been updated...");
				return null;
			}
			
		}
		
		leaseZNode = tempLeaseZNode;
		String lease = new String(client.getData().forPath(leasePath));
		
		JSONObject jsonLease = (JSONObject)new JSONParser().parse(lease);
		setConnectionFailed(false);
		return jsonLease;
		
	}
	
	public JSONObject mockLease() {
		
		LOGGER.info("initializing mock lease...");
		JSONObject ret = new JSONObject();
		ret.put("version", 1.0);		
		ret.put("leaseExpiration", System.currentTimeMillis() + 30*1000);
		ret.put("leaseStart", System.currentTimeMillis());
		ret.put("reportInterval", 10);
		
		
		JSONObject probeTypes = new JSONObject();
		for ( String probeType : new String[]{"PING", "TRACEROUTE", "HTTP"} ) {
			
			JSONArray entries = new JSONArray();
			
			JSONArray entry = new JSONArray();
			Object[] ips = {"0.0.0.0/0", 30, 01., "207.110.223.0/19", 1, 0.2, "192.168.0.1/23", 2, 0.5};
			for (int i =0; i < ips.length; i++ ) {		
				
				entry.add(ips[i]);			
				
				if ( (i+1)%3 == 0) {
			
					entries.add(entry);
					entry = new JSONArray();
				}
			}
			
			probeTypes.put(probeType, entries);
		}
		
		ret.put("budget", new HashMap<String, Object>());
		
		JSONObject budget = new JSONObject();

		budget.put("timeIntervalForBudgetLimits", 15);
		budget.put("probeTypes", probeTypes);
		
		ret.put("budget", budget);
		
		return ret;
				
		
	}
	
	public boolean reportProbedIPs(Map<ProbeType, List<String>> probeTypeToIPsMap) {
		
		boolean ret = true;
		for ( ProbeType probeType : probeTypeToIPsMap.keySet()) {
			ret = ret && reportProbedIPs(probeTypeToIPsMap.get(probeType), probeType);
		}		
		
		return ret;
	}
	
	public boolean reportProbedIPs(List<String> listOfIps, ProbeType probeType) {
		
		if ( listOfIps.size() == 0) {
			LOGGER.info("nothing to report...");
			return false;
		}
		
		String path = String.format(reportProbePath_format + "/%s", UUID, probeType.getEventType());		
				
		try {
						
			try  {
				client.inTransaction().check().forPath(path).and().commit();
				LOGGER.info("path for probetype exists don't create... " + path);
			}
			catch (NoNodeException e) {
				//e.printStackTrace();
				client.inTransaction().create().withMode(CreateMode.PERSISTENT).forPath(path).and().commit();
			}
			
		}
		catch (Exception e) {
			if (DEBUG)
				e.printStackTrace();
			setConnectionFailed(true);
			return !isConnectionFailed();
		}
		
		String ipsText = JSONValue.toJSONString(listOfIps);
		String nodeName = new SimpleDateFormat("yyyy-MM-dd_hh-mm-ss'.tsv'").format(new Date());
		try {
			
			path = path.concat("/" + nodeName);
			client.inTransaction().create().withMode(CreateMode.PERSISTENT).forPath(path).and()
					.setData().forPath(path, ipsText.getBytes()).and().commit();
			LOGGER.info("reporting " + path);

		}
		catch (Exception e) {
			if (DEBUG)
				e.printStackTrace();
			setConnectionFailed(true);
			return !isConnectionFailed();
		}
		
		setConnectionFailed(false);
		return !isConnectionFailed();
		
	}
	
    private CuratorTempFramework createWithOptions(String connectionString, String namespace, RetryPolicy retryPolicy) {
    	
    	if (retryPolicy == null) {
    		retryPolicy = new RetryNTimes(3, 1000);
    	}
    	
        // using the CuratorFrameworkFactory.builder() gives fine grained control
        // over creation options. See the CuratorFrameworkFactory.Builder javadoc
        // details
        return CuratorFrameworkFactory.builder()
            .connectString(connectionString)
            .retryPolicy(retryPolicy)
            .namespace(namespace)
            .buildTemp(2000, TimeUnit.MILLISECONDS);
        
    }

	/**
	 * @param sessionTimeoutMs:
	 * When the ZK client establishes a session with the cluster it provides a
	 * "timeout" value. This value is used by the cluster to determine when the client's session
	 * expires. Expirations happens when the cluster does not hear from the client within the specified
	 * session timeout period (i.e. no heartbeat). Session expired means that the client was partitioned 
	 * off from the ZooKeeper service for more the the session timeout and ZooKeeper decided that the 
	 * client died.
	 * 
	 * @param connectionTimeoutMs:
	 * When a client (session) becomes partitioned from the ZK serving cluster it will begin searching 
	 * the list of servers that were specified during session creation. Eventually, when connectivity 
	 * between the client and at least one of the servers is re-established, the session will either 
	 * again transition to the "connected" state (if reconnected within the session timeout value) or 
	 * it will transition to the "expired" state (if reconnected after the session timeout). 
	 * 
	 */
	
    private CuratorTempFramework  createWithOptions(String connectionString, String namespace, 
    		RetryPolicy retryPolicy, int connectionTimeoutMs, int sessionTimeoutMs) {
    	
    	if (retryPolicy == null) {
    		retryPolicy = new ExponentialBackoffRetry(1000, 3);
    	}
    	
        // using the CuratorFrameworkFactory.builder() gives fine grained control
        // over creation options. See the CuratorFrameworkFactory.Builder javadoc
        // details
        return CuratorFrameworkFactory.builder()
            .connectString(connectionString)
            .retryPolicy(retryPolicy)
            .connectionTimeoutMs(connectionTimeoutMs)
            .sessionTimeoutMs(sessionTimeoutMs)
            .namespace(namespace)
            .buildTemp();
    }

}

 
