/**
 * Dasu Project
 *
 * File:         CuratorConnection.java
 * RCS:          $Id: CuratorConnection.java,v 1.0 Oct 23, 2013 3:09:01 PM mas939 Exp $
 * Description:  CuratorConnection class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 23, 2013 at 3:09:01 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.dynamic.coordination
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2013, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.coordination;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The CuratorConnection class ...
 */

import java.util.HashMap;
import java.util.List;

import org.apache.zookeeper.CreateMode;

import com.netflix.curator.RetryPolicy;
import com.netflix.curator.framework.CuratorFramework;
import com.netflix.curator.framework.CuratorFrameworkFactory;
import com.netflix.curator.framework.api.transaction.CuratorTransaction;
import com.netflix.curator.framework.api.transaction.CuratorTransactionFinal;
import com.netflix.curator.retry.ExponentialBackoffRetry;
import com.netflix.curator.retry.RetryNTimes;

import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
//import edu.northwestern.cs.aqualab.cattle.jsonsimple.parser.JSONParser;

public class CuratorConnectionPersistent {
	
	CuratorFramework client;	
	String taskID;
	String UUID;
	JSONObject defaultNodeInfo;	
	
	private boolean connectionFailed = false;
	
	private final String nameSpace_format = "dasu/%s"; // "dasu/TASKID"; 	
	private final String reportProbePath_format = "%s/probes"; // "/UUID/probes/";
	private final String leasePath_format = "%s/lease"; // "/UUID/lease";
	

	public CuratorConnectionPersistent(String hostPort, String taskID, String UUID, String ip, String cc, String asn) throws Exception {
				
		this.taskID = taskID;
		this.UUID = UUID;
		this.defaultNodeInfo = new JSONObject(); 
		defaultNodeInfo.put("ip", ip);
		defaultNodeInfo.put("cc", cc);
		defaultNodeInfo.put("asn", asn);
		 
		// retry connecting to a server 2 times when disconnected
		client = createWithOptions(hostPort, String.format(nameSpace_format, taskID), new RetryNTimes(2, 1000));
		client.start();

		createhPath(reportProbePath_format, new String[]{UUID});
		createhPath(leasePath_format, new String[]{UUID});

	}
	
	public boolean isConnectionFailed() {
		return connectionFailed;
	}
	
	private void createhPath(String format, String... args) throws Exception {

		String path = "";
		for ( String p : String.format(format, args).split("/") ) {
			path = path.concat("/" + p);
			System.out.println(path);

			if ( client.checkExists().forPath(path) == null ) {
				client.create().withMode(CreateMode.PERSISTENT).forPath(path);
			}
			
			// add node info to UUID znode
			if ( p.equals(UUID)) {
				client.setData().forPath(path, defaultNodeInfo.toJSONString().getBytes());
			}
		}
		
	}
	
	/**
	 * 	@returns JSONObject representing a lease with the following format:
	 *   
	 * { 'leaseExpiration': date,
	 *	'reportInterval': seconds,    	
	 *	'budget': {
	 *				'unitTimeForBudgetLimits': seconds,
	 *				'probeTypes: {
	 *						'PING': [ ('0.0.0.0/0', 30, 0.1), ('207.110.223.0/19', 30, 0.2).... ],
	 *						'TRACEROUTE': [ ('0.0.0.0/0',  -1, 1), (192.168.0.1/24', 22, 0.1)....],
	 *						....
	 *						}
	 *				}
	 * }
	 * 
	 * or null if no lease was successfully fetched.
	 */
	
	public JSONObject PullUpdatedLease(String TaskID) {
		return null;
	}
	
	public void reportProbedIPs(HashMap<String, List<String>> probeTypeToIPsMap) {
	
		for ( String probeType : probeTypeToIPsMap.keySet()) {
			reportProbedIPs(probeTypeToIPsMap.get(probeType), probeType);
		}
	}
	
	private CuratorTransactionFinal recursiveDeleteTransaction(CuratorTransaction transaction, String path) {
		
		System.out.println("****" + path + "****");
		try {
			
			List<String> children = client.getChildren().forPath(path);
			
			if ( children.size() == 0 ) {
				System.out.println("deleting " + path);
				return transaction.delete().forPath(path).and();				
			}
			else {
				
				for (String child : children ) {					
					recursiveDeleteTransaction(transaction, path + "/" + child);
					System.out.println("deleting " + path);					
				}
				return transaction.delete().forPath(path).and();
			}
		}
		catch (Exception e){
			e.printStackTrace();
		}
		
		return null;
		
	}
		
	public boolean reportProbedIPs(List<String> listOfIPs, String probeType) {
		
		String path = String.format(reportProbePath_format + "/%s", UUID, probeType);		
				
		try {
			
			/*
			// if path for probeType exists, delete and report probes in last period
			if ( client.checkExists().forPath(path) != null ) {
				System.out.println("path for probetype exists, delete... " + path);
				recursiveDeleteTransaction(client.inTransaction(), path).commit();
			}
			
			client.create().withMode(CreateMode.PERSISTENT).forPath(path);
			*/
			
			if ( client.checkExists().forPath(path) == null ) {				
				client.create().withMode(CreateMode.PERSISTENT).forPath(path);
			}
			
		}
		catch (Exception e) {
			e.printStackTrace();
			connectionFailed = true;
			return false;
		}
		
		for ( String ip : listOfIPs) {
						
			System.out.println(ip);
			String parts[] = ip.split("\\.");
			if(parts.length != 4) {
				System.out.println("Invalid IP address"+ip);
				return false;
			}

			try {
				
				path = String.format(reportProbePath_format + "/%s", UUID, probeType);
				
				for ( String octect : parts ) {
					
					path = path.concat("/" + octect);
					if ( client.checkExists().forPath(path) == null ) {
						client.create().withMode(CreateMode.PERSISTENT).forPath(path);
						System.out.println("creating subpath " + path);
					}
					else {
						System.out.println("path already exists... " + path);
					}
				}
				
				//byte[]	
				
				//Stat node = client.checkExists().forPath(path);	
				String val = new String(client.getData().forPath(path));
				System.out.println(val);
				//JSONObject json = (JSONObject)new JSONParser().parse(val);
				//System.out.println(json);
				//json.put("count", (Long)json.get("count")+1);				
				//client.setData().forPath(path, json.toJSONString().getBytes());				
				
			}
			catch (Exception e) {
				e.printStackTrace();
				connectionFailed = true;
				return false;
			}
		}
		
		return true;
		
	}
	
    private CuratorFramework createWithOptions(String connectionString, String namespace, RetryPolicy retryPolicy) {
    	
    	if (retryPolicy == null) {
    		retryPolicy = new RetryNTimes(3, 1000);
    	}
    	
        // using the CuratorFrameworkFactory.builder() gives fine grained control
        // over creation options. See the CuratorFrameworkFactory.Builder javadoc
        // details
        return CuratorFrameworkFactory.builder()
            .connectString(connectionString)
            .retryPolicy(retryPolicy)
            .namespace(namespace)
            .build();
    }

	/**
	 * @param sessionTimeoutMs:
	 * When the ZK client establishes a session with the cluster it provides a
	 * "timeout" value. This value is used by the cluster to determine when the client's session
	 * expires. Expirations happens when the cluster does not hear from the client within the specified
	 * session timeout period (i.e. no heartbeat). Session expired means that the client was partitioned 
	 * off from the ZooKeeper service for more the the session timeout and ZooKeeper decided that the 
	 * client died.
	 * 
	 * @param connectionTimeoutMs:
	 * When a client (session) becomes partitioned from the ZK serving cluster it will begin searching 
	 * the list of servers that were specified during session creation. Eventually, when connectivity 
	 * between the client and at least one of the servers is re-established, the session will either 
	 * again transition to the "connected" state (if reconnected within the session timeout value) or 
	 * it will transition to the "expired" state (if reconnected after the session timeout). 
	 * 
	 */
	
    private CuratorFramework  createWithOptions(String connectionString, String namespace, 
    		RetryPolicy retryPolicy, int connectionTimeoutMs, int sessionTimeoutMs) {
    	
    	if (retryPolicy == null) {
    		retryPolicy = new ExponentialBackoffRetry(1000, 3);
    	}
    	
        // using the CuratorFrameworkFactory.builder() gives fine grained control
        // over creation options. See the CuratorFrameworkFactory.Builder javadoc
        // details
        return CuratorFrameworkFactory.builder()
            .connectString(connectionString)
            .retryPolicy(retryPolicy)
            .connectionTimeoutMs(connectionTimeoutMs)
            .sessionTimeoutMs(sessionTimeoutMs)
            .namespace(namespace)
            .build();
    }

}

 
