/**
 * Dasu Project
 *
 * File:         lease.java
 * RCS:          $Id: lease.java,v 1.0 Oct 22, 2013 12:12:39 PM mas939 Exp $
 * Description:  lease class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 22, 2013 at 12:12:39 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.coordination
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2013, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.coordination;

import java.util.HashMap;
import java.util.Map;

import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.util.Pair;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The lease class ...
 */

public class ExperimentLease {

    long FOREVER;
    long ANY;
    
    private float version;
    private long leaseStart;	//absolute value in milliseconds from the epoch
    private long leaseExpiration; //absolute value in milliseconds from the epoch
    private long reportInterval;
    
    private Budget budget;
    
    protected ExperimentLease(float version, long leaseStart, long expiration, long reportInterval, long timeIntervalForBudgetLimits) {
    	this.version = version;
    	this.leaseExpiration = expiration;
    	this.leaseStart = leaseStart;
    	this.reportInterval = reportInterval;       	
    	budget = new Budget(timeIntervalForBudgetLimits);
    }
    
    public long getExpiration() {
    	return this.leaseExpiration;
    }
    
    public boolean isExpired() {
    	return System.currentTimeMillis() > this.leaseExpiration;
    }
    
    public long getLeaseDurationInMillis() {
    	return leaseExpiration - leaseStart;
    }
    
    public long getReportInterval() {
    	return this.reportInterval;
    }
    
    public Budget getBudget() {
    	return budget;
    }     
    
    public String toString() {
    	return "version: " + version + " leaseStart: " + leaseStart + " leaseExpiration: " + leaseExpiration + 
    			" reportInterval: " + reportInterval + " budget: " + budget;
    }
     
	/*
	 {  'version': 1.0,
	 	'leaseExpiration': date,
	 	'leaseStart': date,
		'reportInterval': seconds,    	
		'budget': {
					'timeIntervalForBudgetLimits': minutes,
					'probeTypes: {
							'PING': [ ('0.0.0.0/0', 30, 0.1), ('207.110.223.0/19', 30, 0.2).... ],
							'TRACEROUTE': [ ('0.0.0.0/0',  -1, 1), (192.168.0.1/24', 22, 0.1)....],
							....
							}
					}
	 }
	*/

    
    public class Budget {
    	    	
    	long timeIntervalForBudgetLimits;    	
    	Map<ProbeType, Map<String, BudgetEntry>> probeTypeBudget;
    	
    	
    	public Budget(long timeIntervalForBudgetLimits) {
    		this.timeIntervalForBudgetLimits = timeIntervalForBudgetLimits;
    		this.probeTypeBudget = new HashMap<ProbeType, Map<String, BudgetEntry>>();
    		
    	}
    	
    	public boolean appendProbeTypeLimit(String pt, String prefix, long numberOfProbes, float percentLeewayAboveLimit) {
    		
    		if ( ProbeType.valueOf(pt) != null)
    			return appendProbeTypeLimit( ProbeType.valueOf(pt), prefix, numberOfProbes, percentLeewayAboveLimit);
    			
    		return false;
    		
    	}    	    	
    	
    	public boolean appendProbeTypeLimit(ProbeType pt, String prefix, long numberOfProbes, float percentLeewayAboveLimit) {
    		
    		if (!probeTypeBudget.containsKey(pt)) 
    			probeTypeBudget.put(pt, new HashMap<String, BudgetEntry>());
    		
    		// a new entry for the same prefix will override existing ones		
    		if ( probeTypeBudget.get(pt).containsKey(prefix) )
    			probeTypeBudget.get(pt).remove(prefix);
    		
    		probeTypeBudget.get(pt).put( prefix, new BudgetEntry(numberOfProbes, percentLeewayAboveLimit) );
    		return true;
    		
    	}
    	
    	public boolean probeLimitExistsForPrefix(ProbeType pt, String prefix) {
    		return probeTypeBudget.containsKey(pt) && probeTypeBudget.get(pt).containsKey(prefix);
    	}
    	
    	public Long getMaxNumberOfProbes(ProbeType pt, String prefix) {
    		
    		if ( probeLimitExistsForPrefix(pt, prefix) )
    			return probeTypeBudget.get(pt).get(prefix).getNumberOfProbes();
    		
    		return null;
    			
    	}
    	
    	public long getTimeIntervalForBudgetLimits() {
    		return timeIntervalForBudgetLimits;
    	}
    	
    	public Float getPercentLeewayAboveLimit(ProbeType pt, String prefix) {
    		
    		if ( probeLimitExistsForPrefix(pt, prefix) )
    			return probeTypeBudget.get(pt).get(prefix).getPercentLeewayAboveLimit();
    		
    		return null;
    		
    	}
    	
    	public Pair<Long, Float> getLimits(ProbeType pt, String prefix) {

    		if ( probeLimitExistsForPrefix(pt, prefix) ) {
    			return new Pair<Long, Float>(
    					getMaxNumberOfProbes(pt, prefix),
    					getPercentLeewayAboveLimit(pt, prefix)
    					);
    		}

    		return null;
    	}
    	
    	public String toString() {
    		return "timeIntervalForBudgetLimits: " + this.timeIntervalForBudgetLimits + " " + this.probeTypeBudget;
    	}
    	
    	public class BudgetEntry {
    		
    		private long numberOfProbes;
    		private float percentLeewayAboveLimit;

    		public BudgetEntry( long numberOfProbes, float percentLeewayAboveLimit) {    			
    			this.numberOfProbes = numberOfProbes;
    			this.percentLeewayAboveLimit = percentLeewayAboveLimit;
    		}
    		
        	public long getNumberOfProbes() {
        		return numberOfProbes;
        	}
        	
        	public Float getPercentLeewayAboveLimit() {
        		return percentLeewayAboveLimit;
        	}
        	
        	public String toString() {
        		return "numProbes: " + numberOfProbes + " pctLeewayAboveLimit: " + percentLeewayAboveLimit;
        	}

    	}
    }

}