/**
 * Dasu Project
 *
 * File:         SubnetManager.java
 * RCS:          $Id: SubnetManager.java,v 1.0 Oct 23, 2013 10:25:25 AM mas939 Exp $
 * Description:  SubnetManager class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 23, 2013 at 10:25:25 AM
 * Language:     Java
 * Package:      edu.northwestern.dasu.dynamic.coordination
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2013, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.coordination;

import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The SubnetManager class ...
 */
public class SubnetManager {
	
	Trie<Subnet> subnets = new Trie<Subnet>();
	
	public SubnetManager() {
		
	}
		
    /**
     * @returns The longest prefix that matches the ip, the default
     * 			route if one was defined or null if no match can 
     * 			be found.
     */
	public String getLongestPrefixMatch(String ip) {
		
		Subnet sn = new Subnet(ip);
    	Subnet max = null;
    	for (Subnet s : subnets.getWordsAlongPrefix(sn)) {
    		
    		if ( max == null)
    			max = s;
    		else {
    			if ( s.length > max.length)
    				max = s;
    		}            
    	}
    	
    	if ( max != null)
    		return max.getPrefix();
    	
    	if ( containsPrefix("0.0.0.0/0") )
    		return "0.0.0.0/0";
    	
    	return null;

	}
	
	public boolean containsPrefix(String prefix) {
		return subnets.containsWord(new Subnet(prefix));
	}

	public void addSubnet(String prefix, String label) {
		subnets.addWord(new Subnet(prefix, label));
	}

	public void addSubnet(String prefix) {
		subnets.addWord(new Subnet(prefix));
	}
	
    /**
     *  An ip is simply a /24 prefix 
     */
	public void addIp(String ip) {
		addSubnet(ip);
	}

	public void addIp(String ip, String label) {
		addSubnet(ip, label);
	}
	
 	public class Subnet implements Comparator<Subnet> {

	    String binary;
	    Integer length;
	    String prefix;
	    String label;
	    String format = "%1$08d";
	    
	    public Subnet(String prefix) {
	    	
	    	if ( !prefix.contains("/") ) 
	    		prefix = prefix.concat("/24");
	    	
	    	this.length = Integer.parseInt(prefix.split("/")[1]);
	    	this.binary = toBinary(prefix.split("/")[0]).substring(0, length);
	    	this.prefix = prefix;
	    	this.label = prefix;
	    	
	    }

	    public Subnet(String prefix, String label) {
	    	
	    	this.length = Integer.parseInt(prefix.split("/")[1]);
	    	this.binary = toBinary(prefix.split("/")[0]).substring(0, length);
	    	this.prefix = prefix;
	    	this.label = label;
	    	
	    }

	    private String toBinary(String ip) {
	    	
	    	String ipBinary = "";	    	
	    	
			for ( String octect : ip.split("\\.")) {
				String binary = Integer.toBinaryString(Integer.parseInt(octect));
				ipBinary = ipBinary.concat( String.format(format, Integer.parseInt(binary) ) );
			}
			
			return ipBinary;

	    }
	    
	    public String getBinaryRepresentation() {
	    	return binary;
	    }
	    
	    public String getPaddedBinaryRepresentation() {
	    	return toBinary(prefix.split("/")[0]);
	    }
	    
	    public String getPrefix() {
	    	return prefix;
	    }
	    
	    public Integer getPrefixLength() {
	    	return length;
	    }
	    
	    public String getLabel() {
	    	return label;
	    }
	    
	    public int compare(Subnet s1, Subnet s2) {
    		
    		if ( !s1.getPaddedBinaryRepresentation().equals(s2.getPaddedBinaryRepresentation()) ) {
    			return s1.getPaddedBinaryRepresentation().compareTo(s2.getPaddedBinaryRepresentation());
    		}
    		else {
    			return s1.getPrefixLength().compareTo(s2.getPrefixLength());
    		}
    	}
	    
	    public String toString() {
	    	return getBinaryRepresentation();
	    }
	    
	};
	
	public class Trie<T> {
		 
	    private Node<T> root = new Node<T>('\0', null);

	    public Trie() {}

	    public Trie( List<T> argInitialWords ) {
	            for (T word:argInitialWords) {
	                    addWord(word.toString(), word);
	            }
	    }
	    
	    public void addWord(T value) {
	    	addWord(value.toString(), value);
	    }
	    
	    public void addWord(String argWord, T value) {
	            addWord(argWord.toCharArray(), value);
	    }

	    public void addWord(char[] argWord, T value) {
	            Node<T> currentNode = root;

	            for (int i = 0; i < argWord.length; i++) {
	                    if (!currentNode.containsChildValue(argWord[i])) {
	                    	currentNode.addChild(new Node<T>(argWord[i], value));
	                    }

	                    currentNode = currentNode.getChild(argWord[i]);
	            }
	            
	            currentNode.setValue(value);
	            currentNode.setIsWord(true);
	    }

	    public boolean containsPrefix(T argPrefix) {
	            return contains(argPrefix.toString().toCharArray(), false);
	    }

	    public boolean containsWord(T argWord) {
	            return contains(argWord.toString().toCharArray(), true);
	    }

	    public T getWord(T argString) {
	            Node<T> node = getNode(argString.toString().toCharArray());
	            return node != null && node.isWord() ? node.getValue() : null;
	    }

	    public T getPrefix(T argString) {
	            return getNode(argString.toString().toCharArray()).getValue();
	    }

	    public Set<T> getWordsAlongPrefix(T argString) {
	        return getWords(argString.toString().toCharArray());
	    }
	    
	    private Set<T> getWords(char[] argString) {
	    	
	    	HashSet<T> set = new HashSet<T>();
	        Node<T> currentNode = root;

	        for (int i = 0; i < argString.length && currentNode != null; i++) {
	        	
	                currentNode = currentNode.getChild(argString[i]);                

	                if (currentNode == null) {
	                        return set;
	                }
	                
	                if ( currentNode.isWord() ) {                	                	
	                	set.add(currentNode.getValue());
	                }
	        }
	                
	        return set;
	    }

	    
	    @Override
	    public String toString() {
	            return root.toString();
	    }

	    private boolean contains(char[] argString, boolean argIsWord) {
	            Node<T> node = getNode(argString);
	            return (node != null && node.isWord() && argIsWord) || (!argIsWord && node != null);
	    }

	    private Node<T> getNode(char[] argString) {
	            Node<T> currentNode = root;

	            for (int i = 0; i < argString.length && currentNode != null; i++) {
	                    currentNode = currentNode.getChild(argString[i]);                    

	                    if (currentNode == null) {
	                            return null;
	                    }
	            }
	            
	            return currentNode;
	    }
	    
	    
	    class Node<S> {
	   	 
	        private Character ch;
	        private S value;
	        private Map<String, Node<S>> children = new HashMap<String, Node<S>>();
	        private boolean isValidWord;

	        public Node(char argChar, S argValue) {
	                ch = argChar;
	                value = argValue;
	        }

	        public boolean addChild(Node<S> argChild) {
	        	
	                if (children.containsKey(Character.toString(argChild.getChar()))) {
	                    return false;                    
	                }

	                children.put(Character.toString(argChild.getChar()), argChild);                
	                return true;
	        }

	        public boolean containsChildValue(char c) {
	                return children.containsKey(Character.toString(c));
	        }
	        
	        public void setValue(S value) {
	        	this.value = value;
	        }

	        public S getValue() {
	                return value;
	        }

	        public char getChar() {
	                return ch;
	        }
	        
	        public Node<S> getChild(char c) {        		
	            return children.get(Character.toString(c));
	        }

	        public boolean isWord() {
	                return isValidWord;
	        }

	        public void setIsWord(boolean argIsWord) {
	                isValidWord = argIsWord;
	        }
	        
	        public String toString() {
	        	return value.toString();
	        }

	    }
	}

}
