/**
 * Dasu Project
 *
 * File:         UnlimitedDistributedCoordinator.java
 * RCS:          $Id: ZooKeeperCoordinator.java,v 1.0 Oct 1, 2012 11:14:05 PM mas939 Exp $
 * Description:  UnlimitedDistributedCoordinator class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 1, 2012 at 11:14:05 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.dynamic.coordination
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2012, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.coordination;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONArray;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.dasu.coordination.DistributedCoordinator;
import edu.northwestern.dasu.measurement.AMCoordinatorV2;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.util.Pair;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The UnlimitedDistributedCoordinator class ...
 */
public class ZooKeeperDistributedCoordinator extends DistributedCoordinator {

	
	private Boolean isActive = false;
	public Object halt = new Object();
	private boolean terminate = false;
	long schedulerInterval = 10;
	
	Map<ProbeType, List<Pair<String, Long>>> probed = new Hashtable<ProbeType, List<Pair<String, Long>>>();
	Map<ProbeType, Map<String, List<Long>>> probedSubnet = new Hashtable<ProbeType, Map<String, List<Long>>>();
	
	Map<ProbeType, SubnetManager> subnetManagers = new HashMap<ProbeType, SubnetManager>();
	ExperimentLease lease;
	CuratorConnectionNonPersistent zookeeper;
		
	public ZooKeeperDistributedCoordinator(SourceType engineId,
			String namePrefix, JSONObject distributedCoordinatorParams) throws Exception {
		
		super(engineId, namePrefix, 2.0f);
		isActive = true;
				
		Map<String, Object> envInfo = AMCoordinatorV2.getInstance().getEnvInfo();
		String uuid = envInfo.get("user.uuid").toString();
		String cc = envInfo.get("net.country").toString();
		String asn = envInfo.get("net.asn").toString();
		String myIp = envInfo.get("net.ip.public").toString();
		/*
		String uuid = "1a94acea-42aa-4781-a00b-0d1873d1d9b1";
		String myIp = "10.0.0.1";
		String cc = "DO";
		String asn = "15169";
		*/
		
		String servers = distributedCoordinatorParams.get("serverList").toString();
		String taskId = distributedCoordinatorParams.get("taskId").toString();		
		
		//throws exception if can't connect to the Coordination server
		zookeeper = new CuratorConnectionNonPersistent(servers, taskId, uuid, myIp, cc, asn);
		
		updateLease();
		
		if ( lease == null )
			throw new Exception("no valid lease...");
		
	}

	//an limited distributed coordinator
	public void run() {
				
		while (isActive) {
		
			synchronized (halt) {
	
				schedulerInterval = lease.getReportInterval();
				
				if (lease.isExpired()) {
					try {
						updateLease();
					}
					catch (Exception e) {
						e.printStackTrace();
					}
				}
				
				try {		
					LOGGER.info("COORDINATOR: wating for " + schedulerInterval);
					halt.wait(schedulerInterval*1000);					
					reportProbed();
					
					if (System.currentTimeMillis() - lease.getExpiration() > 30*1000) {
						try {
							zookeeper.updateMockLease();
						}
						catch (Exception e) {
							e.printStackTrace();
						}
					}
					
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			
			if (terminate) {		
				return;
			}
		}
	}

	@Override
	public boolean schedulable(ProbeSubTask pst) {
		
		ProbeType pt = pst.getProbeType();
		
		// if no lease, or lease expired or probetype not present in lease or 
		// last connection to zookeeper failed, reject probe
		if ( lease == null || lease.isExpired() || zookeeper.isConnectionFailed() ||
				!subnetManagers.containsKey(pt) ) {
			LOGGER.info("COORDINATION: not schedulable " + (lease == null) + " " + lease.isExpired() + " " + 
					zookeeper.isConnectionFailed() + " " + !subnetManagers.containsKey(pt));
			return false;
		}
		
		// null should not be returned, the default subnet should be always present
		String subnet = subnetManagers.get(pt).getLongestPrefixMatch( pst.getToProbe() );
		if ( subnet == null ) { 
			LOGGER.warning("COORINATION: subnetManager returned null for destination " + pst.getToProbe());
			return false;
		}
		
		//numProbes, pctLeewayAboveLimit
		Pair<Long, Float> limits = lease.getBudget().getLimits(pt, subnet);
		lease.getBudget().getTimeIntervalForBudgetLimits();
		
		
		Long now = System.currentTimeMillis();
		
		synchronized(probedSubnet) {
			
			if ( !probedSubnet.containsKey(pt) ) 
				probedSubnet.put(pt, new Hashtable<String, List<Long>>());				
			if ( !probedSubnet.get(pt).containsKey(subnet) )
				probedSubnet.get(pt).put(subnet, new ArrayList<Long>());
			
			if (!schedulable(pt, subnet, now)) {
				LOGGER.info("COORDINATION: not schedulable, exceeds maximun " +
						pt + " " + pst.getToProbe() + " "  + subnet + " " + 
						probedSubnet.get(pt).get(subnet).size() + " " + 
						lease.getBudget().getMaxNumberOfProbes(pt, subnet) + " " +
						lease.getBudget().getPercentLeewayAboveLimit(pt, subnet));
				return false;
			}
		}
				
		
		// needs to be synchronized because the run method also accesses this map
		synchronized(probed) {
			if ( !probed.containsKey(pt) )
				probed.put(pt, new ArrayList<Pair<String, Long>>());
			
			probed.get(pt).add(new Pair<String, Long>(pst.getToProbe(), now));
		}
				
		probedSubnet.get(pt).get(subnet).add(now);		
		
		return true;
	}

	@Override
	public void stopCoordinator() {
		isActive = false;
		synchronized(halt) {
			terminate = true;
			halt.notify();
		}		
	}
	
	private boolean schedulable(ProbeType pt, String subnet, Long now) {
		
		while (probedSubnet.get(pt).get(subnet).size() > 0
					&& now - probedSubnet.get(pt).get(subnet).get(0) > 
					1000 * lease.getBudget().getTimeIntervalForBudgetLimits()) {
				probedSubnet.get(pt).get(subnet).remove(0);
		}
		
		return probedSubnet.get(pt).get(subnet).size() < lease.getBudget().getMaxNumberOfProbes(pt, subnet) + 
				lease.getBudget().getMaxNumberOfProbes(pt, subnet) * lease.getBudget().getPercentLeewayAboveLimit(pt, subnet);

	}
	
	/*
	 {  'version': 1.0,
	 	'leaseExpiration': date,
	 	'leaseStart': date,
		'reportInterval': seconds,    	
		'budget': {
					'timeIntervalForBudgetLimits': minutes,
					'probeTypes: {
							'PING': [ ('0.0.0.0/0', 30, 0.1), ('207.110.223.0/19', 30, 0.2).... ],
							'TRACEROUTE': [ ('0.0.0.0/0',  -1, 1), (192.168.0.1/24', 22, 0.1)....],
							....
							}
					}
	 }
	*/
	
	private void reportProbed() {
		
		HashMap<ProbeType, List<String>> toReport = new HashMap<ProbeType, List<String>>();
		HashMap<ProbeType, List<Pair<String, Long>>> removed = new HashMap<ProbeType, List<Pair<String, Long>>>();
		Integer ipsToReport = 0;
		
		synchronized(probed) {
						
			LOGGER.info("CORDINATION: reporting " + probed.size());
			
			synchronized(lease) {
				
				for ( ProbeType pt : probed.keySet() ) {
			
					removed.put(pt, new ArrayList<Pair<String, Long>>());
					ArrayList <String> listOfIps = new ArrayList<String>();
					Iterator<Pair<String, Long>> iter = probed.get(pt).iterator();					
					while ( iter.hasNext() ) {
						Pair<String, Long> pair = iter.next();
						// report all probes older than the specified report interval by the lease
						if ( (System.currentTimeMillis() - pair.getValue())/1000 > lease.getReportInterval() ) {
							iter.remove();
							listOfIps.add(pair.getKey());
							removed.get(pt).add(pair);
							ipsToReport += 1;
						}
					}
					
					toReport.put(pt, listOfIps);
				}
			}
		}
		
		if ( ipsToReport == 0) {
			LOGGER.info("COORDINATION: nothing to report");
			return;
		}
			
		
		// reporting to zookeeper failed, put back
		if (!zookeeper.reportProbedIPs(toReport)) {
			
			LOGGER.warning("COORDINATION: reporting to zookeeper failed, putting back IPs to report");
			synchronized(probed) {
				for ( ProbeType pt : removed.keySet()) {
					probed.get(pt).addAll( removed.get(pt) );
				}
			}
		}
		
	}
	
	private void updateSubnetManager() {
		
		subnetManagers.clear();
		for ( ProbeType pt : lease.getBudget().probeTypeBudget.keySet()) {
			
			subnetManagers.put(pt, new SubnetManager());
			for ( String prefix : lease.getBudget().probeTypeBudget.get(pt).keySet() ) {
				subnetManagers.get(pt).addSubnet(prefix);
			}
		}
	}
	
	private void updateLease() throws Exception {
		
		JSONObject jsonLease = zookeeper.PullUpdatedLease();
		if ( jsonLease == null ) {
			LOGGER.info("COORDINATOR: Lease hasn't been updated");
			return;
		}
		
		LOGGER.info(jsonLease.toJSONString());
		Double version = (Double) jsonLease.get("version");
		Long expiration = (Long) jsonLease.get("leaseExpiration");
		Long leaseStart = (Long) jsonLease.get("leaseStart");
		Long reportInterval = (Long) jsonLease.get("reportInterval");
		
		JSONObject budget = (JSONObject) jsonLease.get("budget");				
		Long timeIntervalForBudgetLimits = (Long) budget.get("timeIntervalForBudgetLimits");
		JSONObject probeTypes = (JSONObject) budget.get("probeTypes");
		
		if ( lease == null)
			lease = new ExperimentLease(version.floatValue(), leaseStart, expiration, reportInterval, timeIntervalForBudgetLimits);
		
		synchronized(lease) {
			
			lease = new ExperimentLease(version.floatValue(), leaseStart, expiration, reportInterval, timeIntervalForBudgetLimits);
			
			for ( String probeType : probeTypes.keySet() ) {
				
				for ( Object obj : ((JSONArray) probeTypes.get(probeType)) ) {
					
					JSONArray array = (JSONArray) obj;
					String prefix = (String) array.get(0);
					Long numberOfProbes = (Long) array.get(1);
					Double percentLeewayAboveLimit = (Double) array.get(2);
				
					ProbeType pt = ProbeType.valueOf(probeType);
					lease.getBudget().appendProbeTypeLimit(pt, prefix, numberOfProbes, percentLeewayAboveLimit.floatValue());
				}
			}

			LOGGER.info(lease.toString());	
			updateSubnetManager();

		}
		
	}

	@Override
	public boolean isFailed() {
		
		if ( zookeeper != null && (zookeeper.getConnectionFailedSince() == -1 || 
				System.currentTimeMillis() - zookeeper.getConnectionFailedSince() < 5*60*1000)) {
			return false;
		}
		
		return true;
		
	}

}
