/**
 * Dasu Project
 *
 * File:         DasuHttpModule.java
 * RCS:          $Id$
 * Description:  DasuHttpModule class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 15, 2010 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.measurement;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.reflect.Field;
import java.net.HttpURLConnection;
import java.net.Socket;
import java.net.URL;
import java.net.URLConnection;
import java.security.cert.X509Certificate;
import java.util.Map;
import java.util.Map.Entry;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.X509TrustManager;

import sun.net.NetworkClient;
import sun.net.www.http.HttpClient;
import edu.northwestern.dasu.dynamic.drools.FactHttpProbeResult;
import edu.northwestern.dasu.measurement.DasuProbeModule;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.util.Util;


/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuHttpModule class manages the set of 
 *         HTT gets performed by the peer.
 */

public class DasuHttpModule extends DasuProbeModule<FactHttpProbeResult> {
		
	private static DasuHttpModule self;
	private final int chunkSize=1024;

	public DasuHttpModule() {
		super("DasuHttpModule", 3.0f);
	}
	
	public synchronized static DasuHttpModule getInstance() {
		if (self == null) {
			self = new DasuHttpModule();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}
	
	public boolean isIdle() {
		return idle;
	}
				
	public FactHttpProbeResult fileUrl(String fAddress, String localFileName, Integer taskId, String target, Map<String, String> headers) {

		FactHttpProbeResult ret = new FactHttpProbeResult(fAddress);		
		ret.setTarget(target);
		URLConnection  uCon = null;
		Long startTime = 0L;
		Long endTime = 0L;
		
		InputStream is = null;
		try {
			URL Url;
			byte[] buf;
			int ByteRead,ByteWritten=0;
			Url= new URL(fAddress);

			uCon = Url.openConnection();
			
			if (headers != null) {
				
				for ( Entry<String, String> e : headers.entrySet() ) {
					uCon.setRequestProperty(e.getKey(), e.getValue());
				}
				
				uCon.connect();
			}
			
			startTime = Util.currentTimeMillis();
			is = uCon.getInputStream();
			buf = new byte[chunkSize];
			while ((ByteRead = is.read(buf)) != -1) {
				//outStream.write(buf, 0, ByteRead);
				ByteWritten += ByteRead;
			}
			endTime = Util.currentTimeMillis();

			float dldTime = (endTime - startTime)/1000F;
			float dldRate = (((float) ByteWritten)*8)/dldTime/1024;
			
			ret = new FactHttpProbeResult(fAddress, dldTime, dldRate, ByteWritten, taskId);
			ret.setTarget(target);
			
			reportProbeResult(ret);

			//System.out.println("DEBUG: DasuHttpModule: " + ret.toString());
			
			LOGGER.info("HTTP: Downloaded Successfully.");
			LOGGER.info("HTTP: File name:\""+localFileName+ "\"\nNo ofbytes: " + ByteWritten);
			LOGGER.info("HTTP: File Size: " + ((float) ByteWritten)/1024F/1024F + " MB");
			LOGGER.info("HTTP: Dld Time: " + (endTime - startTime)/1000F + " seconds");
			Float dr = (((float) ByteWritten)*8)/((endTime - startTime)/1000);
			LOGGER.info("HTTP: Dld Speed: " +  dr/1024 + " kbps");
			
		} catch (Exception e) {
			LOGGER.info("HTTP: Timeout: " + fAddress);
			return ret;
			
		} finally {
			try {
				
				if (is != null)
					is.close();
				//outStream.close();
				
			} catch (IOException e) {
				LOGGER.info("HTTP: File not found!: " + localFileName);
				return ret;
			}
		}
		return ret;

	}

	/*default probe uses default headers server*/
	
	@SuppressWarnings("unchecked")
	protected FactHttpProbeResult probeIp(ProbeSubTask pst) {
	
		if ( !isActive() )
			return null;
		
		Map<String, String> headers = null;
		if(pst.getParams() != null && pst.getParams().containsKey("headers")) {
			headers = (Map<String, String>) pst.getParams().get("headers");
		}
		
		FactHttpProbeResult ret = new FactHttpProbeResult(pst.getToProbe());
		ret.setTarget(pst.getTarget());
		int slashIndex =pst.getToProbe().lastIndexOf('/');
		int periodIndex =pst.getToProbe().lastIndexOf('.');
	
		String fileName=pst.getToProbe().substring(slashIndex + 1);
	
		if (periodIndex >=1 &&  slashIndex >= 0 && slashIndex < pst.getToProbe().length()-1) {
			ret = fileUrl(pst.getToProbe(), fileName, pst.getTaskId(), pst.getTarget(), headers);
		}
		else {
			LOGGER.warning("ERROR: path or file name");
		}

		FactHttpProbeResult result = (FactHttpProbeResult) ret;
		result.setPst(pst);
		return result;
	}
			
	public String getData(String urls, boolean setCookie) {
        
		SSLSocketFactory factory = HttpsURLConnection.getDefaultSSLSocketFactory();
		try {
			URL url = new URL(urls);
			HttpURLConnection conn = null;
			
			if ( urls.contains("https") ) {
				
				TrustManager trm = new X509TrustManager() {
		            public X509Certificate[] getAcceptedIssuers() {
		                return null;
		            }

		            public void checkClientTrusted(X509Certificate[] certs, String authType) {

		            }

		            public void checkServerTrusted(X509Certificate[] certs, String authType) {
		            }
		        };
		       
		        SSLContext sc = SSLContext.getInstance("SSL");
		        sc.init(null, new TrustManager[] { trm }, null);
		        HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());		        

		    	conn = (HttpsURLConnection)url.openConnection();	    	

			}
			else
				conn = (HttpURLConnection)url.openConnection();
	    	
			if ( setCookie ) {
		    	String cookie = conn.getHeaderField("Set-Cookie");
		    	conn = urls.contains("https") ? (HttpsURLConnection) url.openConnection() : 
		    		(HttpURLConnection)url.openConnection();
		    	
		    	conn.setRequestProperty("Cookie", cookie);
			}
	    	    	
	    	conn.connect();
	    	InputStream inputstream = conn.getInputStream();
	    	InputStreamReader inputstreamreader = new InputStreamReader(inputstream);
	    	BufferedReader bufferedreader = new BufferedReader(inputstreamreader);
	
	    	String string = null;
	    	String res = new String();
	    	while ((string = bufferedreader.readLine()) != null) {	    	    
	    	    res = res.concat(string);
	    	}
	    	
	    	//System.out.println(res);
	    	return res;
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		finally {
			HttpsURLConnection.setDefaultSSLSocketFactory(factory);
		}
		
		return null;
    	
	}	
	
	public static void main(String[] args) {
		
//		System.out.println(new Date());
//		DasuHttpModule.getInstance().getData("http://trestian.cs.northwestern.edu:25255", false);
//		System.out.println(new Date());
//		System.exit(0);
		
		try {
			
			HttpURLConnection conn = (HttpURLConnection) new URL("http://a57.foxnews.com/www.foxnews.com/ucat/images/121/91/293514_Henderson_Alvarez.jpg").openConnection();
			//HttpURLConnection conn = (HttpURLConnection) new URL("http://23.72.67.153/www.foxnews.com/ucat/images/121/91/293514_Henderson_Alvarez.jpg").openConnection();
			//conn.setRequestProperty("Host", "a57.foxnews.com");
			
			conn.connect();
			Field field = conn.getClass().getDeclaredField("http");
			field.setAccessible(true);
			HttpClient http = (HttpClient) field.get(conn);
			field = NetworkClient.class.getDeclaredField("serverSocket");
			field.setAccessible(true);
			Socket socket = (Socket) field.get(http);
			System.out.println(socket.getInetAddress().getHostAddress());
			
	    	InputStream inputstream = conn.getInputStream();
	    	InputStreamReader inputstreamreader = new InputStreamReader(inputstream);
	    	BufferedReader bufferedreader = new BufferedReader(inputstreamreader);
	
	    	String string = null;
	    	String res = new String();
	    	while ((string = bufferedreader.readLine()) != null) {	    	    
	    	    res = res.concat(string);
	    	}
	    	
	    	System.out.println(res.length());
			
		}
		catch (Exception e) {
			e.printStackTrace();
		}

	}
	
	@Override
	public Integer getNumberActiveProbes() {
		
		if (!isIdle())
			return 1;
		
		return 0;
		
	}
	
	
}
