/**
 * Dasu Project
 *
 * File:         DasuHttpModule.java
 * RCS:          $Id$
 * Description:  DasuHttpModule class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 15, 2010 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.measurement;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashMap;

import edu.northwestern.dasu.dynamic.drools.FactHttpExtendedProbeResult;
import edu.northwestern.dasu.measurement.DasuProbeModule;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.ProbeType;


/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuHttpModule class manages the set of 
 *         HTT gets performed by the peer.
 */

public class DasuHttpModuleExtended extends DasuProbeModule<FactHttpExtendedProbeResult> {
		
	private static DasuHttpModuleExtended self;	

	public DasuHttpModuleExtended() {
		super("DasuHttpModuleExtended", 1.0f);
	}
	
	public synchronized static DasuHttpModuleExtended getInstance() {
		if (self == null) {
			self = new DasuHttpModuleExtended();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}
	
	public boolean isIdle() {
		return idle;
	}

	/*default probe uses default headers server*/
	
	protected FactHttpExtendedProbeResult probeIp(ProbeSubTask pst) {
	
		if ( !isActive() ) {
			return null;
		}

		Integer port = 80;		
		String userAgent = "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US; rv:1.9.2.8) Gecko/20100806 Firefox/3.6";		
		String hostname = pst.getToProbe().replace("http://",  "").replace("https://", "").split("/")[0];
		String obj = "";
		String[] parts = pst.getToProbe().replace("http://",  "").replace("https://", "").split("/");
		for ( int i=1; i < parts.length; i++) {			
			obj = obj.concat("/" + parts[i]);
		}
		String ip = hostname;
		
		// if ip specified, connect to ip instead of resolving the hostname
		if ( pst.getParams().containsKey("ip") ) {
			ip = (String) pst.getParams().get("ip");
		}
		if ( pst.getParams().containsKey("port") ) {
			port = (Integer) pst.getParams().get("port");
		}
		if ( pst.getParams().containsKey("hostname") ) {
			hostname = (String) pst.getParams().get("hostname");
		}
		if ( pst.getParams().containsKey("obj") ) {
			obj = (String) pst.getParams().get("obj");
		}
		if ( pst.getParams().containsKey("userAgent") ) {
			userAgent = (String) pst.getParams().get("userAgent");
		}
		
		String header;
				
		//create default header
		StringBuilder sb = new StringBuilder();		
		sb.append("GET " + obj + " HTTP/1.1\r\n");
		sb.append("User-Agent: " + userAgent + "\r\n");
		sb.append("Host: " + hostname + "\r\n");
		sb.append("Connection: close\r\n");
		sb.append("Pragma: no-cache\r\n");
		sb.append("Cache-Control: no-cache\r\n");
		sb.append("\r\n");		
		header = sb.toString();
		
		// if a specific header was supplied, use that one instead
		if ( pst.getParams().containsKey("header") ) {
			header = (String) pst.getParams().get("header");
		}

		FactHttpExtendedProbeResult ret = new FactHttpExtendedProbeResult(hostname);
		ret.setTarget(pst.getTarget());
		ret.setIp(ip);
		ret.setHost(hostname);
		ret.setObj(obj); 
		
		// We'll read the data as fast as possible into the ByteArray.
		// We'll store the times that each byte offset is reached, so we can figure out the timing for when things happen. 
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ArrayList<Integer> byteOffsets = new ArrayList<Integer>();
		ArrayList<Long> times = new ArrayList<Long>();
		int chunkSize = 2048;
		byte[] buf = new byte[chunkSize];
		int totalBytes = 0;
		int bytes = 0;
		long currentTime = 0;
		// did this now so we affect timing as little as possible, later on.
		
		long initTime = System.currentTimeMillis();
		
		try {
			
			Socket sock = new Socket(ip, port);
			long connected = System.currentTimeMillis();
			PrintWriter pw = new PrintWriter(sock.getOutputStream());
			pw.write(header.toString());
			pw.flush();
			long headersSent = System.currentTimeMillis();
			
			// Headers are sent. Now, read the response. Tricky because it's in text, then binary.
			
			InputStream inst = sock.getInputStream();
			while ((bytes = inst.read(buf)) != -1) {
				currentTime = System.currentTimeMillis();
				totalBytes += bytes;
				
				baos.write(buf, 0, bytes);
				byteOffsets.add(totalBytes);
				times.add(currentTime);
			}
			long doneTime = System.currentTimeMillis();
			sock.close();
			
			// We have the whole object in memory. (Hope it's not too big! :p )
			byte[] wholeThing = baos.toByteArray();
			
			// Find the end of the headers -- the offset of "\r\n\r\n" in the byte array
			boolean found = false;
			int offset = 0;
			for(int i = 0; i < wholeThing.length - 4; i++) {
				if(wholeThing[i] == 13 && wholeThing[i+1] == 10 && wholeThing[i+2] == 13 && wholeThing[i+3] == 10) {
					found = true;
					offset = i; // offset is the index of the first byte after the last header line's content 
					break;
				}
			}

			ret.setTotalTime(doneTime - initTime);
			ret.setConnect(connected - initTime);
			ret.setSendHead(headersSent - connected);
			
			if(!found) {
				ret.setTime(initTime);
				ret.setError("Parse Error: HTTP header not found");
				ret.setFailed(true);
				return ret;
			}
			
			ByteArrayOutputStream headerByteArr = new ByteArrayOutputStream();
			headerByteArr.write(wholeThing, 0, offset);
			LOGGER.info("HTTP2: header content:\n" + headerByteArr.toString() + "\n\nend header content");
			String[] headerLines = headerByteArr.toString().split("\\r\\n");
			
			String[] resp = headerLines[0].split("\\s+");
			int status = Integer.parseInt(resp[1]);
			
			HashMap<String, String> headers = new HashMap<String, String>();
			for(int i = 1; i < headerLines.length; i++) {
				String line = headerLines[i].trim();				
				int index = line.indexOf(":");
				if(index == -1) continue;
				
				String key = line.substring(0, index);
				String val = line.substring(index + 1).trim();
				headers.put(key, val);
			}
			
			// We have the response code and headers from the server.
			ret.setHeaders(headers);
			ret.setStatus(status);
			
			// Compute when everything arrived...
			int endOfHeaders = offset + 3;
			long eohTime = doneTime;
			for(int i = 0; i < times.size(); i++) {
				int byteOffset = ((Integer) byteOffsets.get(i)).intValue();
				long time = ((Long) times.get(i)).longValue();
				if(byteOffset >= endOfHeaders) {
					eohTime = time;
					break;
				}
			}
			
			ret.setRecvHead(eohTime - headersSent);
			
			int startOfContent = offset + 4;
			long socTime = doneTime;
			for(int i = 0; i < times.size(); i++) {
				int byteOffset = ((Integer) byteOffsets.get(i)).intValue();
				long time = ((Long) times.get(i)).longValue();
				if(byteOffset >= startOfContent) {
					socTime = time;
					break;
				}
			}
			
			ret.setFirstByte(socTime - eohTime);
			ret.setRestOfBytes(doneTime - socTime);
			ret.setObjBytes(wholeThing.length - (offset + 3));
			ret.setHeadBytes(offset + 3);
			
		} catch(Exception e) {
			ret.setError(e.toString());
			ret.setFailed(true);
		}
		
		ret.setTime(initTime);
		
		return ret;
	}	
	
	public static void main(String[] args) {
		
//		System.out.println(new Date());
//		DasuHttpModule.getInstance().getData("http://trestian.cs.northwestern.edu:25255", false);
//		System.out.println(new Date());
//		System.exit(0);
				
	    String obj = "http://a57.foxnews.com/www.foxnews.com/ucat/images/121/91/293514_Henderson_Alvarez.jpg"; // 1.6k
	    
	    int taskId = Integer.parseInt("9920231");
	    int priority = -1; 
	    long startTime = -1; 
	    long endTime = -1; 
	    ArrayList<String> list = new ArrayList<String>();
	    list.add(obj);
	    HashMap<String, String> params = new HashMap<String, String>();
	    params.put("ip", "23.72.67.153");
	    ProbeTask task = new ProbeTask(taskId, priority, startTime, endTime, list, ProbeType.HTTP2);
	    task.setParams(params);

	    for ( ProbeSubTask pst : task.getSubTasks()) {
	    	System.out.println(DasuHttpModuleExtended.getInstance().probeIp(pst).toJSON() );
	    }

	}
	
	@Override
	public Integer getNumberActiveProbes() {
		
		if (!isIdle())
			return 1;
		
		return 0;
		
	}
	
	
}
