/**
 * Dasu Project
 *
 * File:         DasuPingModuleV2.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 15, 2010 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.measurement;


import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.dynamic.drools.FactPingProbeResult;
import edu.northwestern.dasu.measurement.DasuProbeModule;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.profilers.ResourceProfiler;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.PingManager;
import edu.northwestern.dasu.util.Util.PingResponse;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuPingModuleV2 class manages the set of 
 *         Pings performed by the peer.
 */

public class DasuPingModuleV2 extends DasuProbeModule<FactPingProbeResult> {	
		
	private static DasuPingModuleV2 self;
	private Integer maxConcurrentRuns = 5;

	public DasuPingModuleV2() {
		super("DasuPingModule", 2.0f);
	}
	
	public synchronized static DasuPingModuleV2 getInstance() {
		if (self == null) {
			self = new DasuPingModuleV2();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}
	
	public boolean isPaused() {
		return pause && PingManager.getInstance().activePings < maxConcurrentRuns;
	}

	public boolean isIdle() {
		return PingManager.getInstance().activePings < maxConcurrentRuns;
	}
	
	public Integer getNumberActiveProbes() {
		return PingManager.getInstance().activePings;		
	}
	
	@Override
	public void run() {
		
	  while (!Main.isShuttingDown() && isActive()) {
		 
		  	ProbeSubTask pst = null;
			synchronized (toDoTaskList) {
				while ((toDoTaskList.size()==0 || isPause() || PingManager.getInstance().activePings >= maxConcurrentRuns) 
						&& !Main.isShuttingDown() && isActive()) {
					try {
						toDoTaskList.wait(sleepInterval*1000);
						
					} catch (InterruptedException e) {
						setActive(false);
						LOGGER.info("STOP: Adios...");
						return;
					}
				}
				
				if ( toDoTaskList.size() > 0 ) {
					pst = toDoTaskList.remove(0);
					LOGGER.info("Ping server " + pst.getIpAddress());
				}
			}
						
			if ( pst != null ) {
				
				pst.recordExecutionPathTimestamp("probeLaunchTime", System.currentTimeMillis());
				pst.recordExecutionPathTimestamp("cpuProbeModuleUsageAtLaunchTime", DasuManager.getResourceProfiler().getCpuUsage(Thread.currentThread().getId()));
				pst.recordExecutionPathTimestamp("cpuDasuAtLaunchTime", DasuManager.getResourceProfiler().getDasuCpuUsage());
				pst.recordExecutionPathTimestamp("cpuVuzeAtLaunchTime", DasuManager.getResourceProfiler().getJvmCpuUsage());
				pst.recordExecutionPathTimestamp("probeQueueSizeLaunchTime", this.toDoTaskList.size());	
				if ( DasuManager.systemFact.entryList.size() > 1 ) {
					pst.recordExecutionPathTimestamp("shortDetlaAvgDownloadRateSystemLaunchTime", DasuManager.systemFact.getShortDeltaAverageDownloadRate());				
					pst.recordExecutionPathTimestamp("shortDetlaAvgDownloadRateNetworkLaunchTime", DasuManager.localNetworkFact.getAverageTotalBytesReceived());
					pst.recordExecutionPathTimestamp("shortDetlaAvgDownloadRateLpsdLaunchTime", DasuManager.localPeerFact.getShortDeltaAverageDownloadRate());
				}
				
				probeIp(pst);
				
				if ( yieldInterval != -1) {
					try {
						sleep(yieldInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}


	  } // end while active
		
	  LOGGER.info("STOP: Adios...");
	}
		
	@Override
	protected FactPingProbeResult probeIp(final ProbeSubTask pst) {

		if ( isActive() ) {		
		
			PingManager.getInstance().doPing(pst.getIpAddress(), 
					pi, new PingResponse() {
	
						public void response(double rtt) {
							
							pst.recordExecutionPathTimestamp("probeCompletionTime", System.currentTimeMillis());							
							LOGGER.info("REPORTER: " + pst.getExecutionPathTimestamps());
							
							FactPingProbeResult fppr = new FactPingProbeResult(pst.getIpAddress(), rtt);
							fppr.setTarget(pst.getTarget());
							fppr.setPst(pst);
	
							if (rtt >= 0) 
								reportProbeResult(fppr);
							
							updatePingResultQueue(fppr);
						}				
					}
			);
		}
		
		return null;

	}

	private void updatePingResultQueue(FactPingProbeResult fppr) {

		resultTable.add(fppr);
		updateResultQueue(fppr, fppr.getProbeType());
		
	}	
	
//	public Pair<String, FactTemplate> proccessResult(long eventId) {
//		
//		FactPingProbeResult pingResult = resultTable.remove(eventId);		
//		return new Pair<String, FactTemplate>(pingResult.getIp(), pingResult);
//		
//	}

	
}
