/**
 * Dasu Project
 *
 * File:         DasuHttpModule.java
 * RCS:          $Id$
 * Description:  DasuHttpModule class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 15, 2010 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.ono.experiment
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.dynamic.measurement;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;

import org.gudy.azureus2.core3.torrent.TOTorrent;
import org.gudy.azureus2.core3.torrent.TOTorrentException;
import org.gudy.azureus2.core3.torrent.TOTorrentFactory;
import org.gudy.azureus2.core3.tracker.client.TRTrackerAnnouncerFactory;
import org.gudy.azureus2.core3.tracker.client.TRTrackerAnnouncer;
import org.gudy.azureus2.core3.tracker.client.TRTrackerAnnouncerDataProvider;
import org.gudy.azureus2.core3.tracker.client.TRTrackerAnnouncerException;
import org.gudy.azureus2.core3.tracker.client.TRTrackerAnnouncerListener;
import org.gudy.azureus2.core3.tracker.client.TRTrackerAnnouncerResponse;
import org.gudy.azureus2.core3.tracker.client.TRTrackerAnnouncerResponsePeer;
import org.gudy.azureus2.core3.tracker.client.TRTrackerScraper;
import org.gudy.azureus2.core3.tracker.client.TRTrackerScraperResponse;
import org.gudy.azureus2.core3.util.SystemTime;

import com.aelitis.azureus.plugins.dht.DHTPlugin;
import com.aelitis.azureus.plugins.dht.DHTPluginContact;
import com.aelitis.azureus.plugins.dht.DHTPluginOperationListener;
import com.aelitis.azureus.plugins.dht.DHTPluginValue;

import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.dynamic.drools.FactScrapeProbeResult;
import edu.northwestern.dasu.measurement.DasuProbeModule;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuScrapeModule class manages the set of 
 *         Scrapes performed by the peer.
 */

public class DasuScrapeModule extends DasuProbeModule<FactScrapeProbeResult> {	
		
	private static DasuScrapeModule self;
	private HashSet<String> validTypes = 
		new HashSet<String>(Arrays.asList(new String[]{"DHT", "TRACKER_SCRAPE", "TRACKER_ANNOUNCE"}));
	TRTrackerAnnouncer trackerClient;
	
	private static int MIN_INTERVAL = 10; //minimun interval allowed between scrapes to the same server (minutes)
	public Hashtable<String, Long> trackerHistory = new Hashtable<String, Long>();

	
	public DasuScrapeModule() {
		super("DasuScrapeModule", 2.0f);
	}
	
	public synchronized static DasuScrapeModule getInstance() {
		if (self == null) {
			self = new DasuScrapeModule();
			self.setDaemon(true);
		}
		
		return self;
		
	}
	
	public boolean isIdle() {
		return idle;
	}

	protected FactScrapeProbeResult probeIp(ProbeSubTask pst) {
		
		if ( !isActive() )
			return null;
		
		//remove probe history of any tracker over 10 minutes old
		synchronized(trackerHistory) {
			
			Iterator<Entry<String, Long>> iter = trackerHistory.entrySet().iterator();
			while (iter.hasNext()) {
				Entry<String, Long> entry = iter.next();
				if ((Util.currentTimeMillis() - entry.getValue())/1000/60 > MIN_INTERVAL)
					iter.remove();
			}
							
		}
		
		String trackerUrl = "";
		if(pst.getParams() != null && pst.getParams().containsKey("trackerUrl")) trackerUrl = (String) pst.getParams().get("trackerUrl");
		String type = "DHT";
		if(pst.getParams() != null && pst.getParams().containsKey("type")) type = (String) pst.getParams().get("type");
		
		String hash = pst.getIpAddress();
		
		if (trackerHistory.containsKey(trackerUrl) && (Util.currentTimeMillis() - trackerHistory.get(trackerUrl))/1000/60 < MIN_INTERVAL) {
			
				LOGGER.info("new probe to same trackers must be " + MIN_INTERVAL + " minutes appart");
				
				FactScrapeProbeResult fspr = new FactScrapeProbeResult(hash, type, pst.getTaskId(), trackerUrl);
				fspr.setTarget(pst.getTarget());
				fspr.setPst(pst);
				return fspr;
			
		}
				
		else if (hash.length() == 40) {
			
			if (type.equals("DHT")) {
				return dhtScrape(pst);
			}
			else if (type.equals("TRACKER_SCRAPE"))
				return trackerScrape(pst);
			else if (type.equals("TRACKER_ANNOUNCE"))
				return trackerAnnounce(pst);
			
		}
		else {
			LOGGER.info(pst.getParent().probeType +": DasuScrapeModule: ERROR: invalid hash");
		}

		return null;
	}
	
	protected void afterUpdateResultQueue(FactScrapeProbeResult result) {

		if (result.type.equals("TRACKER_ANNOUNCE"))
			trackerClient.destroy();
		
		if (!result.isError()) {

			synchronized(trackerHistory) {
				trackerHistory.put(result.tracker, Util.currentTimeMillis());
			}
			
		}
	}
	
	
	private static volatile FactScrapeProbeResult result = null;
	
	protected FactScrapeProbeResult dhtScrape(final ProbeSubTask pst) {
		
		final String hexHash = pst.getIpAddress();				
		String trackerUrlTmp = "";
		if(pst.getParams() != null && pst.getParams().containsKey("trackerUrl")) 
			trackerUrlTmp = (String) pst.getParams().get("trackerUrl");
		
		final Integer taskId = pst.getTaskId();
		final String trackerUrl = trackerUrlTmp;
		final String target = pst.getTarget();
		
		DHTPlugin dht;
		final org.gudy.azureus2.plugins.PluginInterface dht_pi = 
			Main.getPluginInterface().getPluginManager().getPluginInterfaceByClass(DHTPlugin.class);
				
		if ( dht_pi != null ){			
			dht = (DHTPlugin)dht_pi.getPlugin();
					
			LOGGER.info(pst.getParent().probeType + ": Tracker announce for 'torrent'" );
			dht.get(hexStringToByteArray(hexHash),
					"Tracker announce for torrent",
					DHTPlugin.FLAG_DOWNLOADING, 30,
					2*60*1000, false, false, 
					new DHTPluginOperationListener(){

						List<String>	addresses 	= new ArrayList<String>();
						List<Integer>	ports		= new ArrayList<Integer>();
						List<Integer>	udp_ports	= new ArrayList<Integer>();
						List<Boolean>	is_seeds	= new ArrayList<Boolean>();
						List<String>	flags		= new ArrayList<String>();
						
						int		seed_count;
						int		leecher_count;

						boolean	complete;
						
						public void complete(byte[] key,
								boolean timeout_occurred) {
							synchronized( this ){
								
								if ( complete ){									
									return;
								}
								
								complete = true;
							}
							
							LOGGER.info(pst.getParent().probeType + ": addresses=" + addresses.size() + ", seeds=" + seed_count + ", leechers=" + leecher_count);
							for (String add : addresses) {
								LOGGER.info(pst.getParent().probeType + ": " + add);
							}
							
							FactScrapeProbeResult ret = new FactScrapeProbeResult(hexHash,
									"DHT", 60, taskId, addresses, ports, udp_ports, is_seeds, 
									seed_count, leecher_count, trackerUrl);
							ret.setTarget(target);
							ret.setPst(pst);
							result = ret;
							
						}

						public boolean diversified() {
							return complete;
						}

						public void starts(byte[] key) {
						}

						public void valueRead(DHTPluginContact originator,
								DHTPluginValue value) {

							synchronized( this ){
								
								if ( complete ){
									
									return;
								}

								try{										
									String[]	tokens = new String(value.getValue()).split(";");
								
									String	tcp_part = tokens[0].trim();
									
									int	sep = tcp_part.indexOf(':');
									
									String	ip_str		= null;
									String	tcp_port_str;
									
									if ( sep == -1 ){
										
										tcp_port_str = tcp_part;
										
									}else{
										
										ip_str 			= tcp_part.substring( 0, sep );							
										tcp_port_str	= tcp_part.substring( sep+1 );	
									}
									
									int	tcp_port = Integer.parseInt( tcp_port_str );
										
									if ( tcp_port > 0 && tcp_port < 65536 ){
		
										String	flag_str	= null;
										int		udp_port	= -1;
										
										try{
											for (int i=1;i<tokens.length;i++){
											
												String	token = tokens[i].trim();
												
												if ( token.length() > 0 ){
													
													if ( Character.isDigit( token.charAt( 0 ))){
														
														udp_port = Integer.parseInt( token );
														
														if ( udp_port <= 0 || udp_port >=65536 ){
															
															udp_port = -1;
														}
													}else{
														
														flag_str = token;
													}
												}
											}
										}catch( Throwable e ){
										}
									
										addresses.add( 
												ip_str==null?originator.getAddress().getAddress().getHostAddress():ip_str);
										
										ports.add( new Integer( tcp_port ));
										
										udp_ports.add( new Integer( udp_port==-1?originator.getAddress().getPort():udp_port));
										
										flags.add( flag_str );
										
										if (( value.getFlags() & DHTPlugin.FLAG_DOWNLOADING ) == 1 ){
											
											leecher_count++;
											
											is_seeds.add( new Boolean( false ));
		
										}else{
											
											is_seeds.add( new Boolean( true ));
											
											seed_count++;
										}
									}
								
								}catch( Throwable e ){
									
									// in case we get crap back (someone spamming the DHT) just
									// silently ignore
								}
							}
							
						}

						public void valueWritten(DHTPluginContact target,
								DHTPluginValue value) {
						}});
			}
		
			while (result != null) {
				try {
					Thread.sleep(5*1000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			
			FactScrapeProbeResult ret = result;
			result = null;
			return ret;

	}

	protected FactScrapeProbeResult trackerScrape(ProbeSubTask pst) {

		final String hexHash = pst.getIpAddress();
		final Integer taskId = pst.getTaskId();
		
		String trackerUrlTmp = "";
		if(pst.getParams() != null && pst.getParams().containsKey("trackerUrl")) 
			trackerUrlTmp = (String) pst.getParams().get("trackerUrl");
		
		final String trackerUrl = trackerUrlTmp;
		final String target = pst.getTarget();
		
		try {

			TOTorrent torrent = null;
			
			String file = Main.getDir() + java.io.File.separator+ 
				"temp.torrent";
			
			File sourceFile = new File(file);
			FileOutputStream fos = new FileOutputStream(sourceFile);
			DataOutputStream dos = new DataOutputStream(fos);
		    dos.writeInt(2333);
		    dos.writeChars("Hello");	
		    dos.close();
		    fos.close();
		    
			int piece_size = 1024;
			TOTorrent temp = TOTorrentFactory.createFromFileOrDirWithFixedPieceLength(
					sourceFile,
					new URL( "dht://a.b.c/" ),
					piece_size ).create();
			temp.setHashOverride(hexStringToByteArray(hexHash));
			temp.setAnnounceURL(new URL(trackerUrl));
			torrent = temp;	
			
			//Scrape
			TRTrackerScraper scraper = Main.getGlobalManager().getTrackerScraper();
			TRTrackerScraperResponse response = scraper.scrape(torrent, true);
			
			LOGGER.fine("status: " + response.getStatusString());
			LOGGER.fine("length: " + response.getPeers());
			LOGGER.fine("time next update: " + response.getNextScrapeStartTime());
			
			FactScrapeProbeResult ret = null;
			
			//if the client has not finished but has not returned an error then wait for 5 seconds to allow the tracker
			//to reponse, else report as error
			if ((response.getStatus() != TRTrackerScraperResponse.ST_ERROR) 
					&& (response.getStatus() != TRTrackerScraperResponse.ST_ONLINE)) {
				LOGGER.info(pst.getParent().probeType + ": Waiting for 10 seconds...");
				Thread.sleep(10*1000);
			}
			
			sourceFile.delete();
						
			LOGGER.info(pst.getParent().probeType + ": " + 
					response.getStatusString() + " " + response.getSeeds() + " " + response.getPeers());
				
			if (response.getStatus() == TRTrackerScraperResponse.ST_ONLINE) {
			
				int toWait = (int)((response.getNextScrapeStartTime() - SystemTime.getCurrentTime()) / 1000);
				
				ret = new FactScrapeProbeResult(hexHash,
						"TRACKER_SCRAPE", toWait, taskId, 
						new ArrayList<String>(), new ArrayList<Integer>(), new ArrayList<Integer>(), 
						new ArrayList<Boolean>(), response.getSeeds(), response.getPeers(), trackerUrl);
				ret.setTarget(target);
			}
			else {
				ret = new FactScrapeProbeResult(hexHash, "TRACKER_SCRAPE", taskId, trackerUrl);
				ret.setTarget(target);
			}
			
			ret.setPst(pst);					
			return ret;
			

		} catch (MalformedURLException e1) {
			e1.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (TOTorrentException e) {
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return null;
		
	}
	
	protected FactScrapeProbeResult trackerAnnounce(final ProbeSubTask pst) {
		
		final String hexHash = pst.getIpAddress();
		final Integer taskId = pst.getTaskId();
		
		String trackerUrlTmp = "";
		if(pst.getParams() != null && pst.getParams().containsKey("trackerUrl")) 
			trackerUrlTmp = (String) pst.getParams().get("trackerUrl");
		
		final String trackerUrl = trackerUrlTmp;
		final String target = pst.getTarget();
		
		try {

			TOTorrent torrent = null;
			
			String file = Main.getDir() + java.io.File.separator + 
				"temp.torrent";
			
			File sourceFile = new File(file);
			FileOutputStream fos = new FileOutputStream(sourceFile);
			DataOutputStream dos = new DataOutputStream(fos);
		    dos.writeInt(2333);
		    dos.writeChars("Hello");	
		    dos.close();
		    fos.close();
		    
			int piece_size = 1024;
			TOTorrent temp = TOTorrentFactory.createFromFileOrDirWithFixedPieceLength(
					sourceFile,
					new URL( "dht://a.b.c/" ),
					piece_size ).create();
			temp.setHashOverride(hexStringToByteArray(hexHash));
			temp.setAnnounceURL(new URL(trackerUrl));
			torrent = temp;	
			
			//Announce
			
			trackerClient = TRTrackerAnnouncerFactory.create( torrent, null);
			trackerClient.addListener(new TRTrackerAnnouncerListener() {
			
		
				public void receivedTrackerResponse(
						TRTrackerAnnouncerResponse response) {
					
					FactScrapeProbeResult ret = null;
					int status = response.getStatus();
					
					LOGGER.fine("receivedTrackerResponse()");
					LOGGER.fine("status: " + response.getStatusString());
					if (( (status != TRTrackerAnnouncerResponse.ST_OFFLINE) && (status != TRTrackerAnnouncerResponse.ST_REPORTED_ERROR)) ) {
						LOGGER.fine("length: " + response.getPeers().length);
					}
					LOGGER.fine("time next update: " + response.getTimeToWait());
					//response.print();
				
					List<String>	addresses 	= new ArrayList<String>();
					List<Integer>	ports		= new ArrayList<Integer>();
					List<Integer>	udp_ports	= new ArrayList<Integer>();
					List<Boolean>	is_seeds	= new ArrayList<Boolean>();

					if (( (status != TRTrackerAnnouncerResponse.ST_OFFLINE) 
							&& (status != TRTrackerAnnouncerResponse.ST_REPORTED_ERROR)) ) {
						
						for (TRTrackerAnnouncerResponsePeer p : response.getPeers()) {
							addresses.add(p.getAddress());
							ports.add(p.getPort());
							udp_ports.add(p.getUDPPort());
						}
						
						ret = new FactScrapeProbeResult(hexHash,
								"TRACKER_ANNOUNCE", response.getTimeToWait(), taskId, 
								addresses, ports, udp_ports, is_seeds, 
								response.getScrapeCompleteCount(), response.getScrapeIncompleteCount(), 
								trackerUrl);
						ret.setTarget(target);
					}
					
					else {
						ret = new FactScrapeProbeResult(hexHash, "TRACKER_ANNOUNCE", taskId, trackerUrl);
						ret.setTarget(target);
					}
					
					ret.setPst(pst);
					result = ret;
		
				}
		
				public void urlChanged(TRTrackerAnnouncer announcer,
						URL old_url, URL new_url, boolean explicit) {
				}
		
				public void urlRefresh() {
				}});
			
			trackerClient.setAnnounceDataProvider(
					new TRTrackerAnnouncerDataProvider()
					{
						public String
						getName()
						{
							return( "trackerClient" );
						}
						
						public long
						getTotalSent()
						{
							return( 0 );
						}
						
						public long
						getTotalReceived()
						{
							return( 0 );
						}
						
		    			public long 
		    			getFailedHashCheck() 
		    			{
		    				return( 0 );
		    			}
		    			
						public long
						getRemaining()
						{
							return( 0 );
						}
						
						public String
						getExtensions()
						{
							return( null );
						}
						
						public int
						getMaxNewConnectionsAllowed()
						{							
							return (50);	
						}
						
						public int 
						getUploadSpeedKBSec(
							boolean estimate) 
						{
							return 0;
						}
						
						public int 
						getCryptoLevel() 
						{
							return 0;
						}
						
						public boolean
						isPeerSourceEnabled(
							String		peer_source )
						{
							return( true );
						}
						
						public void
						setPeerSources(
							String[]	sources )
						{
						}

						public int getConnectedConnectionCount() {
							// TODO Auto-generated method stub
							return 0;
						}

						public int getPendingConnectionCount() {
							// TODO Auto-generated method stub
							return 0;
						}
					});
			
			if ( trackerClient != null) {
				trackerClient.setTrackerURL(new URL(trackerUrl));
				trackerClient.update( true );
			}
			else {
				//There was a problem creating the trackerClient 
				//insert empty resultset 
				FactScrapeProbeResult ret = new FactScrapeProbeResult(hexHash, "TRACKER_ANNOUNCE", taskId, trackerUrl);
				ret.setTarget(target);				
				ret.setPst(pst);
				return ret;				
			}
			
			//delete temp file before you are done
			sourceFile.delete();
		
		} catch (MalformedURLException e1) {
			e1.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (TOTorrentException e) {
			e.printStackTrace();
		} catch (TRTrackerAnnouncerException e) {
			e.printStackTrace();
		}
			
		
		while (result != null) {
			try {
				Thread.sleep(5*1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		FactScrapeProbeResult ret = result;
		result = null;
		return ret;
	}
	
	public  String getHexString(byte[] b) {
		
		String result = "";
		for (int i=0; i < b.length; i++) {
			result +=
		          Integer.toString( ( b[i] & 0xff ) + 0x100, 16).substring( 1 );
		}
		return result;
	}
	
	private static byte[] hexStringToByteArray(String s) {
		
		byte[] b = new byte[s.length() / 2];
	    for (int i = 0; i < b.length; i++){
	    	int index = i * 2;
	    	int v = Integer.parseInt(s.substring(index, index + 2), 16);
	    	b[i] = (byte)v;
	    }
	    return b;
	}

	@Override
	public Integer getNumberActiveProbes() {
		
		if (!isIdle())
			return 1;
		
		return 0;
		
	}

}
