/**
 * Dasu Project
 *
 * File:         AMCoordinatorV2.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Dec 15, 2009 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;

import java.io.IOException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;
import java.util.Set;
import java.util.TimeZone;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.logging.Logger;

import org.gudy.azureus2.core3.util.AEThread;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadException;

import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.asn.NetworkAdminASNLookupImpl;
import edu.northwestern.dasu.drools.FactFireAction;
import edu.northwestern.dasu.drools.FactProbeResult;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.drools.FactTestResult;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.drools.RuleBasedEngineInterface;
import edu.northwestern.dasu.measurement.ame.AMEManager;
import edu.northwestern.dasu.measurement.scheduler.LimitedQuotaQueue;
import edu.northwestern.dasu.measurement.scheduler.Scheduler;
import edu.northwestern.dasu.measurement.scheduler.SchedulerAggregator;
import edu.northwestern.dasu.measurement.scheduler.UnlimitedQueue;
import edu.northwestern.dasu.measurement.tasks.DasuSubTask;
import edu.northwestern.dasu.measurement.tasks.DasuTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.tasks.TestTask;
import edu.northwestern.dasu.measurement.tasks.TestTask.TestSubTask;
import edu.northwestern.dasu.measurement.tests.DasuTestAzureusSpeed;
import edu.northwestern.dasu.measurement.tests.DasuTestAzureusUpnp;
import edu.northwestern.dasu.measurement.tests.DasuTestModule;
import edu.northwestern.dasu.measurement.tests.DasuTestPortThrottling;
import edu.northwestern.dasu.measurement.types.AlternateType;
import edu.northwestern.dasu.measurement.types.CronjobType;
import edu.northwestern.dasu.measurement.types.EventType;
import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.measurement.types.TaskType;
import edu.northwestern.dasu.measurement.types.TestType;
import edu.northwestern.dasu.measurement.types.TimeType;
import edu.northwestern.dasu.netinterface.NetInterfaceClassifier;
import edu.northwestern.dasu.profilers.PerProbeProfiler;
import edu.northwestern.dasu.profilers.Profiler;
import edu.northwestern.dasu.signal.holders.HistoricSignalHolder;
import edu.northwestern.dasu.signal.holders.SessionSignalHolder;
import edu.northwestern.dasu.signal.statistics.SignalManipulation;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.timer.CronjobEntry;
import edu.northwestern.dasu.timer.DasuCronScheduler;
import edu.northwestern.dasu.timer.DasuMessageManager;
import edu.northwestern.dasu.timer.DasuTimerManager;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.timer.TimerEntry;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.Util;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The AMCoordinatorV2 class coordinates measurements/tests between all
 *         measurement/test modules.
 **/

@SuppressWarnings({ "unused", "deprecation" })
public class AMCoordinatorV2 implements Runnable {

	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");

	private static boolean active = false;
	private static boolean pause = false;
	private static AMCoordinatorV2 self = null;

	/* sleepInterval in seconds */
	private static int sleepInterval = -1;
	
	/* Map of ProbeType to DasuProbeModule instances */
	@SuppressWarnings("rawtypes")
	public static Hashtable<ProbeType, DasuProbeModule> probeModules = null;
	/* Map of TestType to DasuTestModule instances */
	@SuppressWarnings("rawtypes")
	public static Hashtable<TestType, DasuTestModule> testModules = null;
	/* Holds the tests that are already running */

	@SuppressWarnings("rawtypes")
	private static HashMap<String, Scheduler> schedulers = new HashMap<String, Scheduler>();

	/*
	 * Queue used by every DasuProbeModule (and Drools Engine) used to alert
	 * (awake) the AMC of probe results it needs to collect (or of probe tasks
	 * it needs to perform)
	 */
	public static LinkedBlockingDeque<EventType> events = null;

	/* Used by the AMC to keep track of the probes/tests already dispatched */
	private static Set<EventType> dispatchedEvents = null;
	
	private static LinkedList<Download> activeTorrents = new LinkedList<Download>();
	private static Hashtable<String, HistoricSignalHolder> historicSamples = null;
	private static Hashtable<String, SessionSignalHolder> sessionSamples = new Hashtable<String, SessionSignalHolder>();
 
	public static Profiler profiler = null;
	public static PerProbeProfiler perProbeProfiler = null;

	private AMCoordinatorV2() {

		if (Main.DEBUG)
			LOGGER.info("DEBUG: AMCoordinator.class: startDasuAmModules()");

		sleepInterval = DasuConfiguration.getInstance().getAMCSleepInterval();
		historicSamples = DasuManager.getPersistentDataManager().getHistoricSamplesV2();

	}

	public static AMCoordinatorV2 getInstance() {
		if (self == null) {
			self = new AMCoordinatorV2();
		}
		return self;
	}

	public static void profile(Profiler p) {
		profiler = p;
	}
	
    public static void setPerProbeProfiler(PerProbeProfiler p) {
    	perProbeProfiler = p;
    }

	@SuppressWarnings("rawtypes")
	private static void initialize() {
		
		probeModules = new Hashtable<ProbeType, DasuProbeModule>();		
		testModules = new Hashtable<TestType, DasuTestModule>();
		events = new LinkedBlockingDeque<EventType>();
		dispatchedEvents = new HashSet<EventType>();

		try {
			startDasuAmProbeModules();
			startDasuTestModules();
			initSchedulers();			
		}
		catch (Exception e) {
			LOGGER.severe("STOP: Probe Modules could not be loaded.. abort loading dasu...");
			Main.startUnloadConfigTimer();
			return;
		}
		
		setActive(true);
	}
	
	private static void startDasuTestModules() {
		
		testModules.put(TestType.AZUREUS_UPNP, new DasuTestAzureusUpnp(1.0f));
		testModules.put(TestType.FORCE_SEEDING_THROUGH_PORT, new DasuTestPortThrottling(1.0f));
		DasuTestAzureusSpeed dtas = new DasuTestAzureusSpeed(1.0f);
		testModules.put(TestType.AZUREUS_DL, dtas);
		testModules.put(TestType.AZUREUS_UL, dtas);
		testModules.put(TestType.AZUREUS_DL_ENCRYPTION, dtas);
		testModules.put(TestType.AZUREUS_UL_ENCRYPTION, dtas);
		
	}
	
	/*
	 * Initialize each DasuModule available
	 */
	
	@SuppressWarnings("rawtypes")
	private static void startDasuAmProbeModules() throws IOException {

		Properties map = DasuConfiguration.getInstance()
				.getProbeModulesProperties();
		
		//use default
		if ( map.size() == 0 ) {	        
	         map.load(DasuManager.getDasuClassLoader().getResourceAsStream(
	        		 DasuConfiguration.PROBE_MODULES_PROPERTIES));
		}
		
		LOGGER.info("DYNAMIC: loading modules: " + map);
		
		try {
			
			//add new ProbeTypes if they don't already exit			
			for ( Object probeType : map.keySet()) {
				
				try {
					
					if ( !map.getProperty(probeType.toString()).contains("," ))
						continue;
					
					String maxVals[] = map.getProperty(probeType.toString()).split(":")[1].split(",");
					int maxProbesPerMin = Integer.parseInt(maxVals[0]);
					int maxProbesPerHour = Integer.parseInt(maxVals[1]);
					int consumedBW = Integer.parseInt(maxVals[2]);
					boolean pauseOtherModule = false;
					boolean executeInIsolation = false;
					
					if ( maxVals.length > 3 ) {
						pauseOtherModule = Boolean.parseBoolean(maxVals[3]);
						executeInIsolation = Boolean.parseBoolean(maxVals[4]);
					}
										
					if (!Util.probeTypeExists(probeType.toString())) {
						
						//doesnt exist, add it												 
						DynamicEnumFactory.addEnum(ProbeType.class, probeType.toString(), 
								new Class[]{int.class, int.class, int.class, boolean.class, boolean.class}, 
								new Object[]{maxProbesPerMin,maxProbesPerHour,consumedBW,pauseOtherModule,executeInIsolation});
						LOGGER.info("DYNAMIC: adding a new probeType: "	+ probeType + " " 
								+ maxProbesPerMin + " " + maxProbesPerHour + " " + consumedBW);
						
					}
					else {
						
						HashMap<String, Object> toSet = new HashMap<String, Object>();
						toSet.put("maxProbesPerMin", maxProbesPerMin);
						toSet.put("maxProbesPerHour", maxProbesPerHour);
						toSet.put("consumedBW", consumedBW);
						toSet.put("pauseOtherModule", pauseOtherModule);
						toSet.put("executeInIsolation", executeInIsolation);
												
						try {
							
							LOGGER.info("DYNAMIC: updating probeType: "	+ probeType + " " 
									+ maxProbesPerMin + " " + maxProbesPerHour + " " + consumedBW);
														
							for ( String name : toSet.keySet() ) {
								
								Field f = ProbeType.valueOf(probeType.toString()).getClass().getDeclaredField(name); 
								f.setAccessible(true);
								f.set(ProbeType.valueOf(probeType.toString()), toSet.get(name));
								f.setAccessible(false);
							}
							
						}
						catch (Exception e) {
							e.printStackTrace();
						}
					}
					
				}
				catch (Exception e) {
					//malformed new probeType, skip
					//e.printStackTrace();
					LOGGER.warning("DYNAMIC: malformed new probeType, skipping: "	+ probeType + " " +
							map.getProperty(probeType.toString()));

				}
				
			}

			ClassLoader cl = DasuManager.getDasuClassLoader();
			//ClassLoader cl = Main.getPluginInterface().getPluginClassLoader();

			for (ProbeType pt : ProbeType.values()) {

				try {
					if (map.containsKey(pt.toString())) {
						LOGGER.info("DYNAMIC: adding: "	+ map.getProperty(pt.toString()));
						String probeType = map.getProperty(pt.toString()).split(":")[0];
						probeModules.put(pt, (DasuProbeModule) cl.loadClass(probeType).newInstance());
						LOGGER.info("DYNAMIC: Probe Type: " + pt);
						LOGGER.info("DYNAMIC: Probe Path: " + probeType);
						if (!probeModules.get(pt).isAlive()) {
							probeModules.get(pt).setDaemon(true);
							probeModules.get(pt).start();
							LOGGER.info("DYNAMIC: starting: "+ probeType + 
									" version: " + probeModules.get(pt).getVersion());
						}
					}
				} catch (Exception e1) {
					//e1.printStackTrace();
					LOGGER.warning("DYNAMIC: couldn't load probe module, skipping: " + map.getProperty(pt.toString()));
				}

			}

		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	private static void stopDasuAmModules() {

		for (ProbeType pt : ProbeType.values()) {
			if (probeModules != null && probeModules.get(pt) != null) {
				if (probeModules.get(pt).isAlive()) {

					if (pt.equals(ProbeType.valueOf("NDT"))
							&& !probeModules.get(pt).getState()
									.equals(Thread.State.TIMED_WAITING))
						probeModules.get(pt).interrupt();

					LOGGER.info("STOP: " + pt + " "  + probeModules.get(pt).getName());
					probeModules.get(pt).stopModule();
				}
			}
		}
	}

	/* Add to the *task* queue */

	public static boolean addTestTask(TestType testType) {
		return addTestTask(testType, SourceType.RE, null);
	}

	public static boolean addTestTask(TestType testType, Object[] params) {
		return addTestTask(testType, SourceType.RE, params);
	}

	public static boolean addTestTask(TestType testType, Object param) {
		return addTestTask(testType, SourceType.RE, new Object[] { param });
	}

	public static boolean addTestTask(TestType testType, SourceType sourceType,
			Object[] params) {

		if (!testType.equals(TestType.AZUREUS_UPNP)
				&& !testType.equals(TestType.FORCE_SEEDING_THROUGH_PORT)
				&& !DasuConfiguration.getInstance().isEnableAzureusSpeedTest()) {

			LOGGER.info("DROP: Azureus Speed test disabled at this time...");

			return false;
		}
		
		TestTask tt = new TestTask(testType, params, sourceType);
		return addDasuTask(tt);
	}
			
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private static void initSchedulers() {
		
		if ( schedulers.size() > 0)
			return;
		
		Scheduler reSched = new Scheduler();		
		
		Scheduler<ProbeType, ProbeTask, ProbeSubTask> ameSched = 
				new Scheduler<ProbeType, ProbeTask, ProbeSubTask>(DasuConfiguration.getInstance().getMaxBWUnitConsumptionScheduler());
		for ( ProbeType pt : probeModules.keySet() ) {
			reSched.addQueue(pt, new UnlimitedQueue());
			ameSched.addQueue(pt, new LimitedQuotaQueue(pt.getMaxPerMin(), pt.getMaxPerHour()));
		}
		
		for ( TestType tt : testModules.keySet() ) {
			reSched.addQueue(tt, new LimitedQuotaQueue(tt.getMaxPerMin(), tt.getMaxPerHour()));
		}

		schedulers.put("RE", reSched);
		schedulers.put("not-RE", ameSched);
		schedulers.put("NULL", reSched);
		
		LOGGER.info("AMC: " + schedulers.get("RE").getQueues());
	}
	
	@SuppressWarnings("rawtypes")
	private static Scheduler getScheduler(SourceType st) {
		
		String source = st.equals(SourceType.RE) ? "RE" : "not-RE";
		if (st.equals(SourceType.NULL))
				source = "NULL";
		return schedulers.get(source);
	}

	static boolean disabledMessage = false;
	/* Add to the *task* queue */
	public static boolean addProbeTask(ProbeTask pt) {

		//if (true) {
		//	if (!disabledMessage) {
		//		LOGGER.severe("*********Original Probe Request Path Disabled\n\n");
		//		disabledMessage = true;
		//	}
		//	return true;
		//}
		
		LOGGER.info("AMC: addProbeTask(): " + pt.toJSON());
		return addDasuTask(pt);
	}
	
	public static boolean addProbeTaskUnitTest(ProbeTask pt) {
		
		LOGGER.info("AMC: probe " + pt.toJSON());
		return addDasuTask(pt);
		
	}
	
	/* Add to the *task* queue */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public static boolean addDasuTask(DasuTask dt) {

		if ( dt == null)
			return false;

		Scheduler sched = getScheduler(dt.getSourceType());
		
		if ( sched == null ) {
			LOGGER.severe("AMC: Queue does not exist " + dt.getSourceType());
			return false;
		}
				
		/*high priority probes be limited to the RE*/
		if ( !dt.getSourceType().equals(SourceType.RE ))
			dt.setPriority(0);
		
		// try to schedule
		boolean success = sched.push(dt);

		if ((profiler != null) && ( dt instanceof ProbeTask) ){
			if (success) {
				profiler.probeQueued((ProbeTask)dt);
			} else {
				profiler.probeDenied((ProbeTask)dt);
			}
		}
		
        if ((perProbeProfiler != null ) && ( dt instanceof ProbeTask)) {        	
        	if (!dt.getSourceType().equals(SourceType.NULL) && !dt.getSourceType().equals(SourceType.RE)) {
        		if (!success) {

        			for ( Object dst : dt.getSubTaskList()) { 
        				perProbeProfiler.logProbeDenied((DasuSubTask)dst);
        			}
        		}
        		else {
        			for ( Object dst : dt.getSubTaskList()) { 
        				perProbeProfiler.logProbeRequested((DasuSubTask)dst);
        			}
        		}
        	}

        }
		
		if ( success) {

			/*
			 * If and ONLY if the task is of highest priority, add it to the
			 * *events* queue in order to awake the AMC, otherwise let the AMC deal
			 * with it on next scheduled awake
			 
			if ((dt.getPriority() == -1) || ( dt instanceof TestTask )){
			*/
			
			//awake the AMC right away
	
				synchronized (events) {
									
						LOGGER.info("AMC: append to events: " + dt.toJSON());
					
						if ( dt.getSourceType().equals(SourceType.RE))
							events.offerFirst(AlternateType.RE_REQUEST);		
						else
							events.offerFirst(AlternateType.AME_REQUEST);
						
						events.notifyAll();
									
				}
			//}
		}

		return success;
	} 
	
	public static boolean addLowPriorityProbe(String ip, ProbeType pt,
			SourceType st) {

		ArrayList<String> ipList = new ArrayList<String>();
		ipList.add(ip);

		ProbeTask task = new ProbeTask(-1, -1, 0L, 0L, ipList, pt,TaskType.BULK, st);

		task.sourceType = st;

		return addProbeTask(task);

	}

	/* To be used only by the Detection Rule Engine */
	public static boolean addPriorityProbe(String ip, ProbeType pt) {
		return addPriorityProbe(ip, pt, null);
	}

	/* To be used only by the Detection Rule Engine */
	public static boolean addPriorityProbe(String ip, ProbeType pt,
			String target) {
		
		ArrayList<String> ipList = new ArrayList<String>();
		ipList.add(ip);
		ProbeTask task = new ProbeTask(-1, -1, -1, ipList, pt,TaskType.BULK, SourceType.RE);
		task.setTarget(target);

		return addPriorityProbe(task);
	}

	/* To be used only by the Detection Rule Engine */
	public static boolean addPriorityProbe(ProbeTask task) {
				
		// force the source of the task to be the RE, in case the user didn't
		task.setSourceType(SourceType.RE);
		boolean ret = addProbeTask(task);
		return ret;

	}
	
	private static boolean pauseProbeModulesAndWait(ProbeType except) {
		
		boolean allPaused = true;
		ProbeType pt = except;
		
		for ( ProbeType probetype : probeModules.keySet() ) {
			if ( pt != null && !probetype.equals(pt) )
				probeModules.get(probetype).setPause(true);
		}
		
		int i = 0;
			
		do {	
			
			allPaused = true;
			for ( ProbeType probetype : probeModules.keySet() ) {
				//if ( !probetype.equals(ProbeType.NDT) &&  !probetype.equals(ProbeType.NDT2))
				if ( pt != null && !probetype.equals(pt) )
					allPaused &= probeModules.get(probetype).isPaused();
			}
		
			if ( !allPaused ) {
		
				if (Main.DEBUG)
					LOGGER.info("AMC: AMCoordinatorV2.class: Waiting for ProbeModules to pause...");

				try {
					//wait for 60 seconds, otherwise break
					Thread.sleep(6 * 1000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			
		} while ( i++ < 10  && !allPaused);	
			
		if ( !allPaused ) {
			LOGGER.info("AMC: Not every other probe modules paused within 60 seconds, aborting...");
			resumeProbeModules();
		}
				
		//ensure the 'except' module is not paused
		if ( pt != null )
			probeModules.get(pt).setPause(false);
	
		return allPaused;
		
	}
	
	private static void resumeProbeModules() {
		
		LOGGER.info("AMC: Restarting probe modules...");
		for ( ProbeType probetype : probeModules.keySet() ) {
			probeModules.get(probetype).setPause(false);
		}

	}
	
	/*
	 * Retrieves a probe of the specified DasuSubTask from the Queue and, if
	 * one is available, passes the target to the corresponding module
	 * for execution
	 */
	
	@SuppressWarnings("rawtypes")
	private static void launchEvent(DasuSubTask dst) {
		
		EventType et = dst.getType();
		LOGGER.info("AMC: launching... " + dst);
		
		if ( et instanceof ProbeType && profiler != null) {
			profiler.probeScheduled((ProbeSubTask)dst);			
		}
		
		if ( et.isPauseOtherModules() ) {
		
			/* pause all other modules */
			boolean allPaused = true;
			if ( et instanceof ProbeType)
				allPaused = pauseProbeModulesAndWait((ProbeType)et);
			else
				allPaused = pauseProbeModulesAndWait(null);
				
			
			if ( allPaused ) {
			
				/* pause all torrents */
				if ( et.isExecuteInIsolation() ) {
									
					LOGGER.info("AMC: Temporarily stopping all ("+ activeTorrents.size()	+ ") active torrents");
					stopActiveTorrents();
					
				}
				
				if ( et instanceof ProbeType )
					probeModules.get(et).probe((ProbeSubTask)dst);
				else
					testModules.get(et).runTest((TestSubTask)dst);
				
				synchronized(dispatchedEvents) {
					dispatchedEvents.add(et);
				}
			}
		}
		else {
			
			if ( et instanceof ProbeType ) {
				LOGGER.info("AMC: " + et + " " + dst);
				probeModules.get(et).probe((ProbeSubTask)dst);
			}
			else
				testModules.get(et).runTest((TestSubTask)dst);

			synchronized(dispatchedEvents) {
				dispatchedEvents.add(et);
			}
		}

	}
	
	private boolean cpuLimitExceeded() {
		
		if ( DasuManager.getResourceProfiler() != null) {
			
			if (DasuManager.getResourceProfiler().getDasuCpuUsage()/100 >
				DasuConfiguration.getInstance().getMaxCpuUtilizationAllowedScheduler()) {
				LOGGER.info("AMC: cpuUtilization: " + DasuManager.getResourceProfiler().getDasuCpuUsage()/100 
						+ " maxCpuUtilization: " + DasuConfiguration.getInstance().getMaxCpuUtilizationAllowedScheduler());
				return true;
			}
		}
		
		return false;
			
	}
	
	private boolean bwLimitExceeded() {
		
		if (historicSampleExists("speed_baseline_week_hour_1") &&
				sessionSampleExists("netstat_download_hour") &&
				historicSampleGetSignalHolder("speed_baseline_week_hour_1").getValues("downloadRateNetstat") != null &&				
				sessionSampleGetValues("netstat_download_hour") != null) {

			Collection<Float> downloadRate = historicSampleGetSignalHolder("speed_baseline_week_hour_1").getValues("downloadRate");
			downloadRate.removeAll(Collections.singleton(-0.0f));
			Collection<Float> downloadRateNetstat = historicSampleGetSignalHolder("speed_baseline_week_hour_1").getValues("downloadRateNetstat");
			downloadRateNetstat.removeAll(Collections.singleton(-0.0f));			
			
			Collection<Object> downloadRateNow = AMCoordinatorV2.sessionSampleGetValues("netstat_download_hour");
			
			Float downloadRate95 = SignalManipulation.getPercentile(downloadRate, 95); //kBps
			Float downloadRateNetstat95 = SignalManipulation.getPercentile(downloadRateNetstat, 95); //kBps
			Float downloadRateNowLast = (Float) downloadRateNow.toArray()[downloadRateNow.size()-1]; //kBps

			if (downloadRateNowLast <= 0)
				return false;
			
			//use value recorded through netstat if higher than NDT
			if ( downloadRate95 < downloadRateNetstat95)
				downloadRate95 = downloadRateNetstat95;
			
			//current speed higher than max measured, max measured must be wrong
			if ( downloadRateNowLast > downloadRate95) {
				LOGGER.info("AMC: current speed higher than max measured, max measured must be wrong.  downloadRateNowLast: " + downloadRateNowLast + " downloadRate95 (NDT): " + downloadRate95);
				return false;
			}
				
			if ( downloadRateNowLast > downloadRate95*DasuConfiguration.getInstance().getMaxBWUtilizationAllowedScheduler() ) {
				LOGGER.info("AMC: downloadRateNowLast: " + downloadRateNowLast + " downloadRate95 (NDT): " + downloadRate95);
				LOGGER.info("AMC: downloadRate: " + downloadRate);
				LOGGER.info("AMC: downloadRateNetstat: : " + downloadRateNetstat);
				return true;
			}
		}
		
		return false;
	}
	
	private boolean isDispachedStopOtherModulesEvent() {
		
		boolean stopOtherModulesRequestDispatched = false;
		
		synchronized(dispatchedEvents) {
			
			Iterator<EventType> iter = dispatchedEvents.iterator();
			while (iter.hasNext()) {
				EventType et = iter.next();
				if ( et.isPauseOtherModules() )
					stopOtherModulesRequestDispatched = true;
			}
			
		}
		
		return stopOtherModulesRequestDispatched;
	}	
	
	/**
	 * Poll probe scheduler and launch schedulable probes 
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	private void launchScheduledProbes(SourceType sourceType, ProbeType probeType) {
		
		Scheduler sched = getScheduler(sourceType);	
		
		SchedulerAggregator AMEScheds = null;
		if ( !sourceType.equals(SourceType.RE) && !sourceType.equals(SourceType.NULL)) {			
			
			AMEScheds = new SchedulerAggregator();
			for ( SourceType st : AMEManager.AMEngines.keySet()) 
				AMEScheds.addScheduler(st, AMEManager.AMEngines.get(st));			
		}
		
		Set<ProbeType> probeTypes = new HashSet<ProbeType>();
		if ( probeType == null)
			probeTypes.addAll(probeModules.keySet());
		else
			probeTypes.add(probeType);
			
		StringBuilder sb = new StringBuilder();
		for (ProbeType pt : probeTypes) {
			
			//delay probe due to BW limits
			if (bwLimitExceeded()) {
				LOGGER.info("AMC: BW limit exceeded! Delay probe");
                if (perProbeProfiler != null )
                	perProbeProfiler.logDelayDueToBWLimit(pt);
				return;
			}
			//delay probe due to CPU limits
			if (cpuLimitExceeded()) {
				LOGGER.info("AMC: CPU limit exceeded! Delay probe");
                if (perProbeProfiler != null )
                	perProbeProfiler.logDelayDueToCpuLimit(pt);
				return;				
			}
						
			sb.append("AMC: " + sourceType + " looking for scheduled probe of type: " + pt + 
					" module idle: " + probeModules.get(pt).isIdle() +  " module paused: " + probeModules.get(pt).isPaused() + 
					" minutely allocated: " + sched.getMinutelyAllocated() + " queueSize: " + sched.getQueueSize(pt) + "\n");
			
			if ( probeModules.get(pt).isIdle() && !probeModules.get(pt).isPaused() ) {								
				ProbeSubTask pst = (ProbeSubTask) sched.poll(pt, AMEScheds);
				if ( pst != null ) {
					LOGGER.info("AMC: from run! launch " + pst.toJSON());									
					launchEvent(pst);
				}
				else {
					
					Set<Pair<Long, ProbeSubTask>> delayed = sched.pollDelayedTasks();
					
					if ( delayed.size() > 0) {
						
						HashMap<ProbeType, Integer> delayedProbes = new HashMap();
						for ( Pair<Long, ProbeSubTask> pp : delayed ) {
							
                        	if (perProbeProfiler != null ) {                        		
                        		perProbeProfiler.logProbeDelayed(pp.getValue());
                        	}
                        	
							if ( profiler != null ) { 								
								profiler.probeDelayed(pp.getValue());
								if ( !delayedProbes.containsKey(pp.getValue().getType()) )
									delayedProbes.put(pp.getValue().getType(), 0);
								delayedProbes.put(pp.getValue().getType(),delayedProbes.get(pp.getValue().getType())+1);
							}
						}
						
						LOGGER.info("AMC: delayed " + delayedProbes);										
					}
					
				}
			}
			
		}
		if ( Main.DEBUG)
			LOGGER.info(sb.toString());
	}
	
	private static void startPerProfileTimer() {
		
		//every second
		DasuTimerManager.getInstance().createEvent("PerProbeProfileTimer", new TimerEntry( 
				DasuConfiguration.getInstance().getPerProbeProfilerUpdateInterval()*1000, true ),
				
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {
		
						Scheduler sched = getScheduler(SourceType.AME);
						
						HashMap<EventType, Integer> queueSize = new HashMap<EventType, Integer>();				
						HashMap<EventType, Integer> minutelyAllocated = new HashMap<EventType, Integer>();
						for ( ProbeType pt : ProbeType.values()) {
							queueSize.put(pt, sched.getQueueSize(pt));					
							minutelyAllocated.put(pt, sched.getMinutelyAllocated(pt));
						}
						HashMap<EventType, Integer> active = new HashMap<EventType, Integer>();
						for ( ProbeType pt : probeModules.keySet()) {
							active.put(pt, probeModules.get(pt).getNumberActiveProbes());
						}
						
						perProbeProfiler.recordStatus( 
								DasuManager.getResourceProfiler().getDasuCpuUsage(),						
								DasuManager.systemFact.getShortDeltaAverageDownloadRate(),
								queueSize,
								active,
								minutelyAllocated
							);
					}
		});
		
	}

	public void run() {
		
		initialize();
		
		try {
			Thread.sleep(5 * 1000);
		} catch (InterruptedException e1) {
			e1.printStackTrace();
		}

		if (DasuManager.USE_SIMULATION_V2)
			LOGGER.info("SIMULATOR: AMCoordinatorV2.class: All AMC requests are being skipped dut to SIMULATION mode!!");

		//AMCoordinatorV2.addPriorityProbe("", ProbeType.valueOf("NDT2"));
		//AMCoordinatorV2.addPriorityProbe("8.8.8.8", ProbeType.TRACEROUTE);
		//ProbeProfilerTest.runTests();
		
		if ( perProbeProfiler != null) {
			startPerProfileTimer();		
		}
	
		while (isActive()) {
			
			//launch external rule engine probes
			if ( !isDispachedStopOtherModulesEvent() ) 					
				launchScheduledProbes(SourceType.AME, null);
			
			//launch internal rule engine probes that were delayed
			if ( !isDispachedStopOtherModulesEvent() ) 					
				launchScheduledProbes(SourceType.RE, null);
			
			//launch automatic probes that were delayed (automatically requested by peerStatistics)
			if ( !isDispachedStopOtherModulesEvent() ) 					
				launchScheduledProbes(SourceType.NULL, null);

			try {
				
				EventType event = null;
				synchronized (events) {

					if (isActive() && events.isEmpty() || DasuManager.USE_SIMULATION_V2)
						events.wait(sleepInterval * 1000);
					else
						event = events.poll();
				}

				if (event != null) {
					if (event.equals(AlternateType.HALT))
						setActive(false);
					else {

						if ( event instanceof AlternateType )
							processProbeRequest(event);
						else
							processResult(event);
						
					}
				}

			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			
		}
		
		LOGGER.info("STOP: Adios...");
		
	}

	
	/*
	 * Only used to process high priority measurements submitted. 
	 */
	@SuppressWarnings("rawtypes")
	private void processProbeRequest(EventType event) {
		
		Scheduler sched = null;
		if (event.equals(AlternateType.AME_REQUEST))
			sched = getScheduler(SourceType.AME);
		else if (event.equals(AlternateType.RE_REQUEST)) 
			sched = getScheduler(SourceType.RE);
		
		if (bwLimitExceeded()) {
			LOGGER.info("AMC: BW limit exceeded! Delay probe");
            if (perProbeProfiler != null )
            	perProbeProfiler.logDelayDueToBWLimit(event);
			return;
		}
		//delay probe due to CPU limits
		if (cpuLimitExceeded()) {
			LOGGER.info("AMC: CPU limit exceeded! Delay probe");
            if (perProbeProfiler != null )
            	perProbeProfiler.logDelayDueToCpuLimit(event);
			return;			
		}		

		DasuSubTask pt = sched.poll();
		LOGGER.info("AMC: processProbeRequest " + event + " " + pt);
		if ( pt != null ) {			
			launchEvent(pt);
		}

	}

	private static RuleBasedEngineInterface getEngine(SourceType source) {


		RuleBasedEngineInterface eng = null;
		if ( source != null ) {
			if (source.equals(SourceType.RE))
				eng = edu.northwestern.dasu.DasuManager.engine;
			else if (!source.equals(SourceType.NULL))
				// eng = DasuManager.RmiServer;
				// eng = RmiServerAMCoordinatorV2.getInstance();
				eng = AMEManager.rmiAMCoordinator;
		}

		return eng;

	}

	public static void stop() {
		
		LOGGER.info("STOP: stopping AMCoordinaorV2...");
		synchronized (events) {
			setActive(false);
			//events.add(new Pair<Long, EventType>(0L, AlternateType.HALT));
			events.offer(AlternateType.HALT);
			events.notify();
		}

		if (Main.DEBUG)
			LOGGER.info("STOP: Stopping dasu probe modules...");

		stopDasuAmModules();		

		self = null;
		
	}

	public static void assertToEngine(SourceType sourceType, FactTemplate fact) {

		if (Main.DEBUG)
			LOGGER.info("DEBUG: AMCoordinatorV2: Asserting to engine " + sourceType + " fact " + fact);

		RuleBasedEngineInterface eng = getEngine(sourceType);
		if (eng != null)
			eng.addFact(fact, sourceType);

	}

	private static void processResult(EventType etype) {
					
		FactTemplate factResult = null;
		SourceType st = null;
		
		if ( etype instanceof ProbeType ) {
			factResult = probeModules.get(etype).processResult();
			st = ((FactProbeResult)factResult).getPst().getSourceType();
			if (profiler != null) {
				profiler.probeCompleted(((FactProbeResult)factResult).getPst());
			}
			if (perProbeProfiler != null && !((FactProbeResult)factResult).getPst().getSourceType().equals(SourceType.NULL)
					&& !((FactProbeResult)factResult).getPst().getSourceType().equals(SourceType.RE)) {
				perProbeProfiler.logProbeCompleted(((FactProbeResult)factResult).getPst());
			}
		}
		
		else if ( etype instanceof TestType ) {
			factResult = testModules.get(etype).processResult();
			st = ((FactTestResult)factResult).getTst().getSourceType();			
		}
		
		synchronized(dispatchedEvents) {
			dispatchedEvents.remove(etype);
		}
		assertToEngine(st, factResult);
		LOGGER.info("######SCHEDULER: " + factResult.toJSON() );
		
		if ( etype.isPauseOtherModules() )
			resumeProbeModules();		
		if ( etype.isExecuteInIsolation() )
			restartActiveTorrents();
		
	}
	
	public static void AMECompleted(SourceType engineId) {
		
		//ensure to check with the AMEManager that engine is indeed not active
		if ( !AMEManager.AMEngines.contains(engineId)) {			
			LOGGER.info("SCHEDULER: AME engine " + engineId + " completed, remove any pending probe requests...");
			getScheduler(engineId).discardAllProbesFromSource(engineId);
		}
		
	}
	
	public static String[] getPropertiesFromServer(String url) {		
		HashMap<String, String> map = Main.getPropertiesFromServer(url);
		return (String[]) map.values().toArray(new String[map.size()]);
		
	}
		
	public static HashMap<String, String> getPropertiesFromServer(Integer taskId, String resource) {
		
		HashMap<String, String> ret = new HashMap<String, String>();
		
		JSONObject json = DasuManager.getInstance().getConfigService().GETData( DasuConfiguration.DASU_SCRIPT_AME_ON_DEMAND + 
				DasuConfiguration.getInstance().getResourceScriptRequest() + "&taskid=" + taskId + "&resource=" + resource);
		if ( json != null ) {
									
			Iterator<String> iter = json.keySet().iterator();
			while ( iter.hasNext() ) {					
				String key = iter.next();
				
				try {
					ret.put( key, json.get(key).toString() );
				}
				catch (Exception e) {
					LOGGER.info("DYNAMIC: ERROR: " + e.getMessage());
				}
			}
		}
		//ArrayList<String> l = Arrays.asList(ret.values().toArray(new String[ret.size()]));
		return ret;
		
	}
	
	public static JSONObject getJSONObjectFromServer(Integer taskId, String resource) {
		
		
		JSONObject ret = new JSONObject();
		
		JSONObject json = DasuManager.getInstance().getConfigService().GETData( DasuConfiguration.DASU_SCRIPT_AME_ON_DEMAND + 
				DasuConfiguration.getInstance().getResourceScriptRequest() + "&taskid=" + taskId + "&resource=" + resource);
		
		if ( json != null ) {
			ret = json;
		}
		
		return ret;
		
	}

	/****** CRONJOB SCHEDULER INTERFACE ********/

	public static void removeCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute) {
		removeCronJob(id, hour, day, commandToExecute, CronjobType.CUSTOM);
	}

	public static void removeCronJob(String id, String commandToExecute,
			CronjobType cronjobType) {
		removeCronJob(id, new Integer[] { -1 }, new Integer[] { -1 },
				commandToExecute, cronjobType);
	}

	private static void removeCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute, CronjobType cronjobType) {

		ArrayList<Integer> hours = new ArrayList<Integer>(Arrays.asList(hour));
		ArrayList<Integer> days = new ArrayList<Integer>(Arrays.asList(day));

		if ((hours == null) || (days == null))
			return;

		if ((hours.size() == 1) && (hours.contains(-1))) {
			hours.clear();
			for (int i = 0; i < 24; i++) {
				hours.add(i);
			}
		}

		if ((days.size() == 1) && (days.contains(-1))) {
			days.clear();
			for (int i = 0; i < 7; i++) {
				days.add(i);
			}
		}

		CronjobEntry job = new CronjobEntry(id, new HashSet<Integer>(hours),
				new HashSet<Integer>(days), commandToExecute, cronjobType);

		DasuCronScheduler.getInstance().removeCronJob(job);
	}

	public static void addCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute) {
		addCronJob(id, hour, day, commandToExecute, CronjobType.CUSTOM);
	}

	public static void addCronJob(String id, String commandToExecute,
			CronjobType cronjobType) {
		addCronJob(id, new Integer[] { -1 }, new Integer[] { -1 },
				commandToExecute, cronjobType);
	}

	private static void addCronJob(String id, Integer[] hour, Integer[] day,
			String commandToExecute, CronjobType cronjobType) {

		ArrayList<Integer> hours = new ArrayList<Integer>(Arrays.asList(hour));
		ArrayList<Integer> days = new ArrayList<Integer>(Arrays.asList(day));

		if ((hours == null) || (days == null))
			return;

		if ((hours.size() == 1) && (hours.contains(-1))) {
			hours.clear();
			for (int i = 0; i < 24; i++) {
				hours.add(i);
			}
		}

		if ((days.size() == 1) && (days.contains(-1))) {
			days.clear();
			for (int i = 0; i < 7; i++) {
				days.add(i);
			}
		}

		CronjobEntry job = new CronjobEntry(id, new HashSet<Integer>(hours),
				new HashSet<Integer>(days), commandToExecute, cronjobType);

		DasuCronScheduler.getInstance().registerCronJob(job);
	}

	public static boolean addTimerJob(String id, TimerEntry timerEntry, final FactFireAction fireAction) {
		
		LOGGER.info("AMC: addTimerJob(): " + timerEntry + " " + fireAction);
		boolean ret = DasuTimerManager.getInstance().createEvent(fireAction.getSourceType() + "_" + id, 
				timerEntry, new ITimerEventPerformer(){
			
			public void perform(ITimerEvent event) {				
				AMCoordinatorV2.assertToEngine(fireAction.getSourceType(), fireAction);
			}
		});
		
		return ret;
		
	}
	
	public synchronized static void stopActiveTorrents() {

		LOGGER.info("AMC: stopping active torents...");
		for (Download d : Main.getPluginInterface().getDownloadManager()
				.getDownloads()) {

			if (d.getState() != Download.ST_STOPPED
					&& d.getState() != Download.ST_STOPPING) {

				try {
					LOGGER.info("AMC: stopping ..." + d.getName());
					d.stop();
					activeTorrents.add(d);
				} catch (DownloadException e) {
					e.printStackTrace();
				}
			}
		}

		if (activeTorrents.size() > 0)
			Main.logPlugin("Temporarily stopping all (" + activeTorrents.size()
					+ ") active torrents");

	}

	/**
	 * @deprecated  As of release 4.3, replaced by {@link #getEnvInfo()}
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Deprecated
	public static ArrayList<ArrayList<String>> getMyDnsServers() {

		try {
			
			DasuProbeModule dpm = probeModules.get(ProbeType.valueOf("DNS"));
			Object o = dpm.getClass().getMethod("getMyDnsServers", new Class<?>[]{})
					.invoke(dpm, new Object[]{});
			return (ArrayList<ArrayList<String>>) o;

		} catch (Exception e) {
			LOGGER.warning("DEBUG: Cannot get localhost's DNS servers.\nStack trace: " + e.getStackTrace());
			ArrayList<String> arr = new ArrayList<String>();
			ArrayList<ArrayList<String>> ret = new ArrayList();
			ret.add(arr);
			return ret;
			
		}

	}
	
	public Map<String, Object> getEnvInfo() {
		
		Map<String, Object> ret = new HashMap<String, Object>();
		
		try {
			ret.put("dasu.platform.name", Main.getPluginInterface().getAzureusName());
			ret.put("dasu.platform.version", Main.getPluginInterface().getAzureusVersion());
			ret.put("dasu.os.category",	Main.getPlatform());			

			ret.put("dasu.os.name", System.getProperty("os.name"));
			ret.put("dasu.os.arch", System.getProperty("os.arch"));
			ret.put("dasu.os.version", System.getProperty("os.version"));
			ret.put("dasu.java.version", System.getProperty("java.version"));
			ret.put("dasu.java.vendor", System.getProperty("java.vendor"));			

			ret.put("dasu.client.version", Main.getPluginInterface().getPluginVersion());
			ret.put("dasu.client.uirules.version", DasuConfiguration.getInstance().getInferenceVersion());
			HashMap<String, Float> probeModulesMap = new HashMap<String, Float>();
			for ( ProbeType pt : probeModules.keySet() ) {
				probeModulesMap.put(pt.name(), probeModules.get(pt).getVersion());
			}
			ret.put("dasu.client.modules", probeModulesMap);

			ret.put("net.domain", Main.getDomainName());
			ret.put("net.interface.name", Main.getInterfaceName());
			ret.put("net.interface.mac", Main.getInterfaceMAC());
			ret.put("net.ip.local", Main.getCurrentLocalIpAddress());
			ret.put("net.ip.public", Main.getCurrentPublicIpAddress());

			ret.put("net.dns.servers", getMyDnsServers());
			ret.put("net.asn", NetworkAdminASNLookupImpl.getInstance().getASNInfo().getAS());
			ret.put("net.country", NetworkAdminASNLookupImpl.getInstance().getASNInfo().getCountry());
			
			ret.put("user.uuid", DasuConfiguration.getInstance().getUUID());
			ret.put("user.tz", TimeZone.getDefault().getDisplayName(((TimeZone.getDefault().getDSTSavings() == 0)?false:true), TimeZone.LONG));
			ret.put("user.tz.offset", TimeZone.getDefault().getRawOffset());
			
			ret.put("time.sntp.offset", Util.getDasuSntpTimeOffset());
			ret.put("session.init.time.local.utc", Util.currentTimeMillis());
			ret.put("session.init.time.ntp.local", Util.currentDasuTime());
			ret.put("session.init.time.ntp.utc", Util.currentGMTTime());

			ret.put("homenet.router.privateip", Main.getCurrentLocalIpAddress());
			ret.put("homenet.connection.type", 
					new NetInterfaceClassifier(Main.getInterfaceName()).getInterfaceTypeString());
			
			
		}
		catch (Exception e) {
			LOGGER.warning("AMC: " + e.getMessage());
			e.printStackTrace();
		}
		
		return ret;
	}

	public static String getMyPublicIp() {
		return Main.getPublicIpAddress();
	}

	public synchronized static void restartActiveTorrents() {

		if (Main.DEBUG)
			LOGGER.info("NDT: Restarting all " + activeTorrents.size() + " active torrents");;

		if (activeTorrents.size() > 0)
			Main.logPlugin("Restarting all (" + activeTorrents.size()
					+ ") active torrents");

		Iterator<Download> iter = activeTorrents.iterator();
		while (iter.hasNext()) {
			try {
				Download d = iter.next();
				d.restart();
				iter.remove();
			} catch (DownloadException e) {
				e.printStackTrace();
			}
		}
	}

	/******************
	 * sessionSamples /
	 ******************/

	public static Hashtable<String, SessionSignalHolder> getSessionSample() {
		return sessionSamples;
	}

	public static boolean sessionSampleExists(String name) {
		return sessionSamples.containsKey(name);
	}

	public static boolean sessionSampleCreate(String name, int size) {

		if (!sessionSampleExists(name)) {

			try {
				sessionSamples.put(name, new SessionSignalHolder(size));
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
		}

		return true;

	}

	public static boolean sessionSampleDestroy(String name) {

		if (sessionSampleExists(name))
			sessionSamples.remove(name);
		return true;
	}

	public static boolean sessionSampleSetValue(String name, Object value) {

		if (sessionSampleExists(name)) {
			sessionSamples.get(name).set(value);
			return true;
		}
		return false;
	}

	public static void sessionSamplePrint(String name) {

		if (sessionSampleExists(name)) {
			System.out.println(sessionSamples.get(name));
		}

	}

	@SuppressWarnings("unchecked")
	public static Collection<Object> sessionSampleGetValues(String name) {

		if (sessionSampleExists(name)) {
			return (Collection<Object>) sessionSamples.get(name).getValues().clone();
		}

		return null;
	}

	public static Collection<Long> sessionSampleGetTimeStamps(String name) {

		if (historicSampleExists(name)) {
			return sessionSamples.get(name).getTime();
		}

		return null;
	}
	
	public ArrayList<String> getProbeModules() {
		
		ArrayList<String> ret = new ArrayList<String>();
		for ( ProbeType pt : probeModules.keySet() ) {
			ret.add(pt.toString());
		}
		
		return ret;
	}

	/*******************
	 * historicSamples
	 *******************/

	public static Hashtable<String, HistoricSignalHolder> getHistoricSample() {
		return historicSamples;
	}

	public static boolean historicSampleExists(String name) {
		return historicSamples.containsKey(name);
	}

	public static boolean historicSampleExists(String name, String signalName) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).exists(signalName);
		}

		return false;
	}

	public static boolean historicSampleCreate(String name,
			TimeType timeSpanType, TimeType intervalType, int intervalLength) {

		if (!historicSampleExists(name)
				&& (historicSamples.size() < DasuConfiguration.getInstance()
						.getMaxLogEntries())) {

			try {
				historicSamples.put(name, new HistoricSignalHolder(name,
						timeSpanType, intervalType, intervalLength));
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
		}

		if (Main.DEBUG)
			historicSamplePrint(name);

		return true;

	}

	public static boolean historicSampleDestroy(String name) {

		if (historicSampleExists(name)) {
			historicSamples.remove(name);
			return true;
		}

		return false;

	}

	public static float historicSampleGetValue(String name, String signalName,
			long timeRecorded) {

		if (historicSampleExists(name))
			return historicSamples.get(name).get(signalName, timeRecorded);

		return -1;
	}

	public static HistoricSignalHolder historicSampleGetSignalHolder(String name) {

		if (historicSampleExists(name))
			return historicSamples.get(name);

		return null;

	}

	public static long historicSampleGetTimeStamp(String name, long timeRecorded) {

		if (historicSampleExists(name))
			return historicSamples.get(name).getTimeStamp(timeRecorded);

		return -1;
	}

	public static long historicSampleGetTimeStamp(String name,
			String signalName, float value) {

		if (historicSampleExists(name)) {

			LinkedHashMap<Integer, Float> vals = historicSamples.get(name)
					.getValuesHash(signalName);
			LinkedHashMap<Integer, Long> times = historicSamples.get(name)
					.getTimeStampsHash();

			Iterator<Entry<Integer, Float>> iter = vals.entrySet().iterator();
			int index = -1;
			while (iter.hasNext()) {
				Entry<Integer, Float> entry = iter.next();
				if (entry.getValue().equals(value))
					index = entry.getKey();
			}

			if (index != -1) {
				return times.get(index);
			}
		}

		return -1;

	}

	public static boolean historicSampleSetValue(String name,
			String signalName, long timeRecorded, float value, String unit) {

		if (historicSampleExists(name)) {
			historicSamples.get(name)
					.set(signalName, timeRecorded, value, unit);
			return true;
		}

		return false;
	}

	public static void historicSamplePrint(String name) {

		if (historicSampleExists(name)) {
			System.out.println(historicSamples.get(name));
		}

	}

	public static Collection<Float> historicSampleGetValues(String name,
			String signalName) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getValuesHash(signalName).values();
		}

		return null;
	}

	public static Collection<Long> historicSampleGetTimeStamps(String name) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getTimeStampsHash().values();
		}

		return null;
	}

	public static Collection<Float> historicSampleGetValuesSubsetPast(
			String name, String signalName, long timeRecorded, int numIntervals) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getValuesSubsetPast(signalName,
					timeRecorded, numIntervals);
		}

		return null;
	}

	public static Collection<Long> historicSampleGetTimeStampsSubsetPast(
			String name, String signalName, long timeRecorded, int numIntervals) {

		if (historicSampleExists(name)) {
			return historicSamples.get(name).getTimeStampSubsetPast(
					timeRecorded, numIntervals);
		}

		return null;
	}

	public static void commitReportableFact(ReportableFact fpr) {

		Statistics.getInstance().addReportableFact(fpr);
		
	}

	public static void 	commitReportableFact(String type, Map<String, Object> values)
			throws Exception {
		
		ReportableFact rf = new ReportableFact("ReportableFact");
		rf.addContentsObject("type", type);
		for ( String s : values.keySet() ) {
			rf.addContentsObject(s, values.get(s));
		}		
		Statistics.getInstance().addReportableFact(rf);
		
		LOGGER.info("RMI: " + rf.toJSON().toString());

	}

	/****************
	 * Deprecated
	 ***************/

	@Deprecated
	public static void commitStaticUPnPInfo(HashMap<String, String> vals) {

		if ( vals != null) {
			ReportableFact rf = new ReportableFact("StaticUPnPInfo");
			for (String key : vals.keySet())
				rf.addContentsObject(key, vals.get(key));
			Statistics.getInstance().addReportableFact(rf);
		}

	}

	@Deprecated
	public static void commitInterference(String type, Long timestamp) {

		ReportableFact rf = new ReportableFact("Interference");
		rf.addContentsObject("type", type);
		rf.addContentsObject("timestamp", timestamp);
		Statistics.getInstance().addReportableFact(rf);
	}

	@Deprecated
	public static void commitFeedback(String type, String name, String email,
			String message, String dump) {

		ReportableFact rf = new ReportableFact("ReportableFact");
		rf.addContentsObject("type", type);
		rf.addContentsObject("name", name);
		rf.addContentsObject("email", email);
		rf.addContentsObject("message", message);
		rf.addContentsObject("dump", dump);
		Statistics.getInstance().addReportableFact(rf);	
		
		LOGGER.info("RMI: " + rf.toJSON().toString());
	}

	@Deprecated
	public static void commitUserReports(String dids, String comment) {

		ReportableFact rf = new ReportableFact("UserReports");
		rf.addContentsObject("dids", dids);
		rf.addContentsObject("comment", comment);
		Statistics.getInstance().addReportableFact(rf);
	}

	@Deprecated
	public static void commitIspInfo(String prefix, String ispname) {

		ReportableFact rf = new ReportableFact("IspInfo");
		rf.addContentsObject("prefix", prefix);
		rf.addContentsObject("ispname", ispname);
		rf.addContentsObject("timeRecorded", Util.currentGMTTime());
		Statistics.getInstance().addReportableFact(rf);
	}

	@Deprecated
	public static void commitInterferenceEvent(String type, long timestamp,
			long downloadId, long droppedTo, float mean) {

		ReportableFact rf = new ReportableFact("InterferenceEvent");
		rf.addContentsObject("type", type);
		rf.addContentsObject("timestamp", timestamp);
		rf.addContentsObject("downloadId", downloadId);
		rf.addContentsObject("mean", mean);
		rf.addContentsObject("droppedTo", droppedTo);
		Statistics.getInstance().addReportableFact(rf);

	}

	public static float getPluginVersion() {
		return Float.parseFloat(Main.getPluginInterface().getPluginVersion());
	}

	public static boolean isActive() {
		return active;
	}

	public static boolean isPause() {
		return pause;
	}

	public static void setActive(boolean active) {
		AMCoordinatorV2.active = active;
	}

	public static void addPopup(String msg) {
		DasuMessageManager.getInstance().addMessage("NA", msg, -1, false,
				false, "");
	}

	public static void addPopup(String msg, String id, int timeout,
			boolean repeatable) {
		DasuMessageManager.getInstance().addMessage(id, msg, timeout,
				repeatable, false, "");
	}

	@Deprecated
	public static void addPopup(String msg, String id, int timeout,
			boolean repeatable, boolean audible, String details) {
		DasuMessageManager.getInstance().addMessage(id, msg, timeout,
				repeatable, audible, details);
	}

}
