/**
 * Dasu Project
 *
 * File:         DasuProbeModule.java
 * RCS:          $Id$
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 15, 2008 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;


import java.util.ArrayList;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Logger;

import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.FactProbeResult;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.measurement.AMCoordinatorV2;
import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.EventType;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.PluginInterface;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuProbeModule abstract class servers as 
 *         template for all the ProbeModules
 */

public abstract class DasuProbeModule<T extends FactProbeResult> extends Thread {	

	protected final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	private float version;
	protected Boolean active = false;
	protected boolean pause = false;
	protected PluginInterface pi;
	protected int sleepInterval = 30; //seconds 
	protected volatile int yieldInterval = -1; //milliseconds
	protected boolean idle = false;

	protected ArrayList<ProbeSubTask> toDoTaskList = new ArrayList<ProbeSubTask>();
	public ConcurrentLinkedQueue<T> resultTable;
	
	protected DasuProbeModule(String name, float version) {
		super(name);
		this.version = version;		
		resultTable = new ConcurrentLinkedQueue<T>();		
		this.pi = Main.getPluginInterface();	
		active = true;
	}
	
	protected void setYieldInterval(int timeInMillis) {
		yieldInterval = timeInMillis;
	}
	
	protected void clearYieldInterval() {
		yieldInterval = -1;
	}
	
	protected void setActive(boolean active) {
		this.active = active;
	}
	
	public boolean isActive() {
		return active;
	}
	
	public void setPause(boolean paused) {
		this.pause = paused;
	}
	
	protected boolean isPause() {
		return pause;
	}
	
	public boolean isPaused() {
		return pause && isIdle();
	}
	
	public abstract boolean isIdle();
	
	public abstract Integer getNumberActiveProbes();
	
	protected String getIp() {
		return Main.getPublicIpAddress();
	}
	
	public static boolean isValidIp(final String ip) {
        return ip.matches("^[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}\\.[\\d]{1,3}$");
    }
	
	public float getVersion() {
		return version;
	}
	
	protected synchronized void updateResultQueue(T fact, EventType type) {
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(type);
			AMCoordinatorV2.events.notifyAll();
		}
		
		afterUpdateResultQueue(fact);
	}
	
	protected void reportProbeResult(FactTemplate fact) {

		LOGGER.info("REPORTER: " + fact.toJSON());
		
		Statistics.getInstance().addReportableFact(fact);
	}	
	
	public void run() {

		while (!Main.isShuttingDown() && isActive()) {
			
			ProbeSubTask pst = null;
			synchronized (toDoTaskList) {
				while ( (toDoTaskList.size() == 0 || isPause()) 
						&& !Main.isShuttingDown() && isActive() ) {
					try {
						
						idle = true;				
						toDoTaskList.wait(sleepInterval * 1000);
						
					} catch (InterruptedException e) {
						setActive(false);
						LOGGER.info("STOP: Adios... " + this.getName());
						return;
					}
				}
				
				idle = false;

				if (toDoTaskList.size() > 0){
					pst = toDoTaskList.remove(0);
					LOGGER.info(pst.getParent().probeType + ": " + pst.getToProbe());
				}
			}
						
			if (pst != null) {				
				
				pst.recordExecutionPathTimestamp("probeLaunchTime",System.currentTimeMillis());
				pst.recordExecutionPathTimestamp("cpuProbeModuleUsageAtLaunchTime", DasuManager.getResourceProfiler().getCpuUsage(Thread.currentThread().getId()));
				pst.recordExecutionPathTimestamp("cpuDasuAtLaunchTime", DasuManager.getResourceProfiler().getDasuCpuUsage());
				pst.recordExecutionPathTimestamp("cpuVuzeAtLaunchTime", DasuManager.getResourceProfiler().getJvmCpuUsage());
				pst.recordExecutionPathTimestamp("probeQueueSizeLaunchTime", this.toDoTaskList.size());
				if ( DasuManager.systemFact.entryList.size() > 1 ) {
					pst.recordExecutionPathTimestamp("shortDetlaAvgDownloadRateSystemLaunchTime", DasuManager.systemFact.getShortDeltaAverageDownloadRate());				
					pst.recordExecutionPathTimestamp("shortDetlaAvgDownloadRateNetworkLaunchTime", DasuManager.localNetworkFact.getAverageTotalBytesReceived());
					pst.recordExecutionPathTimestamp("shortDetlaAvgDownloadRateLpsdLaunchTime", DasuManager.localPeerFact.getShortDeltaAverageDownloadRate());
				}
				
				
				T result = probeIp(pst);					
				if ( result != null) {
					long probeCompleteTime = System.currentTimeMillis();
					
					pst.recordExecutionPathTimestamp("probeCompletionTime", probeCompleteTime);
					LOGGER.info("REPORTER: " + pst.getExecutionPathTimestamps());					

					result.setPst(pst);
					resultTable.add(result);
					updateResultQueue(result, pst.getParent().probeType);
				}
				
				if ( yieldInterval != -1) {
					try {
						sleep(yieldInterval);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
				
				

			} // end while active
				
		}

		LOGGER.info("STOP: Adios... " + this.getName());
	}
	
	public void stopModule() {
		
		LOGGER.info("STOP: stopping " + this.getName() + "...");
		setActive(false);
		synchronized(toDoTaskList) {
			toDoTaskList.notify();
		}
	
	}

	public void probeAll(ProbeTask pt) {

		if ( pt != null) {
			synchronized (toDoTaskList) {
				for (ProbeSubTask pst : pt.getSubTasks())
					toDoTaskList.add(pst);
				toDoTaskList.notifyAll();
			}
		}
	}
	
	public void probeWithPriority(ProbeSubTask pst) {
		
		if ( pst != null ) {
			synchronized (toDoTaskList) {
				
				pst.recordExecutionPathTimestamp("cpuProbeModuleUsageAtRequestTime", DasuManager.getResourceProfiler().getCpuUsage(Thread.currentThread().getId()));
				pst.recordExecutionPathTimestamp("cpuDasuAtSubtaskRequestTime", DasuManager.getResourceProfiler().getDasuCpuUsage());
				pst.recordExecutionPathTimestamp("cpuVuzeAtSubtaskRequestTime", DasuManager.getResourceProfiler().getJvmCpuUsage());
				pst.recordExecutionPathTimestamp("probeQueueSizeRequestTime", this.toDoTaskList.size());

				toDoTaskList.add(pst);
				toDoTaskList.notifyAll();
			}
		}
	}

	public void probe(ProbeSubTask pst) {
		
		
		LOGGER.info(pst.getParent().probeType + ": " + pst);
		
		if ( pst != null) {
			synchronized (toDoTaskList) {
				toDoTaskList.add(pst);
				toDoTaskList.notifyAll();
			}
		}

	}
	
	public T processResult() {		
		return resultTable.poll();		
	}
	
	protected abstract T probeIp(ProbeSubTask pst);
	
	protected void afterUpdateResultQueue(T fact) {};	
	
}
