/**
 * Dasu Project
 *
 * File:         Scheduler.java
 * RCS:          $Id: Scheduler.java,v 1.0 Apr 17, 2012 12:22:52 PM mas939 Exp $
 * Description:  Scheduler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 17, 2012 at 12:22:52 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2012, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement.scheduler;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Logger;

import edu.northwestern.dasu.measurement.tasks.ProbeTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.util.Pair;

/**
 * @author John Otto &lt;jotto@eecs.northwestern.edu&gt;
 *
 * The Scheduler class ...
 */

public class ProbeScheduler extends Scheduler<ProbeType, ProbeTask, ProbeTask.ProbeSubTask> {
	// Attempt at multi-queue scheduler.
	//
	// Individual queues can block based on quotas.
	//
	// Usage is as follows:
	// 1) jobs are submitted to the scheduler with a named priority level
	// 2) an individual queue can deny admission to the queue based on
	// quotas (LimitedQuotaQueue)
	// 3) an external thread (e.g. AMC) queries the scheduler for the next
	// job to run
	
	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");

	private Hashtable<ProbeType, Queue> queues = new Hashtable<ProbeType, Queue>();
	private LinkedList<ProbeType> queueOrder = new LinkedList<ProbeType>();

	private TreeMap<Long, Integer> minutelyAllocated = new TreeMap<Long, Integer>();
	
	private int minuteBWLimit = -1; //units per minute
	
	public ProbeScheduler() {		
	}
	
	public ProbeScheduler(int minuteBWLimit) {
		this.minuteBWLimit = minuteBWLimit;
	}

	public boolean addQueue(ProbeType pt, Queue q) {
		if (queues.containsKey(pt))
			return false;
		queueOrder.add(pt);
		queues.put(pt, q);
		return true;
	}

	public Set<ProbeType> getQueues() {
		return queues.keySet();
		//return new ArrayList<String>(queueOrder);
	}

	public boolean push(ProbeTask task) {
		
		ProbeType pt = task.getProbeType();
		if (!queues.containsKey(pt))
			return false;		
		Queue q = queues.get(pt);		
		return q.push(task);
		
	}
	
	public int size() {
		int sum = 0;
		for ( ProbeType pt : queues.keySet() ) {
			sum += queues.get(pt).size();
		}
		return sum;
	}
	
	public void schedule(ProbeType pt) {
		Long now = System.currentTimeMillis();
		if ( minutelyAllocated.containsKey(now) ) {
			minutelyAllocated.put(now, minutelyAllocated.get(now)+pt.getConsumedBW());
		}
		else {
			minutelyAllocated.put(now, pt.getConsumedBW());
		}
	}
	
	public ProbeSubTask poll() {
		return poll( (Scheduler) null);
	}

	public ProbeSubTask poll(ProbeType pt) {
		return poll(pt, null);
	}
	
	public ProbeSubTask poll(ProbeType pt, Scheduler externalSched) {
		
		if (!queues.containsKey(pt))
			return null;
		
		Queue q = queues.get(pt);
		Pair<Long, ProbeSubTask> p = q.peek();		
		if ( p != null ) {
			
			if ( minuteBWLimit == -1) {
				p = q.poll();	
				schedule(p.getValue().getProbeType());
				return p.getValue();
			}
						
			if ( schedulable(p.getValue()) ) {
					
				if (externalSched != null && !externalSched.schedulable(p.getValue()))
					return null;
					
				p = q.poll();	
				schedule(p.getValue().getProbeType());					
				return p.getValue();				

			}
		}
		
		return null;
		
	}
	
	public ProbeSubTask poll(Scheduler externalSched) {

		//go around the queues until the first schedulable subTask is found
		//ensure next time poll() is called the first queue being checked
		//is the next logical one
		int aroundQueue = queueOrder.size();
		while ( aroundQueue-- > 0 ) {
		
			ProbeType pt = queueOrder.poll();
			queueOrder.add(pt);
			ProbeSubTask pst = poll(pt, externalSched);
			if ( pst != null ) {
				return pst;
			}		
		}
		
		return null;
		
	}
	
	public boolean schedulable(ProbeSubTask p) {
		
		return ( getMinutelyAllocated() + p.getProbeType().getConsumedBW()
				<= minuteBWLimit );

	}
	
	public int getMinutelyAllocated() {
		
		Long now = System.currentTimeMillis();
		Iterator<Entry<Long, Integer>> iter = minutelyAllocated.entrySet().iterator();
		
		Integer minutelyAllocatedTotal = 0;
		
		while ( iter.hasNext() ) {
			
			Entry<Long, Integer> entry = iter.next();
			if ( now - entry.getKey() > 1000 * 60 ) 
				iter.remove();
			else
				minutelyAllocatedTotal += entry.getValue();
		}
		
		return minutelyAllocatedTotal;

	}

}