/**
 * Dasu Project
 *
 * File:         Queue.java
 * RCS:          $Id: Queue.java,v 1.0 Apr 17, 2012 12:22:52 PM mas939 Exp $
 * Description:  Queue class (see below)
 * Author:       John Otto
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 17, 2012 at 12:22:52 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement.scheduler
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2012, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement.scheduler;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.logging.Logger;

import edu.northwestern.dasu.measurement.tasks.DasuTask;
import edu.northwestern.dasu.measurement.tasks.DasuSubTask;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.util.HashSetCache;
import edu.northwestern.dasu.util.Pair;

/**
 * @author John Otto &lt;jotto@eecs.northwestern.edu&gt;
 *
 * The Queue class ...
 */
public abstract class Queue<T extends DasuTask<S>, S extends DasuSubTask> {
	// A basic, unlimited queue

	protected PriorityBlockingQueue<Pair<Long, S>> q = new PriorityBlockingQueue<Pair<Long, S>>();
	protected final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");

	//based on the minutely/hourly limits
	public abstract boolean schedulable();

	//based on the size of the queue, if too long, drop!
	public abstract boolean admittable();

	protected abstract void schedule();
	
	protected Set<Pair<Long,S>> delayed = new HashSetCache<Pair<Long,S>>(100);
	
	public Queue() {
		
	}
	
	public Queue(Integer delayCacheSize) {
		delayed = new HashSetCache<Pair<Long,S>>(delayCacheSize);
	}
	
	
	//tasks delayed due to minutely limits
	public Set<Pair<Long,S>> pollDelayedTasks() {
		
		HashSet<Pair<Long,S>> ret = new HashSet<Pair<Long,S>>();
		synchronized(delayed) {
			Iterator<Pair<Long,S>> iter = delayed.iterator();
			while ( iter.hasNext() ) {
				Pair<Long,S> item = iter.next();
				ret.add(item);
			}
			delayed.clear();
		}
		
		return ret;
	}

	public Pair<Long, S> peek() {
		// if the queue has something schedulable, return it here.
		// (follow the semantics of the priorityqueue -- return the
		// first task, but don't remove it from the queue.)
		if (!schedulable() && q.peek() != null) {
			synchronized(delayed) {				
				delayed.add(new Pair<Long, S>(System.currentTimeMillis(), q.peek().getValue()));
			}
			LOGGER.info("SCHEDULER: skip, not schedulable minutely limit reached for probe type...");
			return null;
		}
				
		return q.peek();
		
	} 

	public Pair<Long, S> poll() {
		// follow the semantics of the priorityqueue -- remove and
		// return the first task from the queue.
		if (!schedulable() && q.peek() != null) {
			synchronized(delayed) {
				delayed.add(new Pair<Long, S>(System.currentTimeMillis(), q.peek().getValue()));
			}
			LOGGER.info("SCHEDULER: skip, not schedulable...");
			return null;
		}
		
		if ( q.size() > 0) {
			schedule();
			return q.poll();			
		}
		else
			LOGGER.info("SCHEDULER: empty queue...");
		
		return null;
	}	
	
	public boolean push(T pt) {
		
		if ( pt == null)
			return false;
				
		LinkedList<Boolean> admitted = new LinkedList<Boolean>();
		LinkedList<S> list = (LinkedList<S>) pt.getSubTaskList();
		Iterator<S> iter = list.iterator();
		
		/* 
		 * if high-priority task, the iterate in reverse order to ensure
		 * elements are inserted in the correct order
		 */
		//if ( pt.getPriority() == -1 )
		//	iter = list.descendingIterator();
		
		while ( iter.hasNext() ) {
			S pst = iter.next();

			if ( admittable() ) {

				long t = System.nanoTime();
				
				///* tasks with high priority are pushed to the top (in reverse order)*/
				//if ( pst.getPriority() == -1 && ( q.size() > 0))
				//	t = q.peek().getKey() - 100; 
				
				if (q.contains(new Pair<Long, S>(t, pst))) {
					LOGGER.info("SCHEDULER: Already in the Queue, Overriding! " +
							"granularity not big enough! " + pt + " " + pst);
				}
				pst.recordExecutionPathTimestamp("overallQueueSizeSubTaskCreationTime", q.size() + 1);
				q.add(new Pair<Long, S>(t, pst));
		
				admitted.add(true);
			}
			else {
				admitted.add(false);
			}
		}
		
		boolean ret = true;
		for ( boolean a : admitted)
			ret &= a;
		
		return ret;
	}

	public int size() {		
		return q.size();
	}

	public void discardAllProbesFromSource(SourceType sourceType) {
		
		int before = q.size();
		Iterator<Pair<Long,S>> iter = q.iterator();
		while (iter.hasNext()) {
			
			Pair<Long,S> pt = iter.next();
			if ( pt.getValue().getSourceType().equals(sourceType) ) {
				iter.remove();
			}
		}
		int after = q.size();
		LOGGER.info("SCHEDULER: discarded " + (before-after) + " probe requests..." );
		
	}
}