/**
 * Dasu Project
 *
 * File:         Scheduler.java
 * RCS:          $Id: Scheduler.java,v 1.0 Apr 17, 2012 12:22:52 PM mas939 Exp $
 * Description:  Scheduler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Apr 17, 2012 at 12:22:52 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2012, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement.scheduler;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Logger;

import edu.northwestern.dasu.measurement.tasks.DasuSubTask;
import edu.northwestern.dasu.measurement.tasks.DasuTask;
import edu.northwestern.dasu.measurement.types.EventType;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.util.Pair;

/**
 * @author John Otto &lt;jotto@eecs.northwestern.edu&gt;
 *
 * The Scheduler class ...
 */

@SuppressWarnings({ "unchecked", "rawtypes" })
public class Scheduler<E extends EventType, T extends DasuTask<S>, S extends DasuSubTask<E>> implements SchedulerInterface<S> {
	// Attempt at multi-queue scheduler.
	//
	// Individual queues can block based on quotas.
	//
	// Usage is as follows:
	// 1) jobs are submitted to the scheduler with a named priority level
	// 2) an individual queue can deny admission to the queue based on
	// quotas (LimitedQuotaQueue)
	// 3) an external thread (e.g. AMC) queries the scheduler for the next
	// job to run
	
	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
		
	private Hashtable<E, Queue> queues = new Hashtable<E, Queue>();
	private LinkedList<E> queueOrder = new LinkedList<E>();
	
	private TreeMap<Long, Integer> minutelyAllocated = new TreeMap<Long, Integer>();	
	private TreeMap<Long, EventType> minutelyAllocatedEventTypes = new TreeMap<Long, EventType>();
	
	private int minuteBWLimit = -1; //units per minute
	
	public Scheduler() {		
	}
	
	public Scheduler(int minuteBWLimit) {
		this.minuteBWLimit = minuteBWLimit;
	}
	
	public Set<Pair<Long, S>> pollDelayedTasks() {
		
		Set<Pair<Long, S>> ret = new HashSet<Pair<Long, S>>();
		for ( E qname : queues.keySet() ) {			
			Iterator<Pair<Long,S>> iter = queues.get(qname).pollDelayedTasks().iterator();
			while (iter.hasNext() ) {
				ret.add(iter.next());
			}
		}
		
		return ret;
	}

	public int getQueueSize(EventType e) {
		
		if ( !queues.containsKey(e) )
			return -1;
		
		return queues.get(e).size();
		
	}	
	
	public boolean addQueue(E pt, Queue q) {
		if (queues.containsKey(pt))
			return false;
		queueOrder.add(pt);
		queues.put(pt, q);
		return true;
	}

	public Set<E> getQueues() {
		return queues.keySet();
		//return new ArrayList<String>(queueOrder);
	}

	public boolean push(T task) {
		
		if (task == null) {
			LOGGER.info("SCHEDULER: task cannot be null...");
			return false;			
		}
		
		E pt = (E) task.getType();		
		if (!queues.containsKey(pt)) {
			LOGGER.info("SCHEDULER: reject, queues does not contain type " + pt.getEventType() );
			return false;		
		}
		
		Queue q = queues.get(pt);
		Boolean ret = q.push(task);
		if (!ret) {
			//queue rejected task (too many items in the queue?)
			LOGGER.info("SCHEDULER: reject, some subtasks not admittable...");
		}
		else {
			
			///* if task is of high priority, move the queue to be the first one polled*/
			//if ( task.getPriority() == -1) {			
			//	queueOrder.remove(pt);
			//	queueOrder.push(pt);
			//}
		}
		
		return ret;
		
	}
	
	public int size() {
		int sum = 0;
		for ( E pt : queues.keySet() ) {
			sum += queues.get(pt).size();
		}
		return sum;
	}
	
	public void schedule(E pt) {
		Long now = System.currentTimeMillis();
		if ( minutelyAllocated.containsKey(now) ) {
			minutelyAllocated.put(now, minutelyAllocated.get(now)+pt.getConsumedBW());			
		}
		else {
			minutelyAllocated.put(now, pt.getConsumedBW());
		}
		
		minutelyAllocatedEventTypes.put(now, pt);
	}
	
	public S poll() {
		return poll( (SchedulerInterface) null);
	}

	public S poll(E pt) {
		return poll(pt, null);
	}
	
	public S poll(E pt, SchedulerInterface externalSched) {
		
		if (!queues.containsKey(pt)) {
			LOGGER.info("SCHEDULER: skip, queues do not contain type " + pt);
			return null;
		}
		
		Queue q = queues.get(pt);
		Pair<Long, S> p = q.peek();		
		if ( p != null ) {
			
			if ( schedulable(p.getValue()) ) {
					
				if (externalSched != null && !externalSched.schedulable(p.getValue())) {
					LOGGER.info("SCHEDULER: skip, not schedulable by the externalScheduler..." + pt);
					return null;
				}
					
				p = q.poll();
				if ( p == null ) {
					LOGGER.info("SCHEDULER: skip, not schedulable by Queue..." + pt);
					return null;					
				}
				schedule(p.getValue().getType());					
				return p.getValue();				

			}
			else {
				LOGGER.info("SCHEDULER: skip, not schedulable by BW... " + pt);
			}
		}
		
		return null;
		
	}
	
	public S poll(SchedulerInterface externalSched) {

		//go around the queues until the first schedulable subTask is found
		//ensure next time poll() is called the first queue being checked
		//is the next logical one
		int aroundQueue = queueOrder.size();
		while ( aroundQueue-- > 0 ) {
		
			E pt = queueOrder.poll();
			queueOrder.add(pt);
			S pst = poll(pt, externalSched);
			if ( pst != null ) {
				return pst;
			}		
		}
		
		return null;
		
	}
	
	public boolean schedulable(S p) {
		
		if (minuteBWLimit == -1)
			return true;
		
		return ( getMinutelyAllocated() + p.getType().getConsumedBW()
				<= minuteBWLimit );

	}
	
	public int getMinutelyAllocated(EventType eventType) {

		Long now = System.currentTimeMillis();
		Iterator<Entry<Long, EventType>> iter = minutelyAllocatedEventTypes.entrySet().iterator();
		
		Integer minutelyAllocatedTotal = 0;
		
		while ( iter.hasNext() ) {
			
			Entry<Long, EventType> entry = iter.next();
			if ( now - entry.getKey() > 1000 * 60 ) 
				iter.remove();
			else {
				
				if ( entry.getValue().equals(eventType) )
					minutelyAllocatedTotal += entry.getValue().getConsumedBW();
			}
		}
		
		return minutelyAllocatedTotal;

	}
	
	public int getMinutelyAllocated() {
		
		Long now = System.currentTimeMillis();
		Iterator<Entry<Long, Integer>> iter = minutelyAllocated.entrySet().iterator();
		
		Integer minutelyAllocatedTotal = 0;
		
		while ( iter.hasNext() ) {
			
			Entry<Long, Integer> entry = iter.next();
			if ( now - entry.getKey() > 1000 * 60 ) 
				iter.remove();
			else
				minutelyAllocatedTotal += entry.getValue();
		}
		
		return minutelyAllocatedTotal;

	}
	
	public void discardAllProbesFromSource(SourceType sourceType ) {
		
		LOGGER.info("SCHEDULER: Engine " + sourceType + " terminated... removing remaining probe requests...");
		for ( E e : queues.keySet() ) {			
			LOGGER.info("SCHEDULER: removing probes of type " + e);
			Queue q = queues.get(e);
			q.discardAllProbesFromSource(sourceType);
		}
		
	}


}