/**
 * Dasu Project
 *
 * File:         ProbeTask.java
 * RCS:          $Id$
 * Description:  ProbeTask class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:23:15 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.measurement.tasks;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.dasu.measurement.types.EventType;
import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.measurement.types.SourceType;
import edu.northwestern.dasu.measurement.types.TaskType;
import edu.northwestern.dasu.timer.TimerEntry;
import edu.northwestern.dasu.util.Jsonizable;


/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The ProbeTask class ...
 */

@SuppressWarnings({ "unchecked", "rawtypes" })
public class ProbeTask extends Jsonizable implements DasuTask<ProbeTask.ProbeSubTask>, Serializable {

	private static final long serialVersionUID = 3L;
	
	private long id = System.nanoTime() + new Random().nextLong() + new Random(1000).nextLong();

	public Integer taskId;
	public Integer priority = 0;	/*default priority (highest -1, normal 0)*/
	public List<String> toProbe = new ArrayList<String>();
	public ProbeType probeType = null;
	public Map<Object, Object> params = new HashMap<Object, Object>();
				
	//if set, probe will be dropped if could not be launched within specified number of seconds
	public Integer withinSeconds = -1;
	
	//use to schedule synchronized test in the future
	public TimerEntry launchTime = null;
	
	public TaskType taskType = TaskType.BULK;
	public SourceType sourceType = SourceType.AME;	/* to redirect output for task by the AMC*/	
	public String target = null;	
	
	public ProbeTask(Integer tid, List<String> toProbe, ProbeType probeType) {
		this.taskId = tid;
		this.toProbe = new ArrayList<String>(toProbe);	
		this.probeType = probeType;
	}

	public ProbeTask(Integer tid, Integer prior, Integer withinSeconds, List<String> toProbe, ProbeType probeType) {
		this(tid, toProbe, probeType);
		this.priority = prior;
		this.withinSeconds = withinSeconds;
	}

	public ProbeTask(Integer tid, Integer prior, TimerEntry launchTime, List<String> toProbe, ProbeType probeType) {
		this(tid, toProbe, probeType);
		this.priority = prior;
		this.launchTime = launchTime;
	}

	public ProbeTask(Integer tid, Integer prior, Integer withinSeconds, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType) {
		this(tid, prior, withinSeconds, toProbe, probeType);
		this.taskType = taskType;
		this.sourceType = sourceType;
	}

	public ProbeTask(Integer tid, Integer prior, TimerEntry launchTime, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType) {
		this(tid, prior, launchTime, toProbe, probeType);
		this.taskType = taskType;
		this.sourceType = sourceType;
	}

	public ProbeTask(Integer tid, Integer prior, Integer withinSeconds, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType, Map params) {
		this(tid, prior, withinSeconds, toProbe, probeType, taskType, sourceType);
		this.params = params;
	}
	
	public ProbeTask(Integer tid, Integer prior, TimerEntry launchTime, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType, Map params) {
		this(tid, prior, launchTime, toProbe, probeType, taskType, sourceType);
		this.params = params;
	}
	

	public ProbeTask(Integer tid, Integer prior, Integer withinSeconds, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType, Map params, String target) {
		this(tid, prior, withinSeconds, toProbe, probeType, taskType, sourceType, params);
		this.target = target;
	}

	public ProbeTask(Integer tid, Integer prior, TimerEntry launchTime, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType, Map params, String target) {
		this(tid, prior, launchTime, toProbe, probeType, taskType, sourceType, params);
		this.target = target;
	}

	@Deprecated
	public ProbeTask(Integer tid, Integer prior, Long st, Long et, List<String> toProbe, ProbeType probeType) {
		this(tid, toProbe, probeType);
		this.priority = prior;
	}
		
	@Deprecated
	public ProbeTask(Integer tid, Integer prior, Long st, Long et, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType) {
		this(tid, prior, st, et, toProbe, probeType);
		this.taskType = taskType;
		this.sourceType = sourceType;
	}
	
	@Deprecated
	public ProbeTask(Integer tid, Integer prior, Long st, Long et, List<String> toProbe, ProbeType probeType, TaskType taskType, SourceType sourceType, Map params) {
		this(tid, prior, st, et, toProbe, probeType, taskType, sourceType);
		this.params = params;
	}
	
	@Deprecated
	public List<String> getIpsToProbe() {
		return toProbe;
	}
	
	public List<String> getToProbe() {
		return toProbe;
	}
	
	public Integer getPriority() {
		return priority;
	}
	
	public void setPriority(int prior) {
		priority = prior;
	}
	
	public void setSourceType ( SourceType st ) {
		sourceType = st;
	}
	
	public SourceType getSourceType() {
		return sourceType;
	}
	
	public void setWithinSeconds(int sec){
		withinSeconds = sec;
	}

	public ProbeType getProbeType() {
		return probeType;
	}

	@Override
	public String toString() {
		
		return toJSON().toJSONString();
		
	}
	
	public Integer size() {
		if (toProbe != null)
			return toProbe.size();
		else
			return 0;
	}

	@Override
	public boolean equals(Object obj) {

		if (obj == null || !(obj instanceof ProbeTask)) {
	        return false;
	    }
			
		ProbeTask pt = (ProbeTask) obj;
        return taskId.equals(pt.taskId) && priority.equals(pt.priority) && 
        		probeType.equals(pt.probeType) && withinSeconds.equals(pt.withinSeconds) &&
        		taskType.equals(pt.taskType) && sourceType.equals(pt.sourceType) &&
        		toProbe.equals(pt.toProbe) && params.equals(pt.params) &&
        		((target == null && pt.target == null ) || ( target != null && pt.target != null && target.equals(pt.target)));
	    
	}

	@Override
	public int hashCode()
	{
		return taskId.hashCode() + priority.hashCode() + probeType.hashCode() + withinSeconds.hashCode() +
				taskType.hashCode() + sourceType.hashCode() + toProbe.hashCode() + params.hashCode() +
				((target != null ) ? target.hashCode() : 0);		
	}
	
	public void setTarget(String target) {
		this.target = target;
	}
	
	public String getTarget() {
		return target;
	}
	
	public Collection<ProbeSubTask> getSubTasks() {
		
		List<ProbeSubTask> psts = new LinkedList<ProbeSubTask>();
		for(String ip: toProbe) {
			ProbeSubTask pst = new ProbeSubTask(ip, this);
			pst.recordExecutionPathTimestamp("subTaskCreationTime", System.currentTimeMillis());			
			pst.setParent(this);
			psts.add(pst);
		}
		return psts;
	}
	
	public Collection<ProbeSubTask> getSubTaskList() {
		return getSubTasks();
	}
	
	public EventType getType() {
		return getProbeType();
	}

	public long getId() {
		return id;
	}
	
    public void setParams(Map params) {
    	this.params = params;
    }

	public static class ProbeSubTask extends Jsonizable implements DasuSubTask<ProbeType>, Serializable {
				
		private static final long serialVersionUID = 3L;
		
		private ProbeTask parent;
		private long id;		
		public String toProbe;
		public Map<String,Object> executionPathTimestamps = new HashMap<String, Object>();		
		
		protected ProbeSubTask(String toProbe, ProbeTask parent) {
			this.toProbe = toProbe;
			this.parent = parent;
			this.id = parent.id;
			if ( toProbe != null)
				this.id += toProbe.hashCode();
		}
		
		public long getId() {
			return id;
		}
		
		public Map<String,Object> getExecutionPathTimestamps() {
			return executionPathTimestamps;
		}
		
		public void recordExecutionPathTimestamp(String name, Object time) {
			executionPathTimestamps.put(name, time);
		}
		
		public String getToProbe() {
			return toProbe;
		}
		
		public void setToProbe(String toProbe) {
			this.toProbe = toProbe;
		}
		
		@Deprecated
		public String getIpAddress() {
			return getToProbe();
		}
		
		@Deprecated
		public void setIpAddress(String ip) {
			setToProbe(ip);
		}
		
		public SourceType getSourceType() {
			return parent.sourceType;
		}

		public Integer getTaskId() {
			return parent.taskId;
		}
		
		public Integer getPriority() {
			return parent.priority;
		}
		
		public Integer getWithinSeconds() {
			return parent.withinSeconds;
		}
		
		public Map<Object, Object> getParams() {
			return parent.params;
		}
				
		public String getTarget() {
			return parent.target;
		}
		
		public ProbeTask getParent() {
			return parent;
		}
		
		public void setParent(ProbeTask pt) {
			parent = pt;
		}
		
		public ProbeType getProbeType() {
			return parent.probeType;
		}
				
		@Override
		public String toString() {			
			return toJSON().toJSONString();			
		}
		
		@Override
		public boolean equals(Object obj) {
			
			if (obj == null || !(obj instanceof ProbeSubTask))
				return false;
						
			ProbeSubTask pst = ((ProbeSubTask) obj);
			return ( toProbe.equals(toProbe) && parent.equals(pst.parent));
			
		}
		
		@Override
		public int hashCode() {			
			return toProbe.hashCode() + parent.hashCode();		
		}
		
		/* 
		 * only jsonize these four fields from the parent task to avoid repeated fields when jsonizing 
		 * the probesubtask
		 */
		public JSONObject toJSON() {

			JSONObject json = super.toJSON();	
			JSONObject jsonParent = parent.toJSON();
			json.put("parent", jsonParent);
			
			return json;
		}

        public ProbeType getType() {
            return getProbeType();
        }

        public int getMaxPerHour() {
            return getProbeType().getMaxPerHour();
        }

        public int getMaxPerMin() {
        	return getProbeType().getMaxPerMin();
        }

        public int getConsumedBW() {
        	return getProbeType().getConsumedBW();
        }

        public String getEventType() {
        	return getProbeType().toString();
        }

        public boolean isExecuteInIsolation() {
        	return getProbeType().isExecuteInIsolation();
        }

        public boolean isPauseOtherModules() {
        	return getProbeType().isPauseOtherModules();
        }
        
	}
	
}
