/**
 * Dasu Project
 *
 * File:         PerProbeProfiler.java
 * RCS:          $Id: PerProbeProfiler.java,v 1.0 May 21, 2013 10:54:52 AM mas939 Exp $
 * Description:  PerProbeProfiler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 21, 2013 at 10:54:52 AM
 * Language:     Java
 * Package:      edu.northwestern.dasu.profilers
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2013, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.profilers;

import java.util.Date;
import java.util.Hashtable;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.logging.Logger;

import edu.northwestern.dasu.DasuData;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.measurement.tasks.DasuSubTask;
import edu.northwestern.dasu.measurement.tasks.ProbeTask.ProbeSubTask;
import edu.northwestern.dasu.measurement.types.EventType;
import edu.northwestern.dasu.measurement.types.ProbeType;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.ui.LinePlot;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The PerProbeProfiler class ...
 */
public class PerProbeProfiler {
	
	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	
	Hashtable<EventType, LinkedHashMap<Long, ProbeProfile>> probes = 
			new Hashtable<EventType, LinkedHashMap<Long, ProbeProfile>>();
	
	LinkedList<Long> timeStamps = new LinkedList<Long>();
	LinkedList<Float> dasuCpuUsage = new LinkedList<Float>();
	LinkedList<Float> hostBWUsage = new LinkedList<Float>();	
	Hashtable<EventType, LinkedList<Float>> queueSize = new Hashtable<EventType, LinkedList<Float>>();
	Hashtable<EventType, LinkedList<Float>> activeProbes = new Hashtable<EventType, LinkedList<Float>>();
	Hashtable<EventType, LinkedList<Float>> minutelyAllocated = new Hashtable<EventType, LinkedList<Float>>();
	LinkedList<Float> minutellyAllocatedTotal = new LinkedList<Float>();
	final int MAXNUMENTRIES = 60*5; //60 entries
	final int MAXNUMPROBES = 30;
	
	// timestamp of events
	Hashtable<EventType, LinkedList<Long>> probeDelayedDueToBWLimit = new Hashtable<EventType, LinkedList<Long>>();
	Hashtable<EventType, LinkedList<Long>> probeDelayedDueToCpuLimit = new Hashtable<EventType, LinkedList<Long>>();
	
	private long lastReporterdTimeStamp = -1l;
			
	public PerProbeProfiler() {		
	}
	
	private void purgeStructuresSize() {

		if ( timeStamps.size() >= MAXNUMENTRIES ) {
			
			if ( timeStamps.getFirst() >= lastReporterdTimeStamp) {
				lastReporterdTimeStamp = timeStamps.getLast();
				ReportableFact rf = new ReportableFact("PerProbeProfiler");
				rf.addContentsObject("timeStamps", timeStamps);
				rf.addContentsObject("dasuCpuUsage", dasuCpuUsage);
				rf.addContentsObject("hostBWUsage", hostBWUsage);
				rf.addContentsObject("queueSize", queueSize);
				rf.addContentsObject("activeProbes", activeProbes);
				rf.addContentsObject("minutelyAllocated", minutelyAllocated);
				rf.addContentsObject("minutellyAllocatedTotal", minutellyAllocatedTotal);
				Statistics.getInstance().addReportableFact(rf);
				
				LOGGER.info("PERPROBEPROFILER: " + rf.toJSON());
			}
			
			this.timeStamps.pop();
			this.dasuCpuUsage.pop();
			this.hostBWUsage.pop();
			this.minutellyAllocatedTotal.pop();
			
			for ( EventType pt : queueSize.keySet() ) {
				queueSize.get(pt).pop();
				activeProbes.get(pt).pop();
				minutelyAllocated.get(pt).pop();
			}
		}		

		for ( EventType pt : probeDelayedDueToBWLimit.keySet() ) {
			if ( probeDelayedDueToBWLimit.get(pt).size() >= MAXNUMENTRIES )
				probeDelayedDueToBWLimit.get(pt).pop();
		}

		for ( EventType pt : probeDelayedDueToCpuLimit.keySet() ) {
			if ( probeDelayedDueToCpuLimit.get(pt).size() >= MAXNUMENTRIES )
				probeDelayedDueToCpuLimit.get(pt).pop();
		}
	}
	
	public void recordStatus(Float dasuCpuUsage, Float hostBWUsage, 
			Map<EventType, Integer> queueSize, Map<EventType, Integer> activeProbes,
			Map<EventType, Integer> minutelyAllocated) {
		
		purgeStructuresSize();
		
		this.timeStamps.add(System.currentTimeMillis());
		this.dasuCpuUsage.add(dasuCpuUsage);
		this.hostBWUsage.add(hostBWUsage/1000);
		
		for ( EventType pt : queueSize.keySet() ) {				
			if ( !this.queueSize.containsKey(pt) )
				this.queueSize.put(pt, new LinkedList<Float>());
			
			this.queueSize.get(pt).add(queueSize.get(pt)/1f);
		}

		for ( EventType pt : activeProbes.keySet() ) {				
			if ( !this.activeProbes.containsKey(pt) )
				this.activeProbes.put(pt, new LinkedList<Float>());
			
			this.activeProbes.get(pt).add(activeProbes.get(pt)/1f);
		}

		Float minutelly = 0f;
		for ( EventType pt : minutelyAllocated.keySet() ) {				
			if ( !this.minutelyAllocated.containsKey(pt) )
				this.minutelyAllocated.put(pt, new LinkedList<Float>());
						
			this.minutelyAllocated.get(pt).add(minutelyAllocated.get(pt)/1f/pt.getConsumedBW());
			minutelly+=minutelyAllocated.get(pt)/1f;
		}
		this.minutellyAllocatedTotal.add(minutelly);

		if ( Main.DEBUG) {
		
			// flush to stdout			
			LOGGER.info("PERPROBEPROFILER: " + new Date() + " " + dasuCpuUsage + " " + hostBWUsage + " queue: " + 
					queueSize + " active: " + activeProbes + " minutelyAllocated: " + minutelyAllocated +
					" probeDelayedDueToBWLimit: " + probeDelayedDueToBWLimit );
			
			LinePlot plot_cpu = null;
			plot_cpu = new LinePlot("profiler.scheduler.cpu.vuze.plot",
	                "cpu", timeStamps, this.dasuCpuUsage, "(time)", "(utilization)", LinePlot.FORMAT_TIME);
			DasuData.getInstance().update(plot_cpu.getId(), plot_cpu.generatePlot());
	
			LinePlot plot_bw = null;
			plot_bw = new LinePlot("profiler.scheduler.system.bw.plot",
	                "bw", timeStamps, this.hostBWUsage, "(time)", "(bw)", LinePlot.FORMAT_TIME);
			DasuData.getInstance().update(plot_bw.getId(), plot_bw.generatePlot());
	
			LinePlot plot_minutelly = null;
			plot_minutelly = new LinePlot("profiler.scheduler.minutely.allocated.TRACEROUTE.plot",
	                "minutely", timeStamps, this.minutelyAllocated.get(ProbeType.TRACEROUTE), "(time)", "(probes)", LinePlot.FORMAT_TIME);
			plot_minutelly.addSeries("queued", this.queueSize.get(ProbeType.TRACEROUTE));
			plot_minutelly.addSeries("active", this.activeProbes.get(ProbeType.TRACEROUTE));
			DasuData.getInstance().update(plot_minutelly.getId(), plot_minutelly.generatePlot());
			
			plot_minutelly = new LinePlot("profiler.scheduler.minutely.allocated.PING.plot",
	                "minutely", timeStamps, this.minutelyAllocated.get(ProbeType.PING), "(time)", "(probes)", LinePlot.FORMAT_TIME);
			plot_minutelly.addSeries("queued", this.queueSize.get(ProbeType.PING));
			plot_minutelly.addSeries("active", this.activeProbes.get(ProbeType.PING));
			DasuData.getInstance().update(plot_minutelly.getId(), plot_minutelly.generatePlot());
			
			plot_minutelly = new LinePlot("profiler.scheduler.minutely.allocated.all.plot",
	                "minutely", timeStamps, this.minutellyAllocatedTotal, "(time)", "(bw units)", LinePlot.FORMAT_TIME);
			DasuData.getInstance().update(plot_minutelly.getId(), plot_minutelly.generatePlot());
		}
		
	}
	
	public void logDelayDueToCpuLimit(EventType pt) {
		if ( !probeDelayedDueToCpuLimit.containsKey(pt) )
			probeDelayedDueToCpuLimit.put(pt, new LinkedList<Long>());
		
		probeDelayedDueToCpuLimit.get(pt).add(System.currentTimeMillis());		
	}
	
	public void logDelayDueToBWLimit(EventType pt) {
		if ( !probeDelayedDueToBWLimit.containsKey(pt) )
			probeDelayedDueToBWLimit.put(pt, new LinkedList<Long>());
		
		probeDelayedDueToBWLimit.get(pt).add(System.currentTimeMillis());
	}
		
	@SuppressWarnings("rawtypes")
	private ProbeProfile getProbeProfile(DasuSubTask pst) {
		
		if ( !probes.containsKey(pst.getType()) ) {
			probes.put(pst.getType(), new LinkedHashMap<Long, ProbeProfile>());			
		}
		
		ProbeProfile ppf = probes.get(pst.getType()).get(pst.getId());
		if ( ppf  == null )
			LOGGER.info("PROFILER: probe doesn't exist! " + pst.getType() + " " + pst.getId() + " " + pst);
		return ppf;
		
	}
	
	@SuppressWarnings("rawtypes")
	public void logProbeRequested(DasuSubTask pst) {
				
		if ( !probes.containsKey(pst.getType()) ) {
			
			probes.put(pst.getType(), new LinkedHashMap<Long, ProbeProfile>(){
				int MAX_ENTRIES = MAXNUMPROBES; 
				protected boolean removeEldestEntry(Map.Entry eldest) {
					return size() > MAX_ENTRIES;
				}
			});
		}
		
		probes.get(pst.getType()).put(pst.getId(), new ProbeProfile(pst.getType(), pst.getId()));
						
	}
	
	@SuppressWarnings("rawtypes")
	private void logProbeLaunched(DasuSubTask pst) {
		
		ProbeProfile ppf = getProbeProfile(pst);
		if ( ppf  != null ) {
			
			
			long timeStamp = -1;
			if ( pst instanceof ProbeSubTask ) 
				if (((ProbeSubTask)pst).getExecutionPathTimestamps().containsKey("probeLaunchTime"))
					timeStamp = (Long) ((ProbeSubTask) pst).getExecutionPathTimestamps().get("probeLaunchTime");
				
			ppf.logLaunched(timeStamp);
		}
	}
	
	@SuppressWarnings("rawtypes")
	public void logProbeCompleted(DasuSubTask pst) {
		
		//extract launch timestamp info from task
		logProbeLaunched(pst);
		
		ProbeProfile ppf = getProbeProfile(pst);
		if ( ppf  != null ) {
			
			long timeStamp = -1;
			if ( pst instanceof ProbeSubTask ) 
				if (((ProbeSubTask)pst).getExecutionPathTimestamps().containsKey("probeCompletionTime"))
					timeStamp = (Long) ((ProbeSubTask) pst).getExecutionPathTimestamps().get("probeCompletionTime");
				
			ppf.logCompleted(timeStamp);
		}
		
		//System.out.println(ppf);
	}
	
	@SuppressWarnings("rawtypes")
	public void logProbeDelayed(DasuSubTask pst) {
		
		ProbeProfile ppf = getProbeProfile(pst);
		if ( ppf  != null ) {
			ppf.logDelayed();
		} 
				
	}
	
	@SuppressWarnings("rawtypes")
	public void logProbeDenied(DasuSubTask pst) {
		
		ProbeProfile ppf = getProbeProfile(pst);
		if ( ppf  != null ) {
			ppf.logDenied();
		} 
	}
	
	class ProbeProfile {
		
		EventType EventType;
		Long probeId;
		
		long requested = -1;
		long launched = -1;
		long completed = -1;
		long denied = -1;
		LinkedList<Long> delayed = new LinkedList<Long>();
		
		ProbeProfile(EventType EventType, Long probeId) {
			this.EventType = EventType;
			this.probeId = probeId;
			
			requested = System.currentTimeMillis();
		}
		
		public void logLaunched(long timeStamp) {
			if (launched == -1) {
				
				if (timeStamp == -1)
					timeStamp = System.currentTimeMillis();

				launched = timeStamp;
				
			}
		}
		
		public void logCompleted(long timeStamp) {
			if (completed == -1) {
				
				if (timeStamp == -1)
					timeStamp = System.currentTimeMillis();
				completed = timeStamp;
				
			}
		}
		
		public void logDelayed() {
			delayed.add(System.currentTimeMillis());
		}
		
		public void logDenied() {
			if (denied == -1)
				denied = System.currentTimeMillis();
		}
		
		@Override
		public String toString() {
			
			return EventType + " id: " + probeId + " requested: " + new Date(requested) +
					" launched: " + new Date(launched) + " completed: " + new Date(completed) + " denied: " + denied +
					" requtest-to-launch: " + (launched-requested)/1000 + " launch-to-complete " + (completed-launched)/1000 + 
					" delayed: " +	delayed;
		}
		
	}
}
