/**
 * Dasu Project
 *
 * File:         ResourceProfiler.java
 * RCS:          $Id: ResourceProfiler.java,v 1.0 Nov 16, 2011 10:59:51 AM mas939 Exp $
 * Description:  ResourceProfiler class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Nov 16, 2011 at 10:59:51 AM
 * Language:     Java
 * Package:      edu.northwestern.dasu.profilers
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.profilers;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.DasuData;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.drools.ReportableFact;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.ui.LinePlot;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The ResourceProfiler class ... compute cpu usage on per-thread basis and memory usage.
 */

public class ResourceProfiler implements Runnable {
		
	protected long profileInterval = -1;
	protected long limitUtilizationInterval = 180; //seconds
	protected float maxAllowedUtilization = 0.40f; //15% max cpu utilization
	protected long maxNumTimeEntriesPerThread = -1; 
	protected List<String> threadNameFilter = new LinkedList<String>();	
	protected long totalPhysicalMemory = -1;
	
	protected boolean active = false;
	protected static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");	
	
	protected Hashtable<Long, ThreadCpuInfo> threadIdsMap = new Hashtable<Long, ThreadCpuInfo>();
	protected Hashtable<String, Long> threadNamesToIds = new Hashtable<String, Long>();
	
	protected LinkedList<Float> jvmCpuUsage = new LinkedList<Float>();
	protected LinkedList<Float> dasuCpuUsage = new LinkedList<Float>();
	protected LinkedList<Long> timeAxis = new LinkedList<Long>();	
	protected LinkedList<Long> totalMemoryUsage = new LinkedList<Long>(); 
	
	protected ThreadMXBean threadBean = ManagementFactory.getThreadMXBean();		

	protected long prevUpdate = System.currentTimeMillis();
	protected long prevUIUpdate = System.currentTimeMillis();
	protected long prevReportUpdate = System.currentTimeMillis();
	
	protected long startTime = System.nanoTime();	
	protected long prevUpTime = startTime;
	protected long upTime = 0;
	
	protected static boolean DEBUG = true;
	private final Integer NANO = 1000000000; 
	
	//DEBUG
	protected HashSet<Long> toPlot = new HashSet<Long>();
    
	public ResourceProfiler(long profileInterval, long limitUtilizationInterval, float maxAllowedCpuUtilization,
			String[] threadNameFilterList) {
		
		this.profileInterval = profileInterval;
		this.limitUtilizationInterval = limitUtilizationInterval > this.limitUtilizationInterval 
				? limitUtilizationInterval : this.limitUtilizationInterval;
		this.maxAllowedUtilization = maxAllowedCpuUtilization < this.maxAllowedUtilization 
				? maxAllowedCpuUtilization : this.maxAllowedUtilization;
		
		threadNameFilter.addAll(Arrays.asList(threadNameFilterList));
		maxNumTimeEntriesPerThread = 2*limitUtilizationInterval/profileInterval; //twice the number of entries
				
	}
	
	public LinkedList<Long> getMemoryUsage() {
		return totalMemoryUsage;
	}
	
	public Float getAverageMemoryUsageMB() {
		
		long usage = 0; 
		for ( long m : totalMemoryUsage )
			usage += m;
		
		return ((float)(usage)/totalMemoryUsage.size());
		
	}
	
	public Float getAverageMemoryUsage() {
		
		return getAverageMemoryUsageMB()/this.totalPhysicalMemory;		
		
		
	}
	
	private void updateMemoryUsage() {
		
        int mb = 1024*1024;
        
        //Getting the runtime reference from system
        Runtime runtime = Runtime.getRuntime();
        //totalMemoryUsage.add(runtime.totalMemory()/mb);
        
        totalMemoryUsage.add((runtime.totalMemory() - runtime.freeMemory())/mb);
        
        if ( totalMemoryUsage.size() > maxNumTimeEntriesPerThread)
        	totalMemoryUsage.pop();
        
	}
	
	public void run() {
		
		StringBuilder sb = new StringBuilder();
		sb.append("\n*******\n");
		sb.append("limitUtilizationInterval: " + limitUtilizationInterval + " seconds\n");
		sb.append("profileInterval: " + profileInterval + " seconds\n");
		sb.append("maxAllowedCpuUtilization: " + maxAllowedUtilization);
		sb.append("*******\n");
		LOGGER.info("PROFILER: " + sb.toString());
		
		if (!threadBean.isThreadCpuTimeSupported()) {
					
			LOGGER.info("PROFILER: ThreadCpuTime is not supported!!");
			ReportableFact rf = new ReportableFact("ErrorEvent");
			rf.addContentsObject("event", "ThreadCpuTime not supported!");
			commitReportableFact(rf);			
			
			return;
		}
		
		try {
		    com.sun.management.OperatingSystemMXBean osBean = 
		    		(com.sun.management.OperatingSystemMXBean) java.lang.management.ManagementFactory.getOperatingSystemMXBean();
		    totalPhysicalMemory = osBean.getTotalPhysicalMemorySize()/1024/1024;		    
		    LOGGER.info("PROFILER: totalPhysicalMemory: " + totalPhysicalMemory);
		    
		    /*
		    System.out.println("getArch: " + osBean.getArch());
		    System.out.println("getAvailableProcessors: " + osBean.getAvailableProcessors());
		    System.out.println("getCommittedVirtualMemorySize: " + osBean.getCommittedVirtualMemorySize()/1024/1024);
		    System.out.println("getFreePhysicalMemorySize: " + osBean.getFreePhysicalMemorySize()/1024/1024);
		    System.out.println("getFreeSwapSpaceSize: " + osBean.getFreeSwapSpaceSize()/1024/1024);
		    System.out.println("getName: " + osBean.getName());
		    System.out.println("getProcessCpuTime: " + osBean.getProcessCpuTime());
		    System.out.println("getSystemLoadAverage: " + osBean.getSystemLoadAverage());
		    System.out.println("getTotalPhysicalMemorySize: " + osBean.getTotalPhysicalMemorySize()/1024/1024);
		    System.out.println("getTotalSwapSpaceSize: " + osBean.getTotalSwapSpaceSize()/1024/1024);
		    System.out.println("getVersion: " + osBean.getVersion());
		    System.out.println("totalMemory: " + Runtime.getRuntime().totalMemory()/1024/1024);
		    System.out.println("freeMemory: " + Runtime.getRuntime().freeMemory()/1024/1024);
		    System.out.println("maxMemory: " + Runtime.getRuntime().maxMemory()/1024/1024);
		    */
		    
		}
		catch (Exception e) {
			LOGGER.info("PROFILER: " + e.getMessage());
			LOGGER.info("PROFILER: OperatingSystemMXBean is not supported!!");
			ReportableFact rf = new ReportableFact("ErrorEvent");
			rf.addContentsObject("event", "OperatingSystemMXBean not supported!");
			commitReportableFact(rf);
		}
			
		active = true;	
		
		try {
	    
			while (!Main.isShuttingDown() && isActive()) {
			  				
				Long time = System.currentTimeMillis();

				//ignore the first 60 seconds...
				if ( upTime/NANO > 60 ) {

					try {
						updateCurrentCpuUsage(time);
						updateMemoryUsage();
						
					}
					catch (Exception e) {
						e.printStackTrace();
					}
					
					if ( DEBUG ) {
						if ( System.currentTimeMillis() - prevUpdate > 5*1000)
							debug(time, 5f);
					}
					
					if ( DasuConfiguration.getInstance().isResourceProfilerReportsEnabled() ) {
						
						if ( System.currentTimeMillis() - prevReportUpdate > 
								DasuConfiguration.getInstance().getResourceProfilerReportTimeInterval()*1000)
							sendReport(time, DasuConfiguration.getInstance().getResourceProfilerReportUtilizationLimit()*100);
					}
					
					//update thread list
					if ( System.currentTimeMillis() - prevUpdate > 5*1000) {
						cleanList();
						prevUpdate = System.currentTimeMillis();
					}
									
					
					int numEntries = new Long(limitUtilizationInterval/profileInterval).intValue();
					//if (DEBUG)
					//	LOGGER.info("PROFILER: size: " + dasuCpuUsage.size() + " => " + dasuCpuUsage);
						
					if ( dasuCpuUsage.size() > numEntries ) {
						
						float sum = 0;
						for ( int i = dasuCpuUsage.size()-1; i >= dasuCpuUsage.size() - numEntries; i-- )					
							sum += dasuCpuUsage.get(i);
						
						float usage = sum/numEntries;
						
						if (cpuExceededMaxAllowedUtiliation(usage)) {
							LOGGER.severe("PROFILER: Dasu CPU utilization exceeded predefined limit, unloading... usage: " + 
									usage + ", limit: " + maxAllowedUtilization);
							LOGGER.severe("PROFILER: dasuCpuUsage: " + dasuCpuUsage);
							initiateShutdown("Dasu CPU utilization exceeded predefined limit");
						}
						
						if (memoryExceededMaxAllowedUtiliation()) {
							LOGGER.severe("PROFILER: Dasu Memory utilization exceeded predefined limit, unloading...");
							initiateShutdown("Dasu Memory utilization exceeded predefined limit");						
						}
												
					}
										
				}
				else {
					
					if (DEBUG)
						LOGGER.info("PROFILER: client uptime: " + upTime/NANO + " seconds, wait to start profiling...");
					
					upTime = getUptime();
				}
				
				Thread.sleep(profileInterval*1000);
			}
		  
		} catch (InterruptedException iex) {
			Thread.currentThread().interrupted();
			throw new RuntimeException("Profiler interrupted");
		}
		
	}
	
	protected boolean cpuExceededMaxAllowedUtiliation(float usage) {		
		//if average usage exceeded maxAllowedUtilization in the past interval, shutdown
		return ( usage > maxAllowedUtilization*100);
	}

	protected boolean memoryExceededMaxAllowedUtiliation() {		
		//if average memory exceeded 15% in the past interval, shutdown
		return ( totalPhysicalMemory != -1 && getAverageMemoryUsageMB() > totalPhysicalMemory*maxAllowedUtilization*2  );
	}
	
	protected void initiateShutdown(String reason) {
		
		sendReport(System.currentTimeMillis(), DasuConfiguration.getInstance().getResourceProfilerReportUtilizationLimit());
		
		ReportableFact rf = new ReportableFact("WatchdogTimer");
		rf.addContentsObject("jvmCpuUsage", DasuManager.getResourceProfiler().getJvmCpuUsageList());
		rf.addContentsObject("dasuCpuUsageList", DasuManager.getResourceProfiler().getDasuCpuUsageList());
		rf.addContentsObject("dasuMemoryUsageList", DasuManager.getResourceProfiler().getMemoryUsage());
		rf.addContentsObject("report", "DasuShutdown");
		rf.addContentsObject("reason", reason);
		rf.addContentsObject("profileInterval", DasuConfiguration.getInstance().getResourceProfilerRecordInterval());
		rf.addContentsObject("limitUtilizationInterval", DasuConfiguration.getInstance().getResourceProfilerLimitUtilizationInterval());
		rf.addContentsObject("maxAllowedCpuUtilization", maxAllowedUtilization*100);
		rf.addContentsObject("maxAllowedMemoryUtilization", totalPhysicalMemory*maxAllowedUtilization*2);
		commitReportableFact(rf);
			
		Main.startUnloadConfigTimer();
	}
	
	protected void commitReportableFact(ReportableFact rf) {
		Statistics.getInstance().addReportableFact(rf);
	}
	
	private void cleanList() {
		
		HashSet<Long> activeThreads = new HashSet<Long>();		
		for ( long t : threadBean.getAllThreadIds())
			activeThreads.add(t);
		
		threadIdsMap.keySet().retainAll(activeThreads);		
	}
	
	private LinkedList<Float> fillPlotGap(ThreadCpuInfo tInfo, LinkedList<Long> timeAxis) {
		
		LinkedList<Long> timeStamp = tInfo.getTimeStamp();
		LinkedList<Float> vals = tInfo.getCpuUsageList();
		LinkedList<Float> ret = new LinkedList<Float>();
		
		for ( Long l : timeAxis ) {
			
			int index = timeStamp.indexOf(l);
			if ( index != -1 ) {
				ret.add(vals.get(index));
			}
			else
				ret.add(0f);			
		}
		
		return ret;
		
	}
	
	//update the per-thread cpu usage (for all threads)
	private void updateCurrentCpuUsage(long time) {
			
		upTime = getUptime();		
		updateCpuTime(time);

    	long elapsedCpuTimeTotal = 0;
    	long elapsedCpuTimeDasu = 0;
        long elapsedTime = upTime - prevUpTime;

        if (prevUpTime > 0L && upTime > prevUpTime) {
        	
        	for ( Long threadId : threadIdsMap.keySet() ) {
        		
        		elapsedCpuTimeTotal += threadIdsMap.get(threadId).getElapsedCpuTime(time);
        		
        		boolean filter = true;
        		for ( String threadFilter : threadNameFilter ) {        			
        			if ( threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains(threadFilter) )
        				filter = false;
        		}
        		
        		if ( threadNameFilter.size() == 0 || !filter)
        			elapsedCpuTimeDasu += threadIdsMap.get(threadId).getElapsedCpuTime(time);
        		
        	} 	
        }
        
        if ( jvmCpuUsage.size() > maxNumTimeEntriesPerThread) {
        	jvmCpuUsage.pop();
        	dasuCpuUsage.pop();
        	timeAxis.pop();
        }
        
        float totalCpuUsage = ((float)elapsedCpuTimeTotal / ((float)elapsedTime))*100;
        jvmCpuUsage.add(totalCpuUsage);
        
        float dasuTotalCpuUsage = ((float)elapsedCpuTimeDasu / ((float)elapsedTime))*100;
        dasuCpuUsage.add(dasuTotalCpuUsage);
        
        prevUpTime = upTime;
	}
	
	private HashMap<Long, Thread> getAllThreads() {
		
		HashMap<Long, Thread> ret = new HashMap<Long, Thread>();
		ThreadGroup rootGroup = Thread.currentThread( ).getThreadGroup( );
		ThreadGroup parentGroup;
		while ( ( parentGroup = rootGroup.getParent() ) != null ) {
		    rootGroup = parentGroup;
		}
		
		Thread[] threads = new Thread[ rootGroup.activeCount() ];
		while ( rootGroup.enumerate( threads, true ) == threads.length ) {
		    threads = new Thread[ threads.length * 2 ];
		}
		
		for ( Thread t : threads) {
			
			if ( t != null)
				ret.put(t.getId(), t);
		}
		
		return ret;		
	}
	
	//update the per-thread cpu timing (for all threads)
	private void updateCpuTime(Long time) {
				
		timeAxis.add(time);		
		HashMap<Long, Thread> threads = getAllThreads();
		
		for ( long threadId : threads.keySet() ) {
            
			if (threadId != -1) {
            	    
            	if (!threadIdsMap.containsKey(threadId)) { 
            		threadIdsMap.put(threadId, new ThreadCpuInfo(threadId, threads.get(threadId)));
            		threadNamesToIds.put(threadIdsMap.get(threadId).getThread().getName().toLowerCase(), threadId);
            	}
                
            	threadIdsMap.get(threadId).addCpuTime(threadBean.getThreadCpuTime(threadId), time);
            }
        }
        
	}
	
	public long getUptime() {
		return System.nanoTime() - startTime;		
	}	
	
	public void stop() {
		
		LOGGER.info("STOP: stopping ResourceProfiler...");
		active = false;
	}
	
	public boolean isActive() {
		return active;
	}
	
	public LinkedList<Float> getDasuCpuUsageList() {
		return dasuCpuUsage;
	}
	
	public float getDasuCpuUsage() {
		if ( dasuCpuUsage.size() > 0)
			return dasuCpuUsage.getLast();
		
		return -1f;
	}
	
	public LinkedList<Float> getJvmCpuUsageList() {
		return jvmCpuUsage;
	}

	public float getJvmCpuUsage() {
		
		if ( jvmCpuUsage.size() > 0)
			return jvmCpuUsage.getLast();
		
		return -1f;
	}

	public float getCpuUsage(Long threadId) {
		
		if ( threadIdsMap.containsKey(threadId) && threadIdsMap.get(threadId).getCpuUsageList().size() > 0) {
				return 	threadIdsMap.get(threadId).getCpuUsageList().getLast();
		}
		
		return -1;
 		
	}
	
	public float getCpuUsage(String threadName) {
			
		if (threadNamesToIds.containsKey(threadName.toLowerCase())) {			
			return getCpuUsage(threadNamesToIds.get(threadName.toLowerCase()));
		}
		
		return -1;
	}
	
	public void sendReport(Long time, float pct) {
		
		prevReportUpdate = System.currentTimeMillis();
		HashSet<Long> toPlot = new HashSet<Long>();
		
		for ( Long threadId : threadIdsMap.keySet()) {		

			// report for dasu threads only
			//String tname = threadIdsMap.get(threadId).getThread().getName();
			//if ( !tname.toLowerCase().contains("dasu") )
			//	continue;
				
			//report if last measurement is over the limit
			if ( threadIdsMap.get(threadId).getCpuUsage(time) > pct ) {
				toPlot.add(threadId);
			}
			
			//report if the average is over the limit
			if ( threadIdsMap.get(threadId).getAverageCpuUsage() > pct ) {
				toPlot.add(threadId);
			}

		}
		
		if ( toPlot.size() == 0)
			return;
		
		ReportableFact rf = new ReportableFact("ResourceProfiler");
		rf.addContentsObject("uptime", getUptime());
		rf.addContentsObject("timeStamps", timeAxis);
		rf.addContentsObject("profileInterval", DasuConfiguration.getInstance().getResourceProfilerRecordInterval());
		rf.addContentsObject("reportUtilizationLimit", pct);
		rf.addContentsObject("reportTimeInterval", DasuConfiguration.getInstance().getResourceProfilerReportTimeInterval());
		rf.addContentsObject("dasuCpuUsage", getDasuCpuUsage());
		rf.addContentsObject("jvmCpuUsage", getJvmCpuUsage());
		String jvmName = threadNameFilter.contains("dasu") ? "vuze" : "ame";
		rf.getContentsMap().put("jvmName", jvmName);
		
		for ( Long threadId : threadIdsMap.keySet()) {
					
			if ( toPlot.contains(threadId)) {
				
				if (jvmCpuUsage.size() > 5) {
										
					String tname = threadIdsMap.get(threadId).getThread().getName();
					LinkedList<Float> signal = fillPlotGap(threadIdsMap.get(threadId), timeAxis);
					rf.getContentsMap().put(tname, signal);					
				}
			}
		}
		
		Statistics.getInstance().reportWithWrangler(rf);		
		
	}
	
    public void debug(Long time, float pct) {
    	
    	StringBuilder sb = new StringBuilder();
    	sb.append("\nPROFILER: client uptime: " + upTime/NANO + " seconds \n");
    	sb.append("PROFILER: jvm " + + jvmCpuUsage.getLast() + "%\n");
    	sb.append("PROFILER: dasu: " + dasuCpuUsage.getLast() + "%\n");
		
		LinePlot plot5 = null;
		float reaper_cpu = 0;
		int reaper_count = 0;
		
		for ( Long threadId : threadIdsMap.keySet()) {		

			if ( threadIdsMap.get(threadId).getCpuUsage(time) > pct ) {
				toPlot.add(threadId);
			}

			if ( (threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("dasu") 
				|| !threadNameFilter.contains("dasu") ) &&
					threadIdsMap.get(threadId).getCpuUsageList().size() > 0) {
				
				sb.append("PROFILER: *" + threadIdsMap.get(threadId).getThread().getName() + ": " +
						threadIdsMap.get(threadId).getCpuUsageList().getLast() + "\n");

			}
			
			if ( (threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("reaper") ||
					threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("ping") ||
					threadIdsMap.get(threadId).getThread().getName().toLowerCase().contains("trace") ) && 
					threadIdsMap.get(threadId).getCpuUsageList().size() > 0) {
				
				reaper_cpu += threadIdsMap.get(threadId).getCpuUsageList().getLast();
				reaper_count++;
				//System.out.println("reaper: " + " " + threadBean.getAllThreadIds().length);
					
			}
			
			if ( toPlot.contains(threadId)) {
				
				if (jvmCpuUsage.size() > 5) {
					
					String tname = threadIdsMap.get(threadId).getThread().getName();
					LinkedList<Float> signal = fillPlotGap(threadIdsMap.get(threadId), timeAxis);
					
					if ( plot5 == null) 
						 plot5 = new LinePlot("resource.profiler.cpu.gt5.plot", 
								 tname, timeAxis, signal, "(time)", "(utilization)", LinePlot.FORMAT_TIME);
					else
						plot5.addSeries(tname, signal);
										
				}
			}
			
		}
		
		//don't plot for external jvm
		if ( System.currentTimeMillis() - prevUIUpdate > 15*1000 && threadNameFilter.contains("dasu")) {
			
			if ( plot5 != null )
				DasuData.getInstance().update(plot5.getId(), plot5.generatePlot());
	
							
			if (jvmCpuUsage.size() > 5) {
				LinePlot plot = new LinePlot("resource.profiler.cpu.vuze.plot", "vuze", timeAxis, jvmCpuUsage, "(time)", "(utilization)", LinePlot.FORMAT_TIME);
				plot.addSeries("dasu", dasuCpuUsage);
				DasuData.getInstance().update(plot.getId(), plot.generatePlot());
			}
			
			prevUIUpdate = System.currentTimeMillis();
		}
			
		sb.append("PROFILER: reaper: " + reaper_cpu + "% count: " + reaper_count + "\n");
		sb.append("PROFILER: *******\n");
		LOGGER.info("PROFILER: " + sb);


    }
    
	private class ThreadCpuInfo {
		
		Long threadId;
		LinkedList<Long> cpuTime = new LinkedList<Long>();
		LinkedList<Long> cpuTimeTimeStamp = new LinkedList<Long>();		
		
		LinkedList<Float> cpuUsage = new LinkedList<Float>();
		LinkedList<Long> cpuUsageTimeStamp = new LinkedList<Long>();
		
		//ThreadInfo threadInfo;
		Thread thread;
		
		/*
		public ThreadCpuInfo(long id, ThreadInfo info) {
			this.threadId = id;			
			this.threadInfo = info;			
		}
		*/
		
		public ThreadCpuInfo(long id, Thread t) {
			this.threadId = id;
			this.thread = t;
		}
		
		public void addCpuTime(Long ctime, Long time) {
			
        	if ( cpuTime.size() == maxNumTimeEntriesPerThread) {
        		cpuTime.pop();
        		cpuTimeTimeStamp.pop();
        	}

			cpuTime.add(ctime);
			cpuTimeTimeStamp.add(time);
        	
        	updateCpuUsage();

		}
		
		/*
		public ThreadInfo getThreadInfo() {
			return threadInfo;
		}
		*/
		private Thread getThread() {
			return thread;
		}
		
		
		private long getElapsedCpuTime(long timeStamp) {

			if ( cpuTimeTimeStamp.size() > 1 ) {
				
				int index = -1;
				if ( cpuTimeTimeStamp.get(cpuTimeTimeStamp.size()-1).equals(timeStamp))
					index = cpuTimeTimeStamp.size()-1;
				else 
					index = cpuTimeTimeStamp.indexOf(timeStamp);
				
				if ( index != -1)
					return cpuTime.get(index) - cpuTime.get(index-1);
			}
			
			return 0l;		
			
		}
			
		private long getCpuTime(Long timeStamp) {
			
			if ( cpuTimeTimeStamp.size() > 0 ) {
				
				int index = -1;
				if ( cpuTimeTimeStamp.get(cpuTimeTimeStamp.size()-1).equals(timeStamp) )
					index = cpuTimeTimeStamp.size()-1;
				else
					index = cpuTimeTimeStamp.indexOf(timeStamp);
				
				if ( index != -1)
					return cpuTime.get(index);
			}
			
			return 0l;
			
		}
		
		private float getCpuUsage(Long timeStamp) {

			if ( cpuUsageTimeStamp.size() > 0) {
				
				int index = -1;
				if ( cpuUsageTimeStamp.get(cpuUsageTimeStamp.size()-1).equals(timeStamp) )
					index = cpuUsageTimeStamp.size()-1;
				else
					index = cpuUsageTimeStamp.indexOf(timeStamp);
				
				if ( index != -1)
					return cpuUsage.get(index);
			}
			
			return 0l;
			
		}
		
		public Float getAverageCpuUsage() {
			
			float sum = 0;
			for ( float usage : cpuUsage ) {
				sum += usage;
			}
			return sum/cpuUsage.size();
		}
		
		public Float getMeanCpuUsage() {
			
			LinkedList<Float> usage = (LinkedList<Float>) cpuUsage.clone();
			Collections.sort(usage);
			int index = usage.size()/2;
			return usage.get(index);
		}
		
		public LinkedList<Float> getCpuUsageList() {
			return cpuUsage;
		}
		
		public LinkedList<Long> getTimeStamp() {
			return cpuUsageTimeStamp;
		}
		
		private boolean updateCpuUsage() {
						
			long elapsedTime = upTime - prevUpTime;
			
	        if (prevUpTime > 0L && upTime > prevUpTime) {
				
	        	if ( cpuTime.size() > 1) {
					// elapsedCpu is in ns
		    		int lastIndex = cpuTime.size()-1;
		            long elapsedCpuTime = cpuTime.get(lastIndex) - cpuTime.get(lastIndex-1);
		            	
		        	if ( cpuUsage.size() == maxNumTimeEntriesPerThread) {		        		
		        		cpuUsage.pop();
		        		cpuUsageTimeStamp.pop();
		        	}
	
		            float usage = ((float)elapsedCpuTime / ((float)elapsedTime))*100;	        		
		    		cpuUsage.add(usage);
		    		cpuUsageTimeStamp.add(cpuTimeTimeStamp.get(lastIndex));					
		        
		        	return true;
				}
	        }
	        

			return false;
    		
		}
				
	}
}
