package edu.northwestern.dasu.security;

import java.security.KeyFactory;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.Signature;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.spec.X509EncodedKeySpec;
import java.util.ArrayList;
import java.util.Enumeration;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import edu.northwestern.dasu.DasuConfiguration;
import edu.northwestern.dasu.Main;

public class DigitalSignatureManager {
	
	//private static PublicKey pubKey = loadKey();
	static String sigAlg = "SHA1withDSA";
	private static ArrayList<X509Certificate> certs = loadCertificates();
	private final static java.util.logging.Logger LOGGER = java.util.logging.Logger.getLogger("edu.northwestern.dasu");
		  
	 private static boolean verify(String datafile, byte[] sigbytes, PublicKey pk) {
		 
		 try {
			 Signature sig = Signature.getInstance(sigAlg);		 
			 sig.initVerify(pk);
			 FileInputStream fis = new FileInputStream(datafile);
			 byte[] dataBytes = new byte[1024];
			 int nread = fis.read(dataBytes);
			 while (nread > 0) {
				 sig.update(dataBytes, 0, nread);
				 nread = fis.read(dataBytes);
			 };
			 return sig.verify(sigbytes);
		 }
		 catch (Exception e) {
			 return false;
		 }
	 		 
	 }

	 private static char[] hexChars = {
		  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		  'A', 'B', 'C', 'D', 'E', 'F'};

	 public static String byteArray2Hex(byte[] ba) {
		  StringBuffer sb = new StringBuffer();
		  for (int i = 0; i < ba.length; i++) {
		    int hbits = (ba[i] & 0x000000f0) >> 4;
		    int lbits = ba[i] & 0x0000000f;
		    sb.append("" + hexChars[hbits] + hexChars[lbits] + "");
		  }
		  return sb.toString();
	 }
	 
	 public static byte[] readSignatureFile(String filename) throws IOException {
		 
		 FileInputStream sigfis = new FileInputStream(filename);
		 byte[] sigToVerify = new byte[sigfis.available()]; 
		 sigfis.read(sigToVerify);
		 sigfis.close();
		 return sigToVerify;
		 
	 }
		 
	 public static byte[] getBytesFromFile(String filename) throws IOException {
		  
		  File file = new File(filename);
		  if (!file.isFile()) 
		      throw new IOException("Not a valid file to read "+file.getName());
			  
	      InputStream is = new FileInputStream(file);
	      long length = file.length();
	      byte[] bytes = new byte[(int)length];
	  
	      int offset = 0;
	      int numRead = 0;
	      while (offset < bytes.length
	             && (numRead=is.read(bytes, offset, bytes.length-offset)) >= 0) {
	          offset += numRead;
	      }
	  
	      if (offset < bytes.length)
	          throw new IOException("Could not completely read file "+file.getName());
	  
	      is.close();
	      return bytes;
	 }
	 
	 public static PublicKey getPublicKeyFromFile(String pubKeyFilename) throws Exception {
		 
		 X509EncodedKeySpec pubKeySpec = 
			 new X509EncodedKeySpec(getBytesFromFile(pubKeyFilename));
		 KeyFactory keyFactory = KeyFactory.getInstance("DSA");
		 PublicKey pubk = keyFactory.generatePublic(pubKeySpec);

		 return pubk;

	 }
	 
	 public static PublicKey getPublicKeyFromBytes(byte[] keyBytes) throws Exception {
		 
		 X509EncodedKeySpec pubKeySpec = 
			 new X509EncodedKeySpec(keyBytes);
		 KeyFactory keyFactory = KeyFactory.getInstance("DSA");
		 PublicKey pubk = keyFactory.generatePublic(pubKeySpec);

		 return pubk;

	 }
	 
	 public static boolean verify(String dataFile, String signatureFile) {

		 boolean result = false;
		 try {
			 
			 byte[] sigbytes;
			 sigbytes = readSignatureFile(signatureFile);
			 result = verify(dataFile, sigbytes);
			 
			 if (Main.AME_DEBUG_SKIP_DOWNLOAD) {
				 StringBuilder sb = new StringBuilder();
				 sb.append("AME: verify: " + dataFile + "\n" + signatureFile + "\n");
				 sb.append("AME: Signature(in hex):: " + byteArray2Hex(sigbytes) + "\n");
				 sb.append("Signature Verification Result = " + result + "\n");
				 LOGGER.info(sb.toString());
				 LOGGER.info("AME: certs: " + certs.size());
			 }

		} catch (IOException e) {
			LOGGER.warning("AME: No digital signature was provided for experiment...");			
		}
		 
		 return result;

	 }
	 
	 private static ArrayList<X509Certificate> loadCertificates() {
		 
		ArrayList<X509Certificate> certs = new ArrayList<X509Certificate>();
		try {

		    KeyStore ks = KeyStore.getInstance("JKS");
		    InputStream is = ClassLoader.getSystemResourceAsStream(DasuConfiguration.KEYSTORE_JKS);        
		    ks.load(is, DasuConfiguration.KEYSTORE_PASS.toCharArray());
			 			
	        Enumeration<String> en = ks.aliases();
	        for (; en.hasMoreElements(); ) {
	            String alias = (String)en.nextElement();
	            if (ks.isCertificateEntry(alias)) {
	           	 certs.add((X509Certificate) ks.getCertificate(alias));           	 
	            }
	        }
		}
		catch (KeyStoreException e) {
			e.printStackTrace();
			LOGGER.severe("ERROR: invalid keystore");
		} catch (NoSuchAlgorithmException e) {
			e.printStackTrace();
			LOGGER.severe("ERROR: invalid algorithm");
		} catch (CertificateException e) {
			e.printStackTrace();
			LOGGER.severe("ERROR: invalid certificate");
		} catch (IOException e) {
			e.printStackTrace();
			LOGGER.severe("ERROR: IO problem with keystore");
		}
		return certs;
	 }
	 
	 private static boolean verify(String datafile, byte[] sigbytes) {
		 
		boolean verified = false;
			 
		for ( X509Certificate certificate : certs ) {
						
			 verified = verify(datafile, sigbytes, certificate.getPublicKey());			
			 if (verified)
				 break;
		}
		
		 return verified;

	 }
	 
	 /*
		private static PublicKey loadKey() {
			
			 URL url = ClassLoader.getSystemResource("public-DSA.key");
			 if (url != null) {
			    
				 try {
			
				     InputStream is = new DataInputStream(url.openStream());

				     long length = 1024;
				     byte[] bytes = new byte[(int)length];
				  
				     int offset = 0; 
				     int numRead = 0;
				     while (offset < bytes.length
				             && (numRead=is.read(bytes, offset, bytes.length-offset)) >= 0) {
				          offset += numRead;
				     }

				     return getPublicKeyFromBytes(bytes);
		
				} catch (IOException e) {
		
					try {
						throw (new Exception("No Interferece Detection Rules file from jars"));
					} catch (Exception e1) {
						e1.printStackTrace();
					}
											
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			 
			 return null;

		 }
		 */
		
}

