/**
 * Dasu Project
 *
 * File:         InterferenceSimulatorV3.java
 * RCS:          $Id: InterferenceSimulatorV3.java,v 1.0 Jul 30, 2011 8:25:30 AM mas939 Exp $
 * Description:  InterferenceSimulatorV3 class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Jul 30, 2011 at 8:25:30 AM
 * Language:     Java
 * Package:      edu.northwestern.dasu.simulator
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2011, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.simulator;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;
import java.util.concurrent.TimeUnit;

import org.gudy.azureus2.plugins.download.Download;

import com.azureus.plugins.aznetmon.main.RSTPacketStats.RSTPacketData;

import edu.northwestern.dasu.DasuData;
import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.DasuData.DasuDataListener;
import edu.northwestern.dasu.drools.FactMeasurementDsd;
import edu.northwestern.dasu.drools.FactMeasurementLpsd;
import edu.northwestern.dasu.drools.FactMeasurementSystem;
import edu.northwestern.dasu.drools.FactTemplate;
import edu.northwestern.dasu.netstat.NetstatBWStats.NetstatBWData;
import edu.northwestern.dasu.stats.DownloadStatsDynamic;
import edu.northwestern.dasu.stats.DownloadStatsSummary;
import edu.northwestern.dasu.stats.DynamicLocalPeerStat;
import edu.northwestern.dasu.stats.RSTStat;
import edu.northwestern.dasu.stats.RemotePeerStatDynamic;
import edu.northwestern.dasu.stats.RemotePeerStatStatic;
import edu.northwestern.dasu.stats.StaticLocalPeerStat;
import edu.northwestern.dasu.timer.DasuTimerManager;
import edu.northwestern.dasu.timer.ITimer;
import edu.northwestern.dasu.timer.ITimerEvent;
import edu.northwestern.dasu.timer.ITimerEventPerformer;
import edu.northwestern.dasu.timer.TimerEntry;
import edu.northwestern.dasu.ui.LinePlot;
import edu.northwestern.dasu.upnp.UPnPBWStats.UPnPBWData;
import edu.northwestern.dasu.util.Pair;
import edu.northwestern.dasu.util.TextFile;

/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 * 
 *         The InterferenceSimulatorV3 class ...
 */
public class InterferenceSimulatorV3 {

	
	/***
	 * This version supports simulation for many of the basic facts and stats, but still lacks support for:
	 * 
	 *   1. staticUPnPInfo, 
	 *   2. remotePeerStatDynamic2, 
	 *   3. remotePeerStatStaticDasu, 
	 *   4. localPeerStatStaticDasu, 
	 *   5. 4stStatsDasu,
	 *   
	 *   
	 *   "dir" must be a directory containing individual filenames named after each dasu table to be replayed,
	 *	each file should contain comma separated fields... 
	 *	 		
	 *		completedTask.txt                 localPeerStatDynamicDasu.txt
	 *		configEventDasu.txt               localPeerStatDynamicDasu.txt.orig
	 *		dnsData.txt                       localPeerStatStaticDasu.txt
	 *		dnsTimeout.txt                    pingdata_dasu.txt
	 *		dnsmappings.txt                   remotePeerStatDynamic2.txt
	 *		downloadStatDynamicDasu.txt       remotePeerStatStaticDasu.txt
	 *		downloadStatSummaryDasu.txt       rstStatsDasu.txt
	 *		interference.txt                  staticUPnPInfo.txt
	 *		interferenceEvent.txt             upnpTest.txt
	 *		ipChanges.txt                     userReports.txt
	 *		
	 *	
	 *	mas939$ more upnpTest.txt 
	 *		65601790,1304271125592,24191,'tcp',0,43042
	 *		65601790,1304271145555,50120,'tcp',0,43042
	 *		65601790,1304271165460,6881,'tcp',0,43042
	 *		65601790,1304271185483,6910,'tcp',0,43042
	 *		65601790,1304274201107,24191,'tcp',0,43048
	 *		65601790,1304274220999,50120,'tcp',0,43048
	 *		65601790,1304274241086,6881,'tcp',0,43048
	 *		65601790,1304274261351,6910,'tcp',0,43048
	 *  
	 */
	
	private boolean active = false;
	private long simulationInitialTime = -1l;

	private int tickIncrementSeconds = 1;// default one
	private int tick = 0;
	private ITimer tickTimer;

	public static HashMap<Integer, DownloadStatsSummary> downloadSummary = new HashMap<Integer, DownloadStatsSummary>();
	public PriorityQueue<SimulatedEvent> eventQueue = new PriorityQueue<SimulatedEvent>();

	private HashMap<String, Class> classes = new HashMap<String, Class>();
	private HashMap<String, HashMap<String, ArrayList<Pair<Integer, String>>>> fields = new HashMap();
	private HashMap<String, Iterator<String>> iterators = new HashMap();

	private String dir = null;
	private int numTimeEpocsToLoadInAdvance = 5;
	
	private SimulatedEventList eventList = null;
	private Iterator<Collection<SimulatedEvent>> iterEventList = null;
	private Collection<SimulatedEvent> currEvents = null;
	
	HashMap<Long, PlotObject> factDsDPlotMap = new HashMap<Long, PlotObject>();
	
	List<String> factDsdFields = new ArrayList<String>();
	List<String> factDsdLabels = new ArrayList<String>();

	HashMap<String, List<String>> factFields = new HashMap<String, List<String>>();
	HashMap<String, List<String>> factLabels = new HashMap<String, List<String>>();
	HashMap<String, PlotObject> factPlots = new HashMap<String, PlotObject>();

	public InterferenceSimulatorV3(String dir, int tickIncrementSeconds) {

		/*****

			
		*****/

		
		this.dir = dir;
		setTickIncrementSeconds(tickIncrementSeconds);
		initialize();
		loadIterators();
		loadTimeline();
		setSimulationInitialTime();
		System.out.println(fields);
		active = true;
		
		DasuData.getInstance().registerListener("simulator.second.equivalence", new DasuDataListener(){

			public void updated(String key, Object val) {
				setTickIncrementSeconds(Integer.parseInt(val.toString()));
				
			}});
		DasuData.getInstance().markPersistent("simulator.second.equivalence");

		}
	
	public boolean isActive() {
		return active;
	}

	/*
	 * loads into the eventQueue queue 'numTimeEpocs' (ticks) events in the
	 * future
	 */
	private Collection<SimulatedEvent> loadEvents(String fileName, HashMap<String, ArrayList<Pair<Integer, String>>> typeFields) {

		ArrayList<SimulatedEvent> ret = new ArrayList<SimulatedEvent>();
		Long currTime = nextTickTime();
		long limit = numTimeEpocsToLoadInAdvance * tickIncrementSeconds;

		while (iterators.get(fileName).hasNext()) {

			String line = iterators.get(fileName).next();
			String l[] = line.split(",");

			if (l.length < 3)
				continue;

			Long currTimeStamp = -1l;
			for (String type : typeFields.keySet()) {

				SimulatedEvent event = new SimulatedEvent(type,	classes.get(type));
				for (Pair<Integer, String> pair : typeFields.get(type)) {

					Integer pos = pair.getKey();
					String field = pair.getValue();

					if (field.equals("timeRecorded") || field.equals("timestamp"))
						event.timeStamp = Long.parseLong(l[pos]);
					//else
					event.fields.put(field, l[pos]);
				}

				ret.add(event);
				currTimeStamp = event.timeStamp;
			}

			//System.out.println("loadEvents(): lastTimeStamp: " + new Date(currTimeStamp) + " currTime: " + new Date(currTime) + " difference: " 
			//		+ (currTimeStamp - currTime) / 1000 + " limit: " + limit + " " + typeFields.keySet() + "!!!");

			// load 100 elements from each table if the time has not been corrected yet
			if (simulationInitialTime == -1) {
				if (ret.size() % 4000 == 0) {
					break;					
				}
			}

			if (currTimeStamp - currTime > limit * 1000) {
				System.out.println(new Date(currTimeStamp) + " " + new Date(currTime) + " break!!!");
				break;
			}

		}

		return ret;

	}

	private Collection<SimulatedEvent> getCurrStatsHelper(String statsName, boolean peek) {

		String name = null;
		Long currTime = nextTickTime();

		if (statsName != null) {
			name = statsName.contains(".") ? statsName.substring(statsName
					.lastIndexOf(".") + 1) : statsName;
			name = name.contains("$") ? name
					.substring(name.lastIndexOf("$") + 1) : name;
		}

		ArrayList<SimulatedEvent> events = new ArrayList<SimulatedEvent>();
		ArrayList<SimulatedEvent> eventsToPutBack = new ArrayList<SimulatedEvent>();

		if (eventQueue.size() > 0) {

			try {
				System.out.println("getCurrStatsHelper(): " + name + " " + currTime + " " + eventQueue.peek().timeStamp);
				while (eventQueue.size() > 0
						&& eventQueue.peek().timeStamp.compareTo(currTime) < 0) {

					SimulatedEvent event = eventQueue.remove();
					if (((name != null) && (event.type.contains(name)))
							|| (name == null))
						events.add(event);
					else
						eventsToPutBack.add(event);

				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		eventQueue.addAll(eventsToPutBack);

		if (peek)
			eventQueue.addAll(events);

		return events;
	}

	private Collection<SimulatedEvent> peekCurrStats(String statsName) {

		return getCurrStatsHelper(statsName, true);
	}

	protected static Collection<SimulatedEvent> peekStats(String statsName, Collection<SimulatedEvent> allEvents) {

		ArrayList<SimulatedEvent> ret = new ArrayList();
		if ( allEvents != null) {
			for (SimulatedEvent event : allEvents) {
	
				if (event.type.contains(statsName))
					ret.add(event);
			}
		}

		return ret;
	}
	
	protected static Collection<SimulatedEvent> pollStats(String statsName, Collection<SimulatedEvent> allEvents) {

		ArrayList<SimulatedEvent> ret = new ArrayList();
		Iterator<SimulatedEvent> iter = allEvents.iterator();
		while ( iter.hasNext() ) {
		
			SimulatedEvent event = iter.next();
			if (event.type.contains(statsName)) {
				ret.add(event);
				iter.remove();
			}
		}

		return ret;
	}

	
	protected static Download[] getDownloads(Collection<SimulatedEvent> allEvents) {

		// obtain the summaries first!
		ArrayList<SimulatedEvent> events = (ArrayList<SimulatedEvent>) peekStats("DownloadStatsSummary", allEvents);
		Iterator<SimulatedEvent> iter = events.iterator();
		while (iter.hasNext()) {
			SimulatedEvent event = iter.next();
			try {
				int did = Integer.parseInt(event.fields.get("downloadId"));
				DownloadStatsSummary dss = (DownloadStatsSummary) createObject(event);
				if (!downloadSummary.containsKey(did)) {
					downloadSummary.put(did, dss);
				} else {

					if (downloadSummary.get(did).timeRecorded < dss.timeRecorded)
						downloadSummary.put(did, dss);
					else {
						System.out
								.println("SIMULATOR: Something odd... DownloadStatsSummary in wrong order!\n\n");
					}

				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		// now obtain dynamic vals
		events = (ArrayList<SimulatedEvent>) peekStats("DownloadStatsDynamic", allEvents);
		iter = events.iterator();

		ArrayList<SimulatedDownload> downloads = new ArrayList<SimulatedDownload>();

		while (iter.hasNext()) {
			SimulatedEvent event = iter.next();
			
			//System.out.println("????????????????\t" + event.timeStamp + "\t" + new Date(event.timeStamp) + "\t" + event.type);
			
			try {
				int did = Integer.parseInt(event.fields.get("downloadId"));
				SimulatedDownload dld = new SimulatedDownload(did);

				if (!downloadSummary.containsKey(did))
					System.out
							.println("SIMULATOR: Summary has not appeared yet! for downloadId: "
									+ did + " what do I do, discard?");
				else {

					// System.out.println(event);
					if (downloadSummary.get(did).timeRecorded > event.timeStamp) {
						System.out
								.println("SIMULATOR: Something odd... DownloadStatsStatic timestamp before Summary!");
					} else {

						/* DownloadStatsSummary */
						DownloadStatsSummary dls = downloadSummary.get(did);
						dld.getTorrent().pieceSize = dls.pieceSize;
						dld.getTorrent().size = dls.size;

						long timeDiffSec = (event.timeStamp - dls.timeRecorded) / 1000;
						dld.getStats().downloadingTime = timeDiffSec + dls.downloadingTime;
						dld.getStats().seedingTime = timeDiffSec + dls.seedingTime;

						dld.setMaximumDownloadKBPerSecond(dls.maxDownloadRateLimit);
						dld.setUploadRateLimitBytesPerSecond(dls.maxUploadLimit);

						/* DownloadStatsDynamic */
						dld.getStats().totalUploaded = Long.parseLong(event.fields.get("totalUploaded"));
						dld.getStats().totalDownloaded = Long.parseLong(event.fields.get("totalDownloaded"));
						dld.priority = Integer.parseInt(event.fields.get("priority"));

						dld.getStats().availability = Float.parseFloat(event.fields.get("availability"));
						dld.getStats().completed = Integer.parseInt(event.fields.get("completed"));
						dld.getStats().discarded = Integer.parseInt(event.fields.get("discarded"));
						dld.getStats().hashFails = Integer.parseInt(event.fields.get("hashFails"));
						dld.getStats().downloadRate = Integer.parseInt(event.fields.get("downloadRate"));
						dld.getStats().shareRatio = Integer.parseInt(event.fields.get("shareRatio"));
						dld.getStats().uploadRate = Integer.parseInt(event.fields.get("uploadRate"));
						dld.getLastAnnounceResult().numKnownNonSeeds = Integer.parseInt(event.fields.get("numKnownNonSeeds"));
						dld.getLastAnnounceResult().numKnownSeeds = Integer.parseInt(event.fields.get("numKnownSeeds"));
						dld.getLastScrapeResult().totalNumKnownNonSeeds = Integer.parseInt(event.fields.get("totalNumKnownNonSeeds"));
						dld.getLastScrapeResult().totalNumKnownSeeds = Integer.parseInt(event.fields.get("totalNumKnownSeeds"));
						dld.state = Integer.parseInt(event.fields.get("state"));
						dld.position = Integer.parseInt(event.fields.get("position"));
						dld.swarmPeerSpeed = Long.parseLong(event.fields.get("swarmPeerSpeed"));
						// dld.numPendingPeers =
						// Integer.parseInt(event.fields.get("numPendingPeers"));
						dld.getPeerManager().numPeers = Integer.parseInt(event.fields.get("numPeers"));
						dld.getPeerManager().getStats().connectedSeeds = Integer.parseInt(event.fields.get("connectedSeeds"));
						dld.getPeerManager().getStats().connectedLeechers = Integer.parseInt(event.fields.get("connectedLeechers"));

						downloads.add(dld);
					}

				}

			} catch (Exception e) {
				e.printStackTrace();
			}

		}

		return downloads.toArray(new Download[downloads.size()]);

	}

	protected static Object createObject(SimulatedEvent event) {

		Map<String, String> map = event.fields;
		Class c = event.className;

		if ( event.type.contains("Netstat"))
			System.out.println("NetstatBWData: USE SIMULATION!**** createObject(): " + event + " ");
		if ( event.type.contains("RST"))
			System.out.println("RSTPacketData: USE SIMULATION!**** createObject(): " + event + " ");

		
		if (map != null) {

			try {
				Object obj = c.getConstructor().newInstance();

				for (Field f : c.getFields()) {

					Object value = null;
					if (map.containsKey(f.getName())) {

						if ( event.type.contains("Netstat"))
							System.out.println("NetstatBWData: USE SIMULATION!**** createObject(): " + f.getName());
						if ( event.type.contains("RSTtat"))
							System.out.println("RSTPacketData: USE SIMULATION!**** createObject(): " + f.getName());
						
						String val = map.get(f.getName());

						if (val != null) {
							if (f.getType().equals(String.class))
								value = val;
							else if (f.getType().equals(Boolean.class)
									|| f.getType().toString().equals("boolean"))
								value = val.equals("true");
							else if (f.getType().equals(Float.class)
									|| f.getType().toString().equals("float"))
								value = Float.parseFloat(val);
							else if (f.getType().equals(Integer.class)
									|| f.getType().toString().equals("int"))
								value = Integer.parseInt(val);
							else if (f.getType().equals(Double.class)
									|| f.getType().toString().equals("double"))
								value = Double.parseDouble(val);
							else if (f.getType().equals(Long.class)
									|| f.getType().toString().equals("long"))
								value = Long.parseLong(val);
						}
						
						if ( event.type.contains("Netstat")) {
							System.out.println("NetstatBWData: USE SIMULATION!**** createObject(): " + f.getName() + " " + value);							
						}
						if ( event.type.contains("RST")) {
							System.out.println("RSTPacketData: USE SIMULATION!**** createObject(): " + f.getName() + " " + value);							
						}


					}

					try {

						if (value != null)
							f.set(obj, value);
						
					} catch (Exception e) {
						System.out.println("Field value assignment failed! "
								+ f.getName());
						e.printStackTrace();
					}

				}

				if ( event.type.contains("Netstat")) {
					System.out.println("NetstatBWData: USE SIMULATION!**** createObject(): set " + obj);							
				}
				return obj;

			} catch (Exception e) {
				e.printStackTrace();
			}

		}
		return null;

	}

	private void setSimulationInitialTime() {

		if (simulationInitialTime == -1 && eventQueue.size() > 0) {
			simulationInitialTime = eventQueue.peek().timeStamp;
		}

	}

	private void loadTimeline() {

		// filename, statsName, position, fieldName
		for (String fileName : fields.keySet()) {
			eventQueue.addAll(loadEvents(fileName, fields.get(fileName)));
		}
		// printQueue();

	}

	public void loadIterators() {

		Collection<String> files = fields.keySet();

		for (String fileName : files) {
			TextFile tf = new TextFile(dir + "/" + fileName);
			Iterator<String> iter = tf.iterator();
			iterators.put(fileName, iter);
		}
	}

	public void printQueue() {

		System.out.println("printQueue(): " + eventQueue.size() + "\n");

		ArrayList<SimulatedEvent> queue = new ArrayList<SimulatedEvent>();
		Long prev = 0l;
		while (eventQueue.size() > 0) {
			SimulatedEvent e = eventQueue.remove();
			System.out.print((e.timeStamp - prev) / 1000 / 60 + " "
					+ (e.timeStamp - prev) / 1000 + "\t " + e);
			prev = e.timeStamp;
			queue.add(e);
		}

		eventQueue.addAll(queue);

	}

	public long currentGMTTime() {

		long ret = -1;

		if (simulationInitialTime == -1) {			
			ret = System.currentTimeMillis();
			
		} else {
			
			ret = simulationInitialTime + (tickIncrementSeconds) * tick * 1000;
			
			if ( currEvents != null && currEvents.size() > 0 ) {
				Iterator<SimulatedEvent> iter = currEvents.iterator();
				SimulatedEvent event = iter.next();
				ret = event.timeStamp;
			}
		}
		
		return ret;
	}
	
	
	public long nextTickTime() {

		long ret = -1;

		if (simulationInitialTime == -1) 			
			ret = System.currentTimeMillis();			
		else 			
			ret = simulationInitialTime + (tickIncrementSeconds) * tick * 1000;			
		
		return ret;
		
	}

	private void setTickIncrementSeconds(int incrementSeconds) {

		System.out
				.println("SIMULATOR: tick increase! each second will represent "
						+ incrementSeconds + " seconds instead...");
		tickIncrementSeconds = incrementSeconds;
				
	}
	
	public static void startUiUpdater() {
		
		//DasuManager.interferenceSimulator.tickTimer = Main.createTimer("SimulatorUiTimer");
		//DasuManager.interferenceSimulator.tickTimer.addPeriodicEvent( 100,
		DasuTimerManager.getInstance().createEvent("SimulatorUiTimer", new TimerEntry(1000, true), 
				new ITimerEventPerformer() {

					public void perform(ITimerEvent event) {
						
						try {
							Long ret = DasuManager.interferenceSimulator.currentGMTTime();
							long millis = ret - DasuManager.interferenceSimulator.simulationInitialTime;
							String date = String.format("%d weeks, %d days, %d hour, %d min, %d sec",
									TimeUnit.MILLISECONDS.toDays(millis)/7,
									TimeUnit.MILLISECONDS.toDays(millis) - 7*(TimeUnit.MILLISECONDS.toDays(millis)/7),  		
									TimeUnit.MILLISECONDS.toHours(millis) - TimeUnit.DAYS.toHours(TimeUnit.MILLISECONDS.toDays(millis)),
									TimeUnit.MILLISECONDS.toMinutes(millis) - TimeUnit.HOURS.toMinutes(TimeUnit.MILLISECONDS.toHours(millis)),			    
								    TimeUnit.MILLISECONDS.toSeconds(millis) - TimeUnit.MINUTES.toSeconds(TimeUnit.MILLISECONDS.toMinutes(millis))
								);
							
							System.out.println(" ********************* " + date + " ******************");
							DasuData.getInstance().update("simulator.elapsed.time", date);
						}
						catch (Exception e) {
							e.printStackTrace();
						}
												
					}
		});		
	}


	public int getTickIncrementSeconds() {
		return tickIncrementSeconds;
	}

	public int getTick() {
		return tick;
	}

	public void increaseTick() {
		
		System.out.println("\n\n\n\n\n\n\n\n\n\n\n###########increaseTick(): " + new Date(System.currentTimeMillis()) + "############\n\n\n\n\n\n\n\n\n");
		tick++;
		loadTimeline();
		if ( tickTimer == null)
			startUiUpdater();
	}

	private SimulatedEventList getNextTickEvents() {

		Collection<SimulatedEvent> events = getCurrStatsHelper(null, false);
		SimulatedEventList e = new SimulatedEventList(events);
		System.out.print("\n\n\n\n" + e + "\n\n\n\n");
		return new SimulatedEventList(events);
	}
	
	/**************************************************/
	
	public void iterGetNextTickEvents() {		
		eventList = getNextTickEvents();
		iterEventList = eventList.iterator();		
	}

	public boolean iterHasNext() {
		return iterEventList.hasNext();
	}
	
	public void iterNext() {
		currEvents = iterEventList.next();
		if (currEvents == null)
			System.out.println("\n\n\n\niterNext(): WARNING, iterator is returning null!!\n\n\n\n\n");
	}
		
	public Download[] iterGetDownloads() {
		//System.out.println(currEvents);
		return getDownloads(currEvents);		
	}
		
	public Object iterGetStat(Object stat) {
		
		if ( stat instanceof Class) {

			String statsName = ((Class) stat).getName();
			String name = statsName.contains(".") ? statsName.substring(statsName
					.lastIndexOf(".") + 1) : statsName;
			name = name.contains("$") ? name
					.substring(name.lastIndexOf("$") + 1) : name;
			
			if ( currEvents != null) {
				Iterator<SimulatedEvent> iter = currEvents.iterator();
				while (iter.hasNext()) {
					
					SimulatedEvent event = iter.next();
					if ( event.type.equals(name)) {
						//iter.remove();  //peek don't poll
						System.out.println("????????????????\t" + event.timeStamp + "\t" + new Date(event.timeStamp) + "\t" + event.type);
						return createObject(event);
					}
				}
			}
		}
				
		return null;
	}
	
	public Collection<Object> iterGetStats() {
		
		LinkedList<Object> ret = new LinkedList<Object>();
		Iterator<SimulatedEvent> iter = currEvents.iterator();
		while ( iter.hasNext() ) {
			
			SimulatedEvent event = iter.next();
			Object obj = InterferenceSimulatorV3.createObject(event);
			if ( obj != null)
				ret.add(obj);
		}
		
		return ret;
	}
	
	/**************************************************/

	/*
	private void getNextTickStats() {
		statsList = new SimulatedStatsList(getNextTickEvents());
	}
	*/
	
	public SimulatedStatsList getNextTickStats() {

		SimulatedStatsList list = new SimulatedStatsList(getNextTickEvents());
		return list;

	}

	private void initialize() {

		classes.put("DownloadStatsDynamic", DownloadStatsDynamic.class);
		classes.put("DynamicLocalPeerStat", DynamicLocalPeerStat.class);
		classes.put("UPnPBWData", UPnPBWData.class);
		classes.put("RSTPacketData", RSTPacketData.class);
		classes.put("NetstatBWData", NetstatBWData.class);
		classes.put("RSTStat", RSTStat.class);
		classes.put("DownloadStatsSummary", DownloadStatsSummary.class);
		classes.put("StaticLocalPeerStat", StaticLocalPeerStat.class);
		classes.put("RemotePeerStatStatic", RemotePeerStatStatic.class);
		classes.put("RemotePeerStatDynamic", RemotePeerStatDynamic.class);

		fields.put("downloadStatDynamicDasu.txt", new HashMap());
		fields.put("downloadStatSummaryDasu.txt", new HashMap());
		fields.put("localPeerStatDynamicDasu.txt", new HashMap());
		// fields.put("rstStatsDasu.txt", new HashMap());
		// fields.put("localPeerStatStaticDasu.txt", new HashMap());
		// fields.put("remotePeerStatStaticDasu.txt", new HashMap());
		// fields.put("remotePeerStatDynamic2.txt", new HashMap());

		fields.get("downloadStatDynamicDasu.txt").put("DownloadStatsDynamic", new ArrayList());
		fields.get("downloadStatSummaryDasu.txt").put("DownloadStatsSummary", new ArrayList());
		fields.get("localPeerStatDynamicDasu.txt").put("DynamicLocalPeerStat", new ArrayList());
		fields.get("localPeerStatDynamicDasu.txt").put("RSTPacketData", new ArrayList());
		fields.get("localPeerStatDynamicDasu.txt").put("NetstatBWData", new ArrayList());
		fields.get("localPeerStatDynamicDasu.txt").put("UPnPBWData", new ArrayList());
		// fields.get("rstStatsDasu.txt").put("RSTStat", new ArrayList());
		// fields.get("localPeerStatStaticDasu.txt").put("StaticLocalPeerStat", new ArrayList());
		// fields.get("remotePeerStatStaticDasu.txt").put("RemotePeerStatStatic", new ArrayList());
		// fields.get("remotePeerStatDynamic2.txt").put("RemotePeerStatDynamic", new ArrayList());

		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(0, "peerId"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(1, "downloadId"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(2, "timeRecorded"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(3, "numPeers"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(4, "availability"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(5, "completed"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(6, "discarded"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(7, "downloadRate"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(8, "uploadRate"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(9, "shareRatio"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(10, "numKnownNonSeeds"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(11, "numKnownSeeds"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(12, "state"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(13, "priority"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(14, "position"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(15, "totalDownloaded"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(16, "totalUploaded"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(17, "totalNumKnownNonSeeds"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(18, "totalNumKnownSeeds"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(19, "hashFails"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(20, "UUIDid"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(21, "connectedSeeds"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(22, "connectedLeechers"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(23, "swarmPeerSpeed"));
		fields.get("downloadStatDynamicDasu.txt").get("DownloadStatsDynamic").add(new Pair<Integer, String>(24, "serverTimestamp"));

		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(0, "peerId"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(1, "downloadId"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(2, "timeRecorded"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(3, "downloadingTime"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(4, "seedingTime"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(5, "downloadRateLimit"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(6, "uploadRateLimit"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(7, "totalUploaded"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(8, "totalDownloaded"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(9, "size"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(10, "pieceSize"));
		fields.get("downloadStatSummaryDasu.txt").get("DownloadStatsSummary").add(new Pair<Integer, String>(11, "UUIDid"));
		// maxDownloadRateLimit=4000
		// maxUploadLimit=8000
		// priority
		// downloadName
		// 0, DownloadStatsSummary, downloadId=1, downloadingTime=100, seedingTime=40, maxDownloadRateLimit=4000, maxUploadLimit=8000, totalUploaded=4000, totalDownloaded=90000, size=1234, priority=-1, pieceSize=4192, downloadName=dummy;

		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(0, "peerId"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(1, "timeRecorded"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(2, "uploadRate"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(3, "downloadRate"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(4, "autoSeed"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(5, "rst"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(6, "active"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(7, "passive"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(8, "failed"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(9, "current"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(10, "UUIDid"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(11, "numTorrents"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(12, "numSeedingTorrents"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(13, "numLeechingTorrents"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(14, "numConnectedSeeds"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(15, "numConnectedLeechers"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(16, "bytesSentLpsd"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(17, "bytesReceivedLpsd"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(18, "bytesSentNetstat"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(19, "bytesReceivedNetstat"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(20, "swarmPeerSpeed"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(21, "UPnPNumberOfActiveConnections"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(22, "UPnPBytesSent"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(23, "UPnPBytesReceived"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(24, "UPnPPacketsSent"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(25, "UPnPPacketsReceived"));
		fields.get("localPeerStatDynamicDasu.txt").get("DynamicLocalPeerStat").add(new Pair<Integer, String>(26, "UPnPUptime"));
		// maxUploadRateLimitAllActiveTorrents
		// maxDownloadRateLimitAllActiveTorrents

		fields.get("localPeerStatDynamicDasu.txt").get("UPnPBWData").add(new Pair<Integer, String>(1, "timeRecorded"));
		fields.get("localPeerStatDynamicDasu.txt").get("UPnPBWData").add(new Pair<Integer, String>(21, "numberOfActiveConnections"));
		fields.get("localPeerStatDynamicDasu.txt").get("UPnPBWData").add(new Pair<Integer, String>(22, "totalBytesSent"));
		fields.get("localPeerStatDynamicDasu.txt").get("UPnPBWData").add(new Pair<Integer, String>(23, "totalBytesReceived"));
		fields.get("localPeerStatDynamicDasu.txt").get("UPnPBWData").add(new Pair<Integer, String>(24, "totalPacketsSent"));
		fields.get("localPeerStatDynamicDasu.txt").get("UPnPBWData").add(new Pair<Integer, String>(25, "totalPacketsReceived"));
		fields.get("localPeerStatDynamicDasu.txt").get("UPnPBWData").add(new Pair<Integer, String>(26, "uptime"));
		// 0, UPnPBWData, msg=no message, totalBytesSent=400, totalBytesReceived=500, numberOfActiveConnections=21, totalPacketsSent=1341, totalPacketsReceived=5531, uptime=13134, deltaTime=0, deltaBytesSent=0, deltaBytesReceived=0, isDebugMode=false, debugLines=null

		fields.get("localPeerStatDynamicDasu.txt").get("RSTPacketData").add(new Pair<Integer, String>(1, "timestamp"));
		fields.get("localPeerStatDynamicDasu.txt").get("RSTPacketData").add(new Pair<Integer, String>(5, "nConnReset"));
		fields.get("localPeerStatDynamicDasu.txt").get("RSTPacketData").add(new Pair<Integer, String>(6, "nActiveOpens"));
		fields.get("localPeerStatDynamicDasu.txt").get("RSTPacketData").add(new Pair<Integer, String>(7, "nActiveOpens"));
		fields.get("localPeerStatDynamicDasu.txt").get("RSTPacketData").add(new Pair<Integer, String>(8, "nFailedConnAttempt"));
		fields.get("localPeerStatDynamicDasu.txt").get("RSTPacketData").add(new Pair<Integer, String>(9, "nCurrentOpen"));
		// 0, RSTPacketData, isStored=false, timestamp=1, msg=no message, nConnReset=2, nActiveOpens=3, nPassiveOpens=4, nFailedConnAttempt=5, nCurrentOpen=6, deltaTime=0, deltaConnReset=0, deltaActiveOpens=0, deltaPassiveOpens=0, deltaFailedConnAttempt=0, percentRSTConn=0.0f, percentIncomingConn=0.0f, isDebugMode=false, debugLines=null

		fields.get("localPeerStatDynamicDasu.txt").get("NetstatBWData").add(new Pair<Integer, String>(1, "timeRecorded"));
		fields.get("localPeerStatDynamicDasu.txt").get("NetstatBWData").add(new Pair<Integer, String>(18, "bytesSent"));
		fields.get("localPeerStatDynamicDasu.txt").get("NetstatBWData").add(new Pair<Integer, String>(19, "bytesReceived"));
		// 0, NetstatBWData, msg=no message, bytesSent=200, bytesReceived=300, deltaTime=0, deltaBytesSent=0, deltaBytesReceived=0, isDebugMode=false, debugLines=null

		/*
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(0,"source"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(1,"peerIp"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(2,"downloadId"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(3,"timeRecorded"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(4,"type"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(5,"port"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(6,"bitTorrentClient"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(7,"foundSource"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(8,"isSeed"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(9,"amSeed"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(10,"numSeeds"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(11,"numPeers"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(12,"UUIDid"));
		fields.get("rstStatsDasu.txt").get("RSTStat").add(new Pair<Integer, String>(13,"encryption")); //numRst //0, RSTStat, downloadId=0,
		// peerIp=0, timeRecorded=0, numRst=1, type=1, bitTorrentClient=vuze, port=2000, source=source, isSeed=true, amSeed=false, numSeeds=20, numPeers=22, encryption=encryption

		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(0,"id"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer,String>(1,"uptime"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(2,"maxUploadRate")); //maxUploadRateGlobal
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(3,"maxDownloadRate"));//maxDownloadRateGlobal
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(4,"currentTime"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(5,"UUIDid"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(6,"maxPeerConnections")); //maxPeerConnectionsGlogal
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(7,"maxPeerConnectionsPerTorrent"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(8,"maxPeerConnectionsPerTorrentSeeding"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(9,"maxUploadRateSeeding"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(10,"maxSeedsPerTorrent"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer,String>(11,"lazyBitfield"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(12,"requireEncryption"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(13,"encryptionMinLevel"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(14,"allowNonEncryptedOut"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(15,"allowNonEncryptedIn"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(16,"useCryptoPortExtension"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer,String>(17,"tcpPortIn"));
		fields.get("localPeerStatStaticDasu.txt").get("StaticLocalPeerStat").add(new Pair<Integer, String>(18,"udpPortIn")); //configChanged 
		//0, StaticLocalPeerStat, uptime, maxUploadRate, maxDownloadRate, maxPeerConnections, maxPeerConnectionsPerTorrent, maxPeerConnectionsPerTorrentSeeding, maxUploadRateSeeding, maxSeedsPerTorrent, lazyBitfield,requireEncryption, encryptionMinLevel, allowNonEncryptedOut, allowNonEncryptedIn, useCryptoPortExtension, configChanged, tcpPortIn, udpPortIn
		  
		//TOO MANY UNKNOWNS
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(0,"peerId")); //localId
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(1,"remoteId")); //?
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(2,"peerClient")); //clientId
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(3,"currentLifetime")); //uptime
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(4,"tcpPort"));
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(5,"udpPort"));
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(6,"sameCluster"));
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(7,"currentTime")); //?
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(8,"UUIDid")); //?
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(9,"type"));
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(10,"encryption"));
		fields.get("remotePeerStatStaticDasu.txt").get("RemotePeerStatStatic").add(new Pair<Integer, String>(11,"pieceSize")); 
		//0, RemotePeerStatStatic, Peer peer, peerId, currentLifetime, ArrayList<RemotePeerStatDynamic> dynamicStats, tcpPort, udpPort, sameCluster, firstSeen, lastTraceroute, peerClient, done, encryption, type, pieceSize 

		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"lcoalId")); //?
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"remoteId")); //?
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"currentTime")); //time
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"bytesReceivedFromPeer")); //bytesReceived
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"bytesSentToPeer")); //bytesSent
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"currentReceiveRate")); //receiveRate
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"currentSendRate")); //sendRate
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"isSeed"));
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"rttPending")); //rtt
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"isInterested"));
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"isInteresting"));
		fields.get("remotePeerStatDynamic2.txt").get("RemotePeerStatDynamic").add(new Pair<Integer, String>(0,"UUIDid")); 
		//0, RemotePeerStatDynamic, isInteresting, isInterested, rttPending, ping, currentTime, bytesSentToPeer, bytesReceivedFromPeer, currentSendRate, currentReceiveRate, isSeed
		 
		// iterators.put("staticUPnPInfo.txt", "unk");
		*/
	}

	
	public void addToInputPlot(FactTemplate factObject) {
		
		/*
		String[] factDsd = new String[]{"timeRecorded", "downloadId", "numPeers", "availability", "completed", "downloadRate", "uploadRate", 
				"numKnownNonSeeds", "numKnownSeeds", "totalNumKnownNonSeeds", "totalNumKnownSeeds", "state", "priority", "downloaded", "uploaded", 
				"downloadName", "connectedSeeds", "connectedLeechers", "numPendingPeers", "swarmPeerSpeed", "numRst", "averageUploadNC",
				"averageNumPeersNC", "averageConnectedSeedsNC", "averageConnectedLeechersNC", "averageSwarmPeerSpeed", "averageDownload",
				"averageUpload", "averageNumPeers", "averageConnectedSeeds", "averageConnectedLeechers"};
				
		String[] factSystem = new String[]{"timeRecorded", "rst", "numConnections", "passiveOpens", "activeOpens", "failedConns", "bytesSent",
				"bytesReceived", "averageRst", "averageNumConnections", "averagePassiveOpen", "averageActiveOpen", "averageFailedCons",
		        "averageBytesSent", "averageBytesReceived"};
		        
		String[] factLpsd = new String[]{"timeRecorded", "uploadRate", "downloadRate", "maxUploadRate", "maxDownloadRate", "numTorrents",
			"numLeechingTorrents", "numSeedingTorrents", "numConnectedSeeds", "numConnectedLeechers", "bytesSent", "bytesReceived", longDeltaTime, 
			"shortDeltaTime", "swarmPeerSpeed", "maxPeerConnectionsGlobal", "maxPeerConnectionsPerTorrent", "maxPeerConnectionsPerTorrentSeeding", 
			"maxUploadRateSeeding", "maxSeedsPerTorrent", "encryptionMinLevel", "tcpPortIn", "udpPortIn", "maxUploadRateLimitAllActiveTorrents", "maxDownloadRateLimitAllActiveTorrents",
			"averageUpload", "averageDownload", "averageNumConnectedSeeds", "averageNumConnectedLeechers", "averageBytesSent", "averageBytesReceived",
			"averageSwarmPeerSpeed""};		        
		 */
		
		if ( factDsdFields.size() == 0) {
		
			factDsdFields.addAll(Arrays.asList(new String[]{"numPeers", "availability", "completed", "downloadRate", "uploadRate", 
					"numKnownNonSeeds", "numKnownSeeds", "totalNumKnownNonSeeds", "totalNumKnownSeeds", "downloaded", "uploaded", 
					"connectedSeeds", "connectedLeechers", "numPendingPeers", "swarmPeerSpeed", "numRst"}));
			factDsdLabels.addAll(Arrays.asList(new String[]{"(peers)", "(availability)", "(%)", "(kbps)", "(kbps)", 
					"(nonSeeds)", "(seeds)", "(nonSeeds)", "(seeds)", "(bytes)", "(bytes)", 
					"(seeds)", "(leechers)", "(peers)", "(bps)", "(rst)"}));
						
			factFields.put("FactMeasurementSystem", Arrays.asList(new String[]{"rst", "numConnections", "passiveOpens", "activeOpens", "failedConns", 
					"bytesSent", "bytesReceived"}));
			
			factLabels.put("FactMeasurementSystem", Arrays.asList(new String[]{"(rsts)", "(connections)", "(connections)", "(connections)", "(connections)", 
					"(bytes)", "(bytes)"}));
			
			
			factFields.put("FactMeasurementLpsd", Arrays.asList(new String[]{"uploadRate", "downloadRate", "maxUploadRate", "maxDownloadRate", "numTorrents",
					"numLeechingTorrents", "numSeedingTorrents", "numConnectedSeeds", "numConnectedLeechers", "bytesSent", "bytesReceived", 
					"swarmPeerSpeed", "maxPeerConnectionsGlobal", "maxPeerConnectionsPerTorrent", "maxPeerConnectionsPerTorrentSeeding", 
					"maxUploadRateSeeding", "maxSeedsPerTorrent", "maxUploadRateLimitAllActiveTorrents", "maxDownloadRateLimitAllActiveTorrents",
					"averageUpload", "averageDownload", "averageNumConnectedSeeds", "averageNumConnectedLeechers", "averageBytesSent", "averageBytesReceived",
					"averageSwarmPeerSpeed"}));
			
			factLabels.put("FactMeasurementLpsd", Arrays.asList(new String[]{"(bps)", "(bps)", "(bps)", "(bps)", "(torrents)",
					"(torrents)", "(torrents)", "(seeds)", "(leecers)", "(bps)", "(bps)", 
					"(bps)", "(connections)", "(conections)", "(connections)", 
					"(bps)", "(seeds)", "(bps)", "(bps)",
					"(bps)", "(bps)", "(seeds)", "(leecers)", "(bytes)", "(bytes)",
					"(bps)"}));
		}
				
		ArrayList<String> plotsToShow = (ArrayList<String>) DasuData.getInstance().get("simulator.update.plots");
		if ( plotsToShow == null ) plotsToShow = new ArrayList<String>();
		//System.out.println("\n\n\n\n\n\n\n\n\n\n\n\n\n\n" + plotsToShow + "\n\n\n\n\n\n\n\n\n\n\n\n");		
		
		if ( factObject instanceof FactMeasurementDsd ) {
						
			//System.out.println("\n\n\n\n\nSimulator: PLOT " + factObject.getClass() + " " + factObject + "\n");

			Long did = ((FactMeasurementDsd)factObject).getDownloadId();
			Long time = ((FactMeasurementDsd)factObject).getTimeRecorded();
			
			//initialize
			if (!factDsDPlotMap.containsKey(did)) {
				factDsDPlotMap.put(did, new PlotObject("FactMeasurementDsd", did.toString()));
				factDsDPlotMap.get(did).addFields(factDsdFields, factDsdLabels);
			}			
			
			factDsDPlotMap.get(did).addToPlot(factObject, time);
			
			HashMap<String, ArrayList<Object>> dsdPlots = null;
			for ( String signal : factDsdFields) {

				LinePlot plot = factDsDPlotMap.get(did).getPlot(signal);
				plot.enableAxisTitles();
				
				dsdPlots = (HashMap<String, ArrayList<Object>>) DasuData.getInstance().get(plot.getId());
				if ( dsdPlots == null ) {	
					dsdPlots = new HashMap();
				}

				dsdPlots.put(did.toString(), plot.generatePlot());

				if (plotsToShow.contains("FactMeasurementDsd_" + signal))
					DasuData.getInstance().update(plot.getId(), dsdPlots);

//				dsdPlots = (HashMap<String, ArrayList<Object>>) DasuData.getInstance().get(plot.getId());				
//				System.out.println("*********************\t" + signal);
//				for ( String d : dsdPlots.keySet()) {
//					System.out.println("*********************\t" + d + "\t" + dsdPlots.get(d));
//				}
			}
			
		}
		
		else {
		
			Long time = null;
			String type = null;
			
			if ( factObject instanceof FactMeasurementSystem ) {
				time = ((FactMeasurementSystem)factObject).getTimeRecorded();
				type = "FactMeasurementSystem";
			}
			else if ( factObject instanceof FactMeasurementLpsd ) {
				time = ((FactMeasurementLpsd)factObject).getTimeRecorded();
				type = "FactMeasurementLpsd";
			}

			if ( type != null) {
			
				//System.out.println("\n\n\n\n\nSimulator: PLOT " + factObject.getClass() + " " + factObject + "\n");
					
				if ( !factPlots.containsKey(type) ) {
					factPlots.put(type, new PlotObject(type));
					factPlots.get(type).addFields(factFields.get(type), factLabels.get(type));
				}
							
				factPlots.get(type).addToPlot(factObject, time);			
				for ( String signal : factFields.get(type)) {
		
					LinePlot plot = factPlots.get(type).getPlot(signal);
					if ( plot != null) {
						plot.enableAxisTitles();
						if (plotsToShow.contains(type + "_" + signal))
							DasuData.getInstance().update(plot.getId(), plot.generatePlot());
					}
				}
			}
		}
	}
	
}
