package edu.northwestern.dasu.util;

import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.net.SocketFactory;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLSocketFactory;

import edu.northwestern.cs.aqualab.cattle.socket.SelfSignedSSLSocketFactory;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.JSONObject;
import edu.northwestern.cs.aqualab.cattle.jsonsimple.parser.JSONParser;

/**
 * 
 */

public class HttpsChannel implements Serializable {

	private final static Logger LOGGER = Logger.getLogger("edu.northwestern.dasu");
	private static final long serialVersionUID = 1L;

    private static ArrayList<KeyStore> TRUSTED_KEYSTORES = new ArrayList<KeyStore>();
    private static SocketFactory SSL_FACTORY = SSLSocketFactory.getDefault();

    private long last_attempt = -1;
    private int num_attempts = 0;
    private String dest_host;
    private int dest_port = 443;
    private String resource;
    private List<JSONObject> beefs;


    /**
     * Attempt to add a KeyStore to the trusted list and reinitialize the
     * SSLSocketFactory. Restores the old factory on failure.
     *
     * @param ks
     *            The new KeyStore.
     * @return True if the new KeyStore was added successfully, false otherwise.
     */
    public static synchronized boolean addKeyStore(KeyStore ks) {
        TRUSTED_KEYSTORES.add(ks);
        SocketFactory old_factory = SSL_FACTORY;

        try {
            SSL_FACTORY = new SelfSignedSSLSocketFactory(TRUSTED_KEYSTORES);
        } catch (KeyManagementException kme) {
            SSL_FACTORY = old_factory;
            TRUSTED_KEYSTORES.remove(ks);
            return false;
        } catch (NoSuchAlgorithmException nsae) {
            SSL_FACTORY = old_factory;
            TRUSTED_KEYSTORES.remove(ks);
            return false;
        }

        return true;
    }

    public HttpsChannel(String hostname, Integer port, String resource) {

    	this.setDestinationHost(hostname);
    	this.setDestinationPort(port);
    	this.setResource(resource);
    	
//    	HostnameVerifier allHostsValid = new HostnameVerifier() {	    		
//    		public boolean verify(String arg0, SSLSession arg1) {
//    			return true;
//    		}				
//    	};
//    	HttpsURLConnection.setDefaultHostnameVerifier(allHostsValid);
    	
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(this.dest_host);
        sb.append(":");
        sb.append(String.valueOf(this.dest_port));
        sb.append(this.resource);
        return sb.toString();
    }


    /**
     * Must be called before send().
     */
    public void setDestinationHost(String dest) {
        this.dest_host = dest;
    }


    public void setDestinationPort(int port) {
        this.dest_port = port;
    }


    /**
     * Must be called before send().
     */
    public void setResource(String res) {
        this.resource = res;
    }


    /**
     * @return The timestamp, in millis, when the last call to GET() completed.
     */
    public long getLastAttempt() {
        return this.last_attempt;
    }


    /**
     * @return The number of times GET() has been called.
     */
    public int getNumAttempts() {
        return this.num_attempts;
    }

    public JSONObject GETData(String resource) {
    	
    	return GETData(this.resource + "/" + resource, this.dest_host, this.dest_port);
    	
    }
    
    public JSONObject GETData(String resource, String host, Integer port) {
    	
        JSONObject ret = null; 
        
    	try {
    		
    		LOGGER.info("CONFIG: Contacting " + host + ":" + port + "/" + resource);
    		
    		URL url = new URL(host + ":" + port + "/" + resource);
	    	HttpsURLConnection conn = (HttpsURLConnection)url.openConnection();
	    	conn.setSSLSocketFactory((SSLSocketFactory) SSL_FACTORY);

	    	InputStream inputstream = conn.getInputStream();
	    	InputStreamReader inputstreamreader = new InputStreamReader(inputstream);
	    	BufferedReader bufferedreader = new BufferedReader(inputstreamreader);
	
	    	String string = null;
	    	String res = new String();
	    	while ((string = bufferedreader.readLine()) != null) {	    	    
	    	    res = res.concat(string);
	    	}
	    	
	    	res = res.replace("'", "\"");
	    	LOGGER.info("CONFIG: Received (" + res + ")");
	        ret = (JSONObject)new JSONParser().parse(res);
	    	
    	}
    	catch (Exception e) {
    		LOGGER.info("CONFIG: ERROR: " + e.getMessage());
    	}
    	
        this.last_attempt = System.currentTimeMillis();
        ++this.num_attempts;
    	return ret;
    }
    
    public boolean GETFile(String resource, String file) {
    	
    	boolean ret = true;
		int chunkSize=1024;
        InputStream inputstream = null;
        OutputStream outStream = null;
                
    	try {

    		URL url = new URL(this.dest_host + ":" + this.dest_port + "/" + this.resource + "/" + resource);
	    	HttpsURLConnection conn = (HttpsURLConnection)url.openConnection();
	    	conn.setSSLSocketFactory((SSLSocketFactory) SSL_FACTORY);
	    	inputstream = conn.getInputStream();	    	
			outStream = new BufferedOutputStream(new FileOutputStream(file));

			byte[] buf;
			int ByteRead;

			buf = new byte[chunkSize];
			while ((ByteRead = inputstream.read(buf)) != -1) {
				outStream.write(buf, 0, ByteRead);
			}

			LOGGER.info("CONFIG: Downloaded Successfully (" + url + ")");
	    	
    	}
    	catch (Exception e) {
			LOGGER.info("CONFIG: downloadFileFromURL: Timeout: " + 
					this.dest_host + ":" + this.dest_port + "/" + this.resource + "/" + resource);
			ret = false;
			
		} finally {
			try {

				if (inputstream != null)
					inputstream.close();
				if(outStream != null) {
					outStream.close();
				}
				
			} catch (IOException e) {
				LOGGER.warning("CONFIG: downloadFileFromURL: File not found!: " + file);
				ret = false;
			}
		}
    	
		this.last_attempt = System.currentTimeMillis();
        ++this.num_attempts;
		return ret;
    	
    }

    /**
     * Clear the JSONObjects associated with this Report.
     */
    public void clear() {
        this.beefs = new ArrayList<JSONObject>();
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
        out.writeLong(serialVersionUID);
        out.writeLong(this.last_attempt);
        out.writeInt(this.num_attempts);
        out.writeObject(this.dest_host);
        out.writeInt(this.dest_port);
        out.writeObject(this.resource);
        out.writeInt(this.beefs.size());
        for (int i = 0; i < this.beefs.size(); ++i) {
            out.writeObject(this.beefs.get(i).toString());
        }
    }
    
    public static void main(String[] args) {
    	
    	try {
    		
//    		URL url = new URL("https://signup.netflix.com/nfapi?endpoint=utility/openconnect");
//	    	HttpsURLConnection conn = (HttpsURLConnection)url.openConnection();
//	    	System.out.println(conn.getHeaderField("Set-Cookie"));
//	    	
//	    	String cookie = conn.getHeaderField("Set-Cookie");
//	    	conn = (HttpsURLConnection)url.openConnection();
//	    	conn.setRequestProperty("Cookie", cookie);
//	    	//List<String> cookies = conn.getHeaderFields().get("Set-Cookie");
//	    	
//	    	
//	    	//for (String cookie : cookies) {
//		    //	conn.setRequestProperty("Cookie", cookie);
//	    	//}	        
//
//	    	conn.connect();	        
//	    	InputStream inputstream = conn.getInputStream();
//	    	InputStreamReader inputstreamreader = new InputStreamReader(inputstream);
//	    	BufferedReader bufferedreader = new BufferedReader(inputstreamreader);
//	
//	    	String string = null;
//	    	String res = new String();
//	    	while ((string = bufferedreader.readLine()) != null) {	    	    
//	    	    res = res.concat(string);
//	    	}
//	    	
//	    	LOGGER.info("CONFIG: Received (" + res + ")");
//	    	System.out.println(res);
//
//    		
//	    	KeyStore ks = KeyStore.getInstance("JKS");
//	    	FileInputStream is = new FileInputStream(new File("/Users/mas939/Mario's Stuff/NORTHWESTERN/Dropbox/workspace/ISP Interference/Dasu-Client/dist/dasu_keyStore.jks"));
//	        ks.load(is, "fadazani".toCharArray());
//	        HttpsChannel.addKeyStore(ks);
//	        
//	        HttpsChannel https = new HttpsChannel("https://mas939-imac.cs.northwestern.edu", 443, "config/data/dasu/dynamic/");
//	    	https.GETData("UpdateConfig.php");
////	        HttpsChannel https = new HttpsChannel("https://google.com", 443, "");
////	    	https.GETData("");
    		
	        HttpsChannel https = new HttpsChannel("https://mlab-ns.appspot.com", 443, "");
	    	https.GETData("ndt?format=json");
    		

    	}
    	catch (Exception e) {
    		e.printStackTrace();
    	}
        

    }
}
